/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.entity;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.JobStatus;
import p2pMapReduce.mapreduceModule.TaskID;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.util.LoggingUtil;

/**
 * The backup node view of a job
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class BackupJob implements Serializable {

    /**
     * The associated job
     */
    private Job job;
    /**
     * The id of the primary node
     */
    private String primaryID;
    /**
     * User node info
     */
    private NodeInfo userNodeInfo;
    /**
     * The list of all task attempt submitted for this job by the primary
     */
    private LinkedList<JobTask> jobTaskList;
    /**
     * List of all the node that acts as backup for this job
     */
    private LinkedList<String> backupMasterList;

    public BackupJob(PrimaryJob primaryJob, String primaryID) {
        this.job = primaryJob.getJob();
        this.userNodeInfo = primaryJob.getUserNodeInfo();
        this.jobTaskList = primaryJob.getJobTaskList();
        this.backupMasterList = primaryJob.getBackupMasterList();
        this.primaryID = primaryID;
    }

    public String getJobID() {
        return job.getJobID().toString();
    }

    public LinkedList<String> getBackupMasterList() {
        return backupMasterList;
    }

    public void setBackupMasterList(LinkedList<String> backupMasterList) {
        this.backupMasterList = backupMasterList;
    }

    public Job getJob() {
        return job;
    }

    public void setJob(Job job) {
        this.job = job;
    }

    public JobStatus.State getJobState() {
        return job.getJobState();
    }

    public void setJobState(JobStatus.State jobState) {
        job.setJobState(jobState);
    }

    public LinkedList<JobTask> getJobTaskList() {
        return jobTaskList;
    }

    public void setJobTaskList(LinkedList<JobTask> jobTaskList) {
        this.jobTaskList = jobTaskList;
    }

    public String getPrimaryID() {
        return primaryID;
    }

    public void setPrimaryID(String primaryID) {
        this.primaryID = primaryID;
    }

    public NodeInfo getUserNodeInfo() {
        return userNodeInfo;
    }

    public void setUserNodeInfo(NodeInfo userNodeInfo) {
        this.userNodeInfo = userNodeInfo;
    }

    public void update(JobTask updatedJobTask) {
        //TODO ottimizzare passando la lista di tutti i task updated e ricercare
        //direttamente quelli
        for (Iterator<JobTask> it = jobTaskList.iterator(); it.hasNext();) {
            JobTask jt = it.next();
            if (jt.equals(updatedJobTask)) {
                jt.setSlaveID(updatedJobTask.getSlaveID());
                jt.setTaskStatus(updatedJobTask.getTaskStatus());
                jt.setTask(updatedJobTask.getTask());
                jt.setOutputRetrieved(updatedJobTask.isOutputRetrieved());
                return;
            }
        }
        //Not founded, new JobTask
        jobTaskList.add(updatedJobTask);

    }

    /**
     * Used to update job progress.
     * TODO: used an adhoc structure like JobManager data structure
     */
    public void updateProgress() {

        Set<TaskID> mapTask = new HashSet<TaskID>();
        Set<TaskID> successfulMapTask = new HashSet<TaskID>();
        Set<TaskID> successfulReduceTask = new HashSet<TaskID>();

        if (!jobTaskList.isEmpty()) {
            setJobState(JobStatus.State.RUNNING);
            for (JobTask jobTask : jobTaskList) {
                if (jobTask.getTaskStatus() == Task.TaskStatus.SUCCESS) {
                    if (jobTask.isMap()) {
                        successfulMapTask.add(jobTask.getTask().getTaskID());
                    } else {
                        successfulReduceTask.add(jobTask.getTask().getTaskID());
                    }
                }
                if (jobTask.isMap()) {
                    mapTask.add(jobTask.getTask().getTaskID());
                }
            }

            float mapP = successfulMapTask.size();
            mapP /= mapTask.size();
            float redP = successfulReduceTask.size();
            redP /= job.getNumReduceTasks();

            job.getStatus().setMapProgress(mapP);
            job.getStatus().setReduceProgress(redP);
        }


    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("BackupJob [jobID:").append(job.getJobID()).append(" jobState:").append(job.getJobState()).append(" userID:").append(userNodeInfo.getID()).append("\n");
        sb.append("JobTaskList=\n").append(LoggingUtil.elementByLine(jobTaskList));
        sb.append("BackupMasterList=\n").append(LoggingUtil.elementByLine(backupMasterList));
        return sb.toString();
    }
}
