/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.entity;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.channels.FileChannel;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.fs.FileSystemView;
import p2pMapReduce.ftp.FTPManager;
import p2pMapReduce.mapreduceModule.MRJobConfig;
import p2pMapReduce.mapreduceModule.MapContext;
import p2pMapReduce.mapreduceModule.MapContextImpl;
import p2pMapReduce.mapreduceModule.Mapper;
import p2pMapReduce.mapreduceModule.Path;
import p2pMapReduce.mapreduceModule.util.ReflectionUtils;
import p2pMapReduce.mapreduceModule.TaskAttemptContext;
import p2pMapReduce.mapreduceModule.TaskAttemptContextImpl;
import p2pMapReduce.mapreduceModule.TaskAttemptID;
import p2pMapReduce.mapreduceModule.WrappedMapper;
import p2pMapReduce.mapreduceModule.conf.Configuration;
import p2pMapReduce.mapreduceModule.input.InputFormat;
import p2pMapReduce.mapreduceModule.input.InputSplit;
import p2pMapReduce.mapreduceModule.input.RecordReader;
import p2pMapReduce.mapreduceModule.RawComparator;
import p2pMapReduce.mapreduceModule.io.serializer.Deserializer;
import p2pMapReduce.mapreduceModule.io.serializer.SerializationFactory;
import p2pMapReduce.mapreduceModule.io.serializer.Serializer;
import p2pMapReduce.mapreduceModule.output.MapRecordWriter;
import p2pMapReduce.mapreduceModule.output.RecordWriter;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.nodeModule.NodeModuleIF;
import p2pMapReduce.entity.Task.TaskStatus;

/**
 *
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class MapTaskRunner extends TaskRunner {

    private static final transient Logger LOG = Logger.getLogger(MapTaskRunner.class.getName());
    private String LOG_PREFIX;

    public MapTaskRunner(SlaveTask slaveTask, NodeModuleIF nodeModule) {
        super(slaveTask, nodeModule);
        LOG_PREFIX = "MapTaskRunner for " + slaveTask.getTask().getTaskAttIDAsString();
    }

    public TaskStatus call() throws Exception {

        //Represents the parallel state EXECUTION
        setupEnv();
        InputSplit inputSplit = retrieveInput();

        runMapper(inputSplit);
        //local sort phase
        boolean success = sortOutput();
        if (success) {
            return TaskStatus.SUCCESS;
        } else {
            LOG.log(Level.WARNING, "{0} failed to sorting output", new Object[]{LOG_PREFIX});
            return TaskStatus.FAILED;
        }

    }

    ///////////////////////////////////////////////////////////////////////////
    //  Private methods
    ///////////////////////////////////////////////////////////////////////////
    /**
     *
     */
    private void setupEnv() {
        LOG.log(Level.INFO, LOG_PREFIX + "Setup environment");
        taskAttContext = new TaskAttemptContextImpl(slaveTask.getTask().getConf(), slaveTask.getTask().getTaskAttID());
        updateWorkingDir();

    }

    /**
     * Sets the job working dir as {@link FTPManager#getHomeDir() }
     * @param job
     */
    private void updateWorkingDir() {
        FileSystemView fsv = nodeModule.getFileSystemView();
        String workingDir = fsv.getRootDir();
        //TODO valutare se utilizzare o meno path e se introdurre in metodo
        //set working dir in JobContext
        taskAttContext.getConfiguration().set(MRJobConfig.WORKING_DIR, new Path(workingDir).toString());
    }

    private InputSplit retrieveInput() throws FileNotFoundException, IOException, Exception {

        LOG.log(Level.INFO, LOG_PREFIX + "Retrieving input");
        NetworkModuleIF netModule = nodeModule.getNetworkModule();
        NodeInfo primary = netModule.find(slaveTask.getPrimaryID());
        FTPManager ftpM = nodeModule.getFtpManager();
        FileSystemView fsv = nodeModule.getFileSystemView();
        Configuration conf = slaveTask.getTask().getConf();

        String inputSplitPath = conf.get(MRJobConfig.JOB_INPUTSPLIT_DIR_ATTR) + slaveTask.getTask().getPartition() + "_split.txt";

        boolean success = ftpM.downloadFile(primary, inputSplitPath);
        if (!success) {
            throw new RuntimeException("Unable to retrive input split from " + primary.getIpAddress() + " at " + primary.getFtpPort() + " on " + slaveTask.getPrimaryID());
        }

        String inputSplitDesc = inputSplitPath + InputSplit.DESCRIPTOR_SUFFIX;
        success = ftpM.downloadFile(primary, inputSplitDesc);
        if (!success) {
            throw new RuntimeException("Unable to retrive input split desc from " + primary.getIpAddress() + " at " + primary.getFtpPort() + " on " + slaveTask.getPrimaryID());
        }
        String jarsPath = conf.get(MRJobConfig.JOB_JAR_DIR_ATTR);
        success = ftpM.downloadDir(primary, jarsPath);
        if (!success) {
            throw new RuntimeException("Unable to retrive jars from " + primary.getIpAddress() + " at " + primary.getFtpPort() + " on " + slaveTask.getPrimaryID());
        }
        updateClassPath(fsv.createPath(jarsPath));
        Class<? extends InputSplit> inputSplitClass = conf.getClass(MRJobConfig.INPUTSPLIT_CLASS_ATTR, null, InputSplit.class);
        InputSplit split = ReflectionUtils.newInstance(inputSplitClass, conf);

        DataInputStream dis = new DataInputStream(new FileInputStream(fsv.createPath(inputSplitDesc)));
        split.readFields(dis);

        LOG.log(Level.INFO, "{0} Input retrieved", LOG_PREFIX);

        return split;
    }

    private <INKEY, INVALUE, OUTKEY, OUTVALUE> void runMapper(InputSplit split) throws IOException, ClassNotFoundException, InterruptedException {
        // make a task context so we can get the classes
        Configuration conf = taskAttContext.getConfiguration();
        TaskAttemptID taskAttID = slaveTask.getTask().getTaskAttID();

        TaskAttemptContext taskContext = new TaskAttemptContextImpl(conf, taskAttID);
        // make a mapper
        Mapper<INKEY, INVALUE, OUTKEY, OUTVALUE> mapper = (Mapper<INKEY, INVALUE, OUTKEY, OUTVALUE>) ReflectionUtils.newInstance(taskContext.getMapperClass(), conf);
        // make the input format
        InputFormat<INKEY, INVALUE> inputFormat = (InputFormat<INKEY, INVALUE>) ReflectionUtils.newInstance(taskContext.getInputFormatClass(), conf);
        // rebuild the input split
        RecordReader<INKEY, INVALUE> input = inputFormat.createRecordReader(split, taskContext);
        //            job.setBoolean(JobContext.SKIP_RECORDS, isSkipping());
        RecordWriter output = null;
        // get an output object
        //            if (job.getNumReduceTasks() == 0) {
        //                output = new NewDirectOutputCollector(taskContext, job, umbilical, reporter);
        //            } else {
        //                output = new NewOutputCollector(taskContext, job, umbilical, reporter);
        //            }
        output = new MapRecordWriter(taskAttID, taskContext, taskContext.getMapOutputKeyClass(), taskContext.getMapOutputValueClass());
        MapContext<INKEY, INVALUE, OUTKEY, OUTVALUE> mapContext = new MapContextImpl<INKEY, INVALUE, OUTKEY, OUTVALUE>(conf, taskAttID, input, output, null, null, split);
        Mapper<INKEY, INVALUE, OUTKEY, OUTVALUE>.Context mapperContext = new WrappedMapper<INKEY, INVALUE, OUTKEY, OUTVALUE>().getMapContext(mapContext);
        input.initialize(split, mapperContext);
        mapper.run(mapperContext);
        //            mapPhase.complete();
        //            setPhase(TaskStatus.Phase.SORT);
        //            statusUpdate(umbilical);
        input.close();
        output.close(mapperContext);
    }

    /**
     * Reads the temporary map output, sort it by key and
     * write the definitive map output in the map output dir
     *
     */
    public <KEYIN, VALUEIN> boolean sortOutput() throws IOException {

        Configuration c = taskAttContext.getConfiguration();
        int numMapTask = taskAttContext.getTaskAttemptID().getTaskID().getId();

        Class<KEYIN> keyInClass = (Class<KEYIN>) taskAttContext.getMapOutputKeyClass();
        Class<VALUEIN> valueInClass = (Class<VALUEIN>) taskAttContext.getMapOutputValueClass();

        SerializationFactory sf = new SerializationFactory(c);
        Deserializer<KEYIN> keyDeserializer = sf.getDeserializer(keyInClass);
        Deserializer<VALUEIN> valueDeserializer = sf.getDeserializer(valueInClass);
        Serializer<KEYIN> keySerializer = sf.getSerializer(keyInClass);
        Serializer<VALUEIN> valueSerializer = sf.getSerializer(valueInClass);

        for (int numPart = 0; numPart < taskAttContext.getNumReduceTasks(); numPart++) {
            //For each partition try to write the relative file in the output dir
            String inputPartPath = taskAttContext.getWorkingDirectory() + "/" + c.get(MRJobConfig.TASK_TEMP_DIR) + "map/" + taskAttContext.getTaskAttemptID().toString() + "/";
            inputPartPath += "partition_" + numPart + "_" + numMapTask + ".tmp";
            String outputPartTmpPath = taskAttContext.getWorkingDirectory() + "/" + c.get(MRJobConfig.TASK_TEMP_DIR) + "map/" + taskAttContext.getTaskAttemptID().toString() + "/";
            outputPartTmpPath += "partition_" + numPart + "_" + numMapTask;
            String outputPartDir = taskAttContext.getWorkingDirectory() + "/" + c.get(MRJobConfig.MAP_OUTPUT_DIR_ATTR);
            String outputPartPath = outputPartDir + "partition_" + numPart + "_" + numMapTask;
            String outputPartCheckPath = outputPartPath + ".check";

            File inputPartFile = new File(inputPartPath);
            File outputPartTmpFile = new File(outputPartTmpPath);
            File outputPartFile = new File(outputPartPath);
            File outputPartCheckFile = new File(outputPartCheckPath);

            if (outputPartCheckFile.exists()) {
                //An output for this map task has already been written
                continue;
            } else {
                //Create sorted output file
                createSortedOutput(inputPartFile, outputPartTmpFile, keyDeserializer, valueDeserializer, keySerializer, valueSerializer, keyInClass, valueInClass, c);

                //Try to write output file
                FileChannel src = null;
                FileChannel dst = null;
                try {
                    //create map output dir if it not exists
                    File mapOutputDir = new File(outputPartDir);
                    mapOutputDir.mkdir();

                    src = new FileInputStream(outputPartTmpFile).getChannel();
                    dst = new FileOutputStream(outputPartFile).getChannel();
                    dst.transferFrom(src, 0, src.size());
                    /*
                     * Atomic creation of the file, if the file was not already
                     * created
                     */
                    outputPartCheckFile.createNewFile();

                } catch (IOException ex) {
                    LOG.log(Level.INFO, "{0} Failed to write output file: {1}", new Object[]{LOG_PREFIX, ex.getLocalizedMessage()});
                    return false;
                } finally {
                    if (src != null) {
                        src.close();
                    }
                    if (dst != null) {
                        dst.close();
                    }
                }

            }
        }

        return true;
    }

    /**
     * Reads all the key.value pairs contained in the inputPartFile
     * and write them ordered by keys in outputPartTmpFile
     *
     * @param inputPartFile
     * @param outputPartTmpFile
     */
    private <KEYIN, VALUEIN> void createSortedOutput(
            File inputPartFile,
            File outputPartTmpFile,
            Deserializer<KEYIN> keyDeserializer,
            Deserializer<VALUEIN> valueDeserializer,
            Serializer<KEYIN> keySerializer,
            Serializer<VALUEIN> valueSerializer,
            Class<KEYIN> keyInClass,
            Class<VALUEIN> valueInClass,
            Configuration c)
            throws IOException {

        FileInputStream fis = new FileInputStream(inputPartFile);
        DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));
        keyDeserializer.open(dis);
        valueDeserializer.open(dis);

        // Sort the output
        LOG.log(Level.INFO, "{0} Sort output", new Object[]{LOG_PREFIX});
        RawComparator<KEYIN> keyComparator = (RawComparator<KEYIN>) taskAttContext.getSortComparator();
        TreeMap<KEYIN, LinkedList<VALUEIN>> keyValuesMap = new TreeMap<KEYIN, LinkedList<VALUEIN>>(keyComparator);
        //Reads all key-value pairs and put them into the ordered map
        while (dis.available() > 0) {
            KEYIN key = null;
            VALUEIN value = null;
            key = keyDeserializer.deserialize(key);
            value = valueDeserializer.deserialize(value);
            LinkedList<VALUEIN> list = keyValuesMap.get(key);
            if (list == null) {
                list = new LinkedList<VALUEIN>();
                keyValuesMap.put(key, list);
            }
            list.add(value);
        }
        dis.close();
        LOG.log(Level.INFO, "{0} output sorted", new Object[]{LOG_PREFIX});

        LOG.log(Level.INFO, "{0} Starting to write sorted output", new Object[]{LOG_PREFIX});
        //Save ordered key,value pairs
        DataOutputStream dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(outputPartTmpFile)));
        keySerializer.open(dos);
        valueSerializer.open(dos);
//        File x = new File(outputPartTmpFile.getParent(), outputPartTmpFile.getName() + ".txt");
        //TODO remove pw
//        PrintWriter pw = new PrintWriter(x);
        for (Entry<KEYIN, LinkedList<VALUEIN>> entry : keyValuesMap.entrySet()) {
            for (VALUEIN v : entry.getValue()) {
                keySerializer.serialize(entry.getKey());
                valueSerializer.serialize(v);
//                pw.println(entry.getKey() + " " + v);
            }
        }
        dos.close();
//        pw.close();
        LOG.log(Level.INFO, "{0} Sorted output writed", new Object[]{LOG_PREFIX});

    }
}
