/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.entity;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.channels.FileChannel;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.entity.JobTask;
import p2pMapReduce.fs.FileSystemView;
import p2pMapReduce.ftp.FTPManager;
import p2pMapReduce.mapreduceModule.Counter;
import p2pMapReduce.mapreduceModule.MRJobConfig;
import p2pMapReduce.mapreduceModule.OutputCommitter;
import p2pMapReduce.mapreduceModule.Path;
import p2pMapReduce.mapreduceModule.ReduceContext;
import p2pMapReduce.mapreduceModule.ReduceContextImpl;
import p2pMapReduce.mapreduceModule.Reducer;
import p2pMapReduce.mapreduceModule.util.ReflectionUtils;
import p2pMapReduce.mapreduceModule.StatusReporter;
import p2pMapReduce.mapreduceModule.TaskAttemptContextImpl;
import p2pMapReduce.mapreduceModule.TaskAttemptID;
import p2pMapReduce.mapreduceModule.TaskID;
import p2pMapReduce.mapreduceModule.WrappedReducer;
import p2pMapReduce.mapreduceModule.conf.Configuration;
import p2pMapReduce.mapreduceModule.RawComparator;
import p2pMapReduce.mapreduceModule.io.serializer.Deserializer;
import p2pMapReduce.mapreduceModule.io.serializer.SerializationFactory;
import p2pMapReduce.mapreduceModule.io.serializer.Serializer;
import p2pMapReduce.mapreduceModule.output.OutputFormat;
import p2pMapReduce.mapreduceModule.output.RecordWriter;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.nodeModule.NodeModuleIF;
import p2pMapReduce.entity.Task.TaskStatus;

/**
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class ReduceTaskRunner extends TaskRunner {

    private static final transient Logger LOG = Logger.getLogger(ReduceTaskRunner.class.getName());
    private String LOG_PREFIX;

    public ReduceTaskRunner(SlaveTask slaveTask, NodeModuleIF nodeModule) {
        super(slaveTask, nodeModule);
        LOG_PREFIX = "ReduceTaskRunner for " + slaveTask.getTask().getTaskAttIDAsString();
    }

    public TaskStatus call() throws Exception {
        setupEnv();
        Map<TaskID, File> retrievedInputs = retrieveInput();
        Configuration configuration = taskAttContext.getConfiguration();
        RawComparator keyComparator = taskAttContext.getSortComparator();

        String partFilePath = taskAttContext.getWorkingDirectory() + "/" + configuration.get(MRJobConfig.REDUCE_INPUT_DIR_ATTR);
        partFilePath += "partition_" + taskAttContext.getTaskAttemptID().getTaskID().getId();
        String partCheckFilePath = partFilePath + ".check";

        File partFile = new File(partFilePath);
        File partCheckFile = new File(partCheckFilePath);

        if (!partCheckFile.exists()) {
            partFile.getParentFile().mkdirs();
            merge(retrievedInputs, partFile, keyComparator);
            /*
             * Atomic creation of the file, if the file was not already
             * created
             */
            partCheckFile.createNewFile();
        }

        LOG.log(Level.INFO, "{0} Starting  reducer", new Object[]{LOG_PREFIX});
        runReducer(partFilePath, keyComparator);
        LOG.log(Level.INFO, "{0} Reducer ended", new Object[]{LOG_PREFIX});
        return TaskStatus.SUCCESS;
//        LOG.log(Level.INFO, "{0} Starting upload output file", new Object[]{LOG_PREFIX});
//        boolean success = uploadOutput();
//        if (success) {
//            LOG.log(Level.INFO, "{0} Output file uploaded", new Object[]{LOG_PREFIX});
//            return TaskStatus.SUCCESS;
//        } else {
//            return TaskStatus.FAILED;
//        }


    }

    ///////////////////////////////////////////////////////////////////////////
    //  Private methods
    ///////////////////////////////////////////////////////////////////////////
    /**
     *
     */
    private void setupEnv() {
        LOG.log(Level.INFO, LOG_PREFIX + "Setup environment");
        taskAttContext = new TaskAttemptContextImpl(slaveTask.getTask().getConf(), slaveTask.getTask().getTaskAttID());
        updateWorkingDir();

    }

    /**
     * Sets the job working dir as {@link FTPManager#getHomeDir() }
     * @param job
     */
    private void updateWorkingDir() {
        FileSystemView fsv = nodeModule.getFileSystemView();
        String workingDir = fsv.getRootDir();
        taskAttContext.getConfiguration().set(MRJobConfig.WORKING_DIR, new Path(workingDir).toString());
    }

    /**
     * For each map task retrieves the associate output from a node where
     * the map task has been executed with success.
     * The output of map tasks executed by this node are retrieved by a local
     * copy.
     * Others are retrieved by remote FTP download.
     *
     *
     * @return A File refers to the output of a map task
     * @throws FileNotFoundException
     * @throws IOException
     */
    private Map<TaskID, File> retrieveInput() throws FileNotFoundException, IOException, Exception {

        LOG.log(Level.INFO, "{0} Retrieving input", new Object[]{LOG_PREFIX});

        NetworkModuleIF netModule = nodeModule.getNetworkModule();
        NodeInfo primary = netModule.find(slaveTask.getPrimaryID());
        FTPManager ftpM = nodeModule.getFtpManager();
        FileSystemView fsv = nodeModule.getFileSystemView();
        ReduceTask task = (ReduceTask) slaveTask.getTask();
        Configuration conf = taskAttContext.getConfiguration();

        String jarsPath = conf.get(MRJobConfig.JOB_JAR_DIR_ATTR);
        boolean success = ftpM.downloadDir(primary, jarsPath);
        if (!success) {
            throw new RuntimeException("Unable to retrive jars from " + primary.getIpAddress() + " at " + primary.getFtpPort() + " on " + slaveTask.getPrimaryID());
        }
        updateClassPath(fsv.createPath(jarsPath));

        //TODO sostituire con costante
        String destDir = conf.get(MRJobConfig.TASK_TEMP_DIR) + "/reduce/" + taskAttContext.getTaskAttemptID() + "/";
        fsv.mkdirs(destDir);

        HashMap<TaskID, TreeSet<JobTask>> successfullMapTaskAttempt = task.getSuccessfullMapTaskAttempt();
        Map<TaskID, File> ris = new HashMap<TaskID, File>(successfullMapTaskAttempt.size());

        //XXX performance optimization
        //in order to not continuosly search the network for slave node info
        //let's perform a global search
        //FIXME allSlaves or allNodes ?? a slave could be become master?
        Set<NodeInfo> nodes = nodeModule.getNetworkModule().getAllNodes();

        //TODO optimization: make the search with a thread pool
        for (Map.Entry<TaskID, TreeSet<JobTask>> entry : successfullMapTaskAttempt.entrySet()) {
            TaskID taskID = entry.getKey();
            TreeSet<JobTask> mapTasks = entry.getValue();
            long start = System.currentTimeMillis();
            File f = retrieve(taskID, mapTasks, destDir, nodes);
            LOG.log(Level.INFO, "{0} retrieving time for {1}: {2} ms", new Object[]{LOG_PREFIX, taskID, (System.currentTimeMillis() - start)});
            if (f != null) {
                ris.put(taskID, f);
            } else {
                throw new RuntimeException("Unable to retrive input file for " + taskID);
            }
        }

        LOG.log(Level.INFO, "{0} input retrieved", new Object[]{LOG_PREFIX});
        return ris;
    }

    private <INKEY, INVALUE, OUTKEY, OUTVALUE> void runReducer(String partitionFile, RawComparator<INKEY> comparator) throws IOException, InterruptedException, ClassNotFoundException {

        Configuration conf = taskAttContext.getConfiguration();
        // make a reducer
        Reducer<INKEY, INVALUE, OUTKEY, OUTVALUE> reducer = (Reducer<INKEY, INVALUE, OUTKEY, OUTVALUE>) ReflectionUtils.newInstance(taskAttContext.getReducerClass(), conf);
        OutputFormat<OUTKEY, OUTVALUE> outputFormat = (OutputFormat<OUTKEY, OUTVALUE>) ReflectionUtils.newInstance(taskAttContext.getOutputFormatClass(), conf);
        RecordWriter<OUTKEY, OUTVALUE> output = (RecordWriter<OUTKEY, OUTVALUE>) outputFormat.getRecordWriter(taskAttContext);
        //            RecordWriter<OUTKEY, OUTVALUE> trackedRW = new NewTrackingRecordWriter<OUTKEY, OUTVALUE>(output, reduceOutputCounter);
        //            job.setBoolean("mapred.skip.on", isSkipping());
        //            job.setBoolean(JobContext.SKIP_RECORDS, isSkipping());
        Reducer.Context reducerContext = createReduceContext(reducer, conf, taskAttContext.getTaskAttemptID(), partitionFile, null, null, output, null, null, comparator, (Class<INKEY>) taskAttContext.getMapOutputKeyClass(), (Class<INVALUE>) taskAttContext.getMapOutputValueClass());
        reducer.run(reducerContext);
        output.close(reducerContext);
    }

    protected static <INKEY, INVALUE, OUTKEY, OUTVALUE> Reducer<INKEY, INVALUE, OUTKEY, OUTVALUE>.Context createReduceContext(
            Reducer<INKEY, INVALUE, OUTKEY, OUTVALUE> reducer,
            Configuration job,
            TaskAttemptID taskId,
            String partitionFile,
            Counter inputKeyCounter,
            Counter inputValueCounter,
            RecordWriter<OUTKEY, OUTVALUE> output,
            OutputCommitter committer,
            StatusReporter reporter,
            RawComparator<INKEY> comparator,
            Class<INKEY> keyClass,
            Class<INVALUE> valueClass)
            throws IOException, InterruptedException {
        //            ReduceContext<INKEY, INVALUE, OUTKEY, OUTVALUE> reduceContext = new ReduceContextImpl2<INKEY, INVALUE, OUTKEY, OUTVALUE>(
        //                    job, taskId,
        //                    rIter,
        //                    inputKeyCounter,
        //                    inputValueCounter,
        //                    output,
        //                    committer,
        //                    reporter,
        //                    comparator,
        //                    keyClass,
        //                    valueClass);
        ReduceContext<INKEY, INVALUE, OUTKEY, OUTVALUE> reduceContext = new ReduceContextImpl<INKEY, INVALUE, OUTKEY, OUTVALUE>(job, taskId, partitionFile, output, committer, reporter, keyClass, valueClass);
        Reducer<INKEY, INVALUE, OUTKEY, OUTVALUE>.Context reducerContext = new WrappedReducer<INKEY, INVALUE, OUTKEY, OUTVALUE>().getReducerContext(reduceContext);
        return reducerContext;
    }

    /**
     * Retrieves the map output from one of the available node in mapTasks set
     * Prefers this node if available
     * 
     * @param taskID
     * @param mapTasks
     * @return
     */
    private File retrieve(TaskID taskID, TreeSet<JobTask> mapTasks, String destDir, Set<NodeInfo> nodes) throws IOException {
        NodeInfo thisNode = nodeModule.getNodeInfo();

        for (JobTask jobTask : mapTasks) {
            if (jobTask.getSlaveID().equals(thisNode.getID())) {
                //The output to retrieve is available locally
                return retrieveLocal(taskID, destDir);
            }
        }

        //No local available
        Iterator<JobTask> it = mapTasks.iterator();
        for (int i = 0; i < mapTasks.size(); i++) {
            JobTask selected = it.next();
            NodeInfo nodeInfo = find(selected.getSlaveID(), nodes);
            if (nodeInfo == null) {
                LOG.log(Level.INFO, "{0} unable to find " + selected.getSlaveID() + ".Try a direct search", new Object[]{LOG_PREFIX});
                nodeInfo = nodeModule.getNetworkModule().find(selected.getSlaveID());
                if (nodeInfo == null) {
                    LOG.log(Level.INFO, "{0} unable to find " + selected.getSlaveID(), new Object[]{LOG_PREFIX});
                    continue;
                }
            }
            File f = retrieveRemote(taskID, nodeInfo, destDir);
            if (f != null) {
                return f;
            }
        }

        return null;
    }

    private File retrieveLocal(TaskID taskID, String destDir) throws IOException {
        //FIXME if it is local need to copy?
        //TODO optimization: why not using map output file direcly ??
        LOG.log(Level.INFO, "{0} local retrieving for {1}", new Object[]{LOG_PREFIX, taskID});
        int reduceIndex = slaveTask.getTask().getTaskID().getId();
        FileSystemView fsv = nodeModule.getFileSystemView();

        String mapOutputDir = taskAttContext.getConfiguration().get(MRJobConfig.MAP_OUTPUT_DIR_ATTR);
        String fileName = "partition_" + reduceIndex + "_" + taskID.getId();
        String srcFilePath = mapOutputDir + fileName;
        String destFilePath = destDir + fileName;

        File srcFile = fsv.createFile(srcFilePath);
        File destFile = fsv.createFile(destFilePath);

        FileChannel src = new FileInputStream(srcFile).getChannel();
        FileChannel dest = new FileOutputStream(destFile).getChannel();

        //TODO check that all the bytes have been written
        long byteWritten = dest.transferFrom(src, 0, src.size());

        src.close();
        dest.close();

        return destFile;
    }

    private File retrieveRemote(TaskID taskID, NodeInfo slaveNode, String destDir) throws IOException {

        LOG.log(Level.INFO, "{0} remote retrieving for {1}", new Object[]{LOG_PREFIX, taskID});
        int reduceIndex = slaveTask.getTask().getTaskID().getId();
        FileSystemView fsv = nodeModule.getFileSystemView();
        FTPManager ftpM = nodeModule.getFtpManager();

        String mapOutputDir = taskAttContext.getConfiguration().get(MRJobConfig.MAP_OUTPUT_DIR_ATTR);
        String fileName = "partition_" + reduceIndex + "_" + taskID.getId();
        String srcFilePath = mapOutputDir + fileName;
        String destFilePath = destDir + fileName;

        boolean success = ftpM.downloadFile(slaveNode, srcFilePath, destFilePath);

        if (success) {
            return fsv.createFile(destFilePath);
        } else {
            return null;
        }
    }

    /**
     * 
     * @param <KEYIN>
     * @param <VALUEIN>
     * @param retrievedInputs
     * @param dest
     * @param keyComparator
     * @throws IOException
     */
    private <KEYIN, VALUEIN> void merge(Map<TaskID, File> retrievedInputs, File dest, RawComparator<KEYIN> keyComparator) throws IOException {
        //TODO optimization: every single map output file is already ordered: use a n-merger
        LOG.log(Level.INFO, "{0} Merging map output files", new Object[]{LOG_PREFIX});
        Configuration c = taskAttContext.getConfiguration();
        Class<KEYIN> keyInClass = (Class<KEYIN>) taskAttContext.getMapOutputKeyClass();
        Class<VALUEIN> valueInClass = (Class<VALUEIN>) taskAttContext.getMapOutputValueClass();
        SerializationFactory sf = new SerializationFactory(c);
        Deserializer<KEYIN> keyDeserializer = sf.getDeserializer(keyInClass);
        Deserializer<VALUEIN> valueDeserializer = sf.getDeserializer(valueInClass);
        Serializer<KEYIN> keySerializer = sf.getSerializer(keyInClass);
        Serializer<VALUEIN> valueSerializer = sf.getSerializer(valueInClass);

        Collection<File> files = retrievedInputs.values();

        TreeMap<KEYIN, LinkedList<VALUEIN>> keyValuesMap = new TreeMap<KEYIN, LinkedList<VALUEIN>>(keyComparator);
        for (File mapOutput : files) {
            DataInputStream dis = new DataInputStream(new BufferedInputStream(new FileInputStream(mapOutput)));
            keyDeserializer.open(dis);
            valueDeserializer.open(dis);
            KEYIN key = null;
            VALUEIN value = null;
            while (dis.available() > 0) {
                key = keyDeserializer.deserialize(null);
                value = valueDeserializer.deserialize(null);
                LinkedList<VALUEIN> list = keyValuesMap.get(key);
                if (list == null) {
                    list = new LinkedList<VALUEIN>();
                    keyValuesMap.put(key, list);
                }
                list.add(value);
            }
            dis.close();
        }

        DataOutputStream dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(dest)));
        keySerializer.open(dos);
        valueSerializer.open(dos);
        for (Iterator<Entry<KEYIN, LinkedList<VALUEIN>>> it = keyValuesMap.entrySet().iterator(); it.hasNext();) {
            Entry<KEYIN, LinkedList<VALUEIN>> entry = it.next();
            KEYIN key = entry.getKey();
//                System.out.println("Writing " + key + " " + entry.getValue());
            for (Iterator<VALUEIN> it1 = entry.getValue().iterator(); it1.hasNext();) {
                VALUEIN value = it1.next();
                keySerializer.serialize(key);
                valueSerializer.serialize(value);
            }
        }
        dos.close();

    }

    private NodeInfo find(String slaveID, Set<NodeInfo> nodes) {
        for (NodeInfo nodeInfo : nodes) {
            if (nodeInfo.getID().equals(slaveID)) {
                return nodeInfo;
            }
        }

        return null;
    }
}
