/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.impl.ftp;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.net.ftp.FTP;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPFile;
import org.apache.ftpserver.ConnectionConfig;
import org.apache.ftpserver.ConnectionConfigFactory;
import org.apache.ftpserver.FtpServer;
import org.apache.ftpserver.FtpServerFactory;
import org.apache.ftpserver.ftplet.FtpException;
import org.apache.ftpserver.ftplet.UserManager;
import org.apache.ftpserver.listener.Listener;
import org.apache.ftpserver.listener.ListenerFactory;
import org.apache.ftpserver.usermanager.Md5PasswordEncryptor;
import org.apache.ftpserver.usermanager.PropertiesUserManagerFactory;
import org.apache.log4j.BasicConfigurator;
import p2pMapReduce.ftp.FTPManagerFactory;
import p2pMapReduce.ftp.FTPManager;
import p2pMapReduce.networkModule.NodeInfo;

/**
 * An implementation of the {@link FTPManager} interface based on apache
 * libraries.
 * An instance could be obtained by {@link FTPManagerFactory#createFTPManager() }
 *
 * Every download path parameters must be a relative path
 * that refers to local ftp base directory as current directory.
 * No absolute paths are allowed
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class FTPManagerImpl implements FTPManager {

    private static final transient Logger LOG = Logger.getLogger(FTPManagerImpl.class.getName());
    private File userPropFile;
    private String ipAddress;
    private int ftpPort;
    private FtpServerFactory serverFactory;
    private FtpServer localFtpServer;
    private String user = "admin";
    private String pass = "admin";
    private String baseDir;

    public FTPManagerImpl(File userPropFile, int ftpPort) {
        this.userPropFile = userPropFile;
        this.ftpPort = ftpPort;
    }

    public void startLocalServer() throws IOException {
        try {
            serverFactory = new FtpServerFactory();
            //User configuration
            PropertiesUserManagerFactory userManagerFactory = new PropertiesUserManagerFactory();
            userManagerFactory.setFile(userPropFile);
            userManagerFactory.setPasswordEncryptor(new Md5PasswordEncryptor());
            UserManager userManager = userManagerFactory.createUserManager();
            String homeDir = userManager.getUserByName(user).getHomeDirectory();

//            File homeDirFile = new File(homeDir);
//            if (homeDirFile.exists()) {
//                for (File file : homeDirFile.listFiles()) {
//                    Util.recursiveDelete(file);
//                }
//            }
//            this.baseDir = homeDir + "/" + baseDir + "/";

            this.baseDir = homeDir;
            File baseDirFile = new File(this.baseDir);
            baseDirFile.mkdirs();

//            User u = userManager.getUserByName(user);
//            UserFactory userFactory = new UserFactory();
//            userFactory.setEnabled(true);
//            userFactory.setHomeDirectory(baseDir);
//            userFactory.setAuthorities(u.getAuthorities());
//            userFactory.setName(u.getName());
//            userFactory.setPassword(u.getPassword());
//            userFactory.setMaxIdleTime(u.getMaxIdleTime());
//            User current = userFactory.createUser();
//            userManager.save(current);
            serverFactory.setUserManager(userManager);
            //Connection configuration
            ConnectionConfigFactory ccf = new ConnectionConfigFactory();
            ccf.setAnonymousLoginEnabled(false);
            //By default max login is 10 also if 0 is specified in users.properties
            ccf.setMaxLogins(100);
            serverFactory.setConnectionConfig(ccf.createConnectionConfig());
            //Listener
            ListenerFactory factory = new ListenerFactory();
            factory.setPort(ftpPort);
            Listener createListener = factory.createListener();
            serverFactory.addListener("default", createListener);
            //Starting server
            localFtpServer = serverFactory.createServer();
            localFtpServer.start();

            setLocalAddress();

        } catch (FtpException ex) {
            LOG.log(Level.SEVERE, ex.getLocalizedMessage(), ex);
            throw new IOException(ex.getLocalizedMessage(), ex);
        }

    }

    private void setLocalAddress() {
        try {
            InetAddress localHostIA = InetAddress.getLocalHost();
            //                System.out.println("localHostIA = " + localHostIA);
            //                System.out.println("localHostIA host name = " + localHostIA.getHostName());
            byte[] address = localHostIA.getAddress();
            //                System.out.println(InetAddress.getByAddress(address));
            String canonicalName = localHostIA.getCanonicalHostName();
            //                System.out.println("canonicalName = " + canonicalName);
            InetAddress[] allByName = InetAddress.getAllByName(canonicalName);
            for (InetAddress inetAddress : allByName) {
                //                    System.out.println("inetAddress = " + inetAddress);
                if (!inetAddress.isLoopbackAddress() && !(inetAddress instanceof Inet6Address)) {
                    ipAddress = inetAddress.getHostAddress();
                }
            }
            if (ipAddress == null) {
                ipAddress = localHostIA.getHostAddress();
            }
            //                System.out.println("ipAddress = " + ipAddress);
            //
            //                Socket s = new Socket("localhost", ftpPort);
            //                System.out.println("inet address "+ s.getInetAddress());
            //                System.out.println("remote socket address "+s.getRemoteSocketAddress());
        } catch (UnknownHostException ex) {
            LOG.log(Level.SEVERE, ex.getLocalizedMessage(), ex);
        }
    }

    /**
     * Returns the path to the directory used as root directory by FTP server
     * 
     * @return The path to FTP server root directory, {@code null} if the
     * server has not been stared yet
     */
    public String getBaseDir() {
        if (localFtpServer == null) {
            return null;
        }
        return baseDir;
    }

    /**
     * Find for a {@link InetAddress} which is not a loopback address
     * and return {@link InetAddress#getHostAddress()} result.
     * If all {@link InetAddress} are loopback addresses utilize
     * {@link InetAddress#getLocalHost() }
     * 
     * @return The string representation of the ip address for this FTP server
     */
    public String getIpAddress() {


        return ipAddress;
    }

    /**
     * The local port on which FTP server listen
     * @return
     */
    public int getFtpPort() {
        return ftpPort;
    }

    public boolean downloadFile(NodeInfo remoteNode, String filePath) {
        return downloadFile(remoteNode, filePath, filePath);
    }

    public boolean downloadFile(NodeInfo remoteNode, String srcFilePath, String dstFilePath) {
        return downloadFile(remoteNode.getIpAddress(), remoteNode.getFtpPort(), srcFilePath, dstFilePath);
    }

    /**
     * Download the remote file in a local file which path is the same
     * of the remote file, relatively to the FTP home directory
     * If necessary the directory tree of the file is created
     * 
     * @param host
     * @param port
     * @param filePath
     * @return
     */
    public boolean downloadFile(String host, int port, String filePath) {
        return downloadFile(host, port, filePath, filePath);
    }

    /**
     * Download the remote file in a local file which path is {@code dstFilePath}
     * relatively to the FTP home directory.
     * If necessary the directory tree of the file is created
     *
     * @param host
     * @param port
     * @param filePath
     * @param dstFilePath
     * @return
     */
    public boolean downloadFile(String host, int port, String srcFilePath, String dstFilePath) {
        FTPClient ftpClient = null;
        OutputStream os = null;
        try {
            ftpClient = login(host, port);
            if (ftpClient != null) {
                File dest = getFile(dstFilePath);
                dest.getParentFile().mkdirs();
                FTPFile ftpf = ftpClient.listFiles(srcFilePath)[0];
                if (dest.createNewFile() || dest.length() != ftpf.getSize()) {
                    os = new BufferedOutputStream(new FileOutputStream(dest));
                    return ftpClient.retrieveFile(srcFilePath, os);
                }
                return true;

            } else {
                return false;
            }
        } catch (SocketException ex) {
            LOG.log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, null, ex);
        } finally {
            if (ftpClient != null && ftpClient.isConnected()) {
                try {
                    closeFTPClient(ftpClient);
                } catch (IOException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
            if (os != null) {
                try {
                    os.close();
                } catch (IOException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
        }

        return false;
    }

    public boolean downloadDir(NodeInfo remoteNode, String dirPath) {
        return downloadDir(remoteNode, dirPath, dirPath);
    }

    public boolean downloadDir(NodeInfo remoteNode, String dirPath, String destDir) {
        return downloadDir(remoteNode.getIpAddress(), remoteNode.getFtpPort(), dirPath, destDir);
    }

    /**
     * Download the remote directory in a local directory which path is the same
     * of the remote directory, relatively to the FTP home directory
     * If necessary the directory is created
     *
     * @param host
     * @param port
     * @param dirPath
     * @return
     */
    public boolean downloadDir(String host, int port, String dirPath) {

        return downloadDir(host, port, dirPath, dirPath);
    }

    /**
     * Download all the content from remote directory dir in the local
     * directory destDir
     * 
     * @param host
     * @param port
     * @param dir
     * @param destDir
     * @return
     */
    public boolean downloadDir(String host, int port, String dir, String destDir) {
        FTPClient ftpClient = null;
        try {
            ftpClient = login(host, port);
            if (ftpClient != null) {
                boolean fullSuccess = true;
                getFile(destDir).mkdir();
                FTPFile names[] = ftpClient.listFiles(dir);
                for (FTPFile fTPFile : names) {
                    recursiveTransfer(fTPFile, dir, destDir, ftpClient);
                }
                return fullSuccess;
            } else {
                return false;
            }
        } catch (SocketException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            if (ftpClient != null && ftpClient.isConnected()) {
                try {
                    closeFTPClient(ftpClient);
                } catch (IOException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
        }

        return false;
    }

    private void recursiveTransfer(FTPFile fTPFile, String curDir, String destDir, FTPClient ftpClient) throws IOException {
        if (fTPFile.isFile()) {
            File file = getFile(destDir + fTPFile.getName());
            /*
             * XXX Per evitare di sorvrascrivere file già esistenti si controlla
             * prima che il file non esista.
             * Se esiste si scarica solo se non è completo
             */
            if (file.createNewFile() || file.length() != fTPFile.getSize()) {
                OutputStream os = null;
                try {
                    os = new BufferedOutputStream(new FileOutputStream(file));
                    ftpClient.retrieveFile(curDir + fTPFile.getName(), os);
                } catch (IOException ex) {
                    LOG.log(Level.INFO, ex.getLocalizedMessage(), ex);
                } finally {
                    if (os != null) {
                        os.close();
                    }
                }
            }
        } else {
            String dest = destDir + fTPFile.getName() + "/";
            getFile(dest).mkdir();
            String current = curDir + fTPFile.getName() + "/";
            FTPFile files[] = ftpClient.listFiles(current);
            for (FTPFile fTPFile1 : files) {
                recursiveTransfer(fTPFile1, current, dest, ftpClient);
            }
        }
    }

    /**
     * 
     * @param host
     * @param port
     * @param user
     * @param pass
     * @param path
     * @param dest
     * @param results
     * @deprecated
     * @return
     */
    public boolean download(String host, int port, String user, String pass, String[] path, File[] dest, boolean[] results) {
        FTPClient ftpClient = null;
        try {
            ftpClient = new FTPClient();
            ftpClient.connect(host, port);
            boolean logged = ftpClient.login(user, pass);
            if (logged) {
                boolean fullSuccess = true;
                for (int i = 0; i < results.length; i++) {

                    results[i] = ftpClient.retrieveFile(path[i], new FileOutputStream(dest[i]));
                    if (!results[i]) {
                        fullSuccess = false;
                    }
                }
                return fullSuccess;
            } else {
                return false;
            }
        } catch (SocketException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            if (ftpClient != null && ftpClient.isConnected()) {
                try {
                    closeFTPClient(ftpClient);
                } catch (IOException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
        }

        return false;
    }

    /**
     * 
     * @param host
     * @param port
     * @param user
     * @param pass
     * @param path
     * @param dest
     * @deprecated 
     * @return
     */
    public boolean download(String host, int port, String user, String pass, String path, File dest) {
        FTPClient ftpClient = null;
        try {
            ftpClient = new FTPClient();
            ftpClient.connect(host, port);
            boolean logged = ftpClient.login(user, pass);
            if (logged) {

                return ftpClient.retrieveFile(path, new FileOutputStream(dest));
            } else {
                return false;
            }
        } catch (SocketException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            if (ftpClient != null && ftpClient.isConnected()) {
                try {
                    closeFTPClient(ftpClient);
                } catch (IOException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
        }

        return false;

    }

    public boolean uploadFile(NodeInfo remoteNode, String srcFilePath, String destPath, String destFileName) {
        return uploadFile(remoteNode.getIpAddress(), remoteNode.getFtpPort(), srcFilePath, destPath, destFileName);
    }

    /**
     * @param srcFilePath File da uplodare
     * @param dest Path del file (dove salvarlo)
     * @param dim la dimensione del file
     * @return True se l'upload è stato eseguito con successo
     */
    public boolean uploadFile(String host, int port, String srcFilePath, String destPath, String destFileName) {

        FTPClient ftpClient = null;
        FileInputStream fis = null;
        try {
            ftpClient = login(host, port);
            if (ftpClient != null) {
//                ftpClient.makeDirectory(destPath);
                fis = new FileInputStream(srcFilePath);
                return ftpClient.storeFile(destPath + "/" + destFileName, fis);

            } else {
                return false;
            }
        } catch (SocketException ex) {
            LOG.log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, null, ex);
        } finally {
            if (ftpClient != null && ftpClient.isConnected()) {
                try {
                    closeFTPClient(ftpClient);
                } catch (IOException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
            if (fis != null) {
                try {
                    fis.close();
                } catch (IOException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
        }

        return false;
    }

    /**
     * 
     * @param host
     * @param port
     * @param user
     * @param pass
     * @param srcFiles
     * @param destPath
     * @param destFileNames
     * @param results
     * @deprecated to test
     * @return
     */
    public boolean upload(String host, int port, String user, String pass, File[] srcFiles, String destPath, String[] destFileNames, boolean[] results) {
        try {
            FTPClient ftpClient = new FTPClient();
            ftpClient.connect(host, port);
            boolean logged = ftpClient.login(user, pass);
            if (logged) {
                ftpClient.setFileType(FTP.BINARY_FILE_TYPE);
                boolean fullSuccess = false;
                if (ftpClient.changeWorkingDirectory(destPath)) {
                    for (int i = 0; i < results.length; i++) {
                        results[i] = ftpClient.storeFile(destFileNames[i], new FileInputStream(srcFiles[i]));
                        fullSuccess = fullSuccess && results[i];
                    }

                }

                return fullSuccess;
            }

        } catch (SocketException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        }

        return false;
    }

    /**
     * 
     * @param host
     * @param port
     * @param user
     * @param pass
     * @param srcFiles
     * @param destPaths
     * @param destFileNames
     * @param results
     * @deprecated to test
     * @return
     */
    public boolean upload(String host, int port, String user, String pass, File[] srcFiles, String[] destPaths, String[] destFileNames, boolean[] results) {
        try {
            FTPClient ftpClient = new FTPClient();
            ftpClient.connect(host, port);
            boolean logged = ftpClient.login(user, pass);
            if (logged) {
                ftpClient.setFileType(FTP.BINARY_FILE_TYPE);
                boolean fullSuccess = true;
                for (int i = 0; i < results.length; i++) {
                    if (ftpClient.changeWorkingDirectory(destPaths[i])) {
                        results[i] = ftpClient.storeFile(destFileNames[i], new FileInputStream(srcFiles[i]));
                        if (!results[i]) {
                            fullSuccess = false;
                        }
                    } else {
                        fullSuccess = false;
                    }
                }
                return fullSuccess;
            }

        } catch (SocketException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(FTPManagerImpl.class.getName()).log(Level.SEVERE, null, ex);
        }

        return false;
    }

    private FTPClient login(String host, int port) {
        FTPClient ris = null;
        try {
            ris = new FTPClient();
//            ris.setSocketFactory(null)
            ris.connect(host, port);
            boolean logged = ris.login(user, pass);
            if (!logged) {
                ris.disconnect();
                return null;
            }
            boolean succ = ris.setFileType(FTPClient.BINARY_FILE_TYPE);
//            /**
//             * FIXME
//             * on hided network the simple ftp doesn't work.
//             * It is needed to enterLocalPassiveMode to made it worked.
//             * Needs to be investigate
//             */
//            ris.enterLocalPassiveMode();
            if (!succ) {
                closeFTPClient(ris);
                return null;
            }

            //TODO if possible add COMPRESSED TRANSFER MODE
//            succ = ris.setFileTransferMode(FTP.COMPRESSED_TRANSFER_MODE);
//            if (!succ) {
//                ris.disconnect();
//                return null;
//            }

            return ris;

        } catch (IOException ex) {
            LOG.log(Level.SEVERE, null, ex);
        }

        return null;
    }

    private void closeFTPClient(FTPClient ftpClient) throws IOException {
        ftpClient.logout();
        ftpClient.disconnect();
    }

    private File getFile(String filePath) {
        File dest = new File(getBaseDir() + "/" + filePath);
        return dest;
    }

    private static void setlog() {

        BasicConfigurator.configure();

    }

    private void show(ConnectionConfig connectionConfig) {
        System.out.println("maxLogins = " + connectionConfig.getMaxLogins());
        System.out.println("login failure delay = " + connectionConfig.getLoginFailureDelay());
        System.out.println("maxThreads = " + connectionConfig.getMaxThreads());
    }
}
