/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.impl.networkModule.jxta;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.Serializable;
import java.net.URI;
import java.util.LinkedList;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.jxta.endpoint.EndpointAddress;
import net.jxta.endpoint.Message;
import net.jxta.endpoint.Messenger;
import net.jxta.endpoint.MessengerStateListener;
import net.jxta.peer.PeerID;
import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.PipeMsgEvent;
import net.jxta.pipe.PipeMsgListener;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.util.JxtaBiDiPipe;
import net.jxta.util.PipeEventListener;
import net.jxta.util.PipeStateListener;
import p2pMapReduce.networkModule.ConnectionIF;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.nodeModule.event.EventIF;
import p2pMapReduce.nodeModule.event.InnerEvent;
import p2pMapReduce.nodeModule.event.EventNotifier;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.MessageEvent;
import p2pMapReduce.util.LoggingUtil;

/**
 * During the creation of a JxtaBidiPipe an exchange of the node info is made
 * between the two connection endpoint.
 * If no node info is received after a timeout the constructor throws
 * an IOException.
 *
 * After that the node info exchanging was completed successfully a thread is
 * started to check that the remote endpoint is alive. This thread periodically
 * exchanges a message with the remote endpoint. If no remote response
 * is received in time the connection is marked as broken.
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class JxtaBidiPipeConnection implements ConnectionIF, PipeMsgListener, PipeEventListener, MessengerStateListener {

    private static final transient Logger LOG = Logger.getLogger(JxtaBidiPipeConnection.class.getName());
    /**
     * Time to wait for receiving remote node info.
     * If it is not received in time an {@link IOException} is thrown
     *
     * //TODO tuning this parameter and provide setter method
     */
    public static final int CONNECTION_OPENING_TIMEOUT = 5 * 1000;
    /**
     * @deprecated 
     */
    public static final int RECEIVE_TIMEOUT = 10 * 1000;
    /**
     * Time period of the connection check
     */
    public static final int CHECK_TIMEOUT = 60 * 1000;
    public static final String MESSAGE_ELEMENT_NAME = "cont";
    public static final String MESSAGE_NAMESPACE_NAME = "P2MapReduce_message_namespace";
    PeerGroup peerGroup;
    NodeInfo localNodeInfo;
    NodeInfo remoteNodeInfo;
    NodeInfo receivedNodeInfo;
    /**
     * XXX From JXTA documentation:
     * Sending messages vis sendMessage(Message) from within a PipeMsgListener
     * may result in a deadlock due to contention between
     * the sending and receiving portions of BiDi pipes.
     */
    JxtaBiDiPipe bidiPipe;
    JxtaNetworkModule netModule;
    EventNotifier eventNotifier;
    Notifier notifier;
    ConnectionHandler connectionHandler;
    /**
     * Indicates if this connection was started by this local node
     */
    boolean startedLocally;
    boolean broken;
    boolean disconnected;
    boolean nodeInfoExchanged;
    boolean pipeOpenEventReceived;
    final Object remoteLock = new Object();
    final Object openLock = new Object();

    /**
     * Use to open a new connection towards the specified pipe adv
     * @param peerGroup
     * @param pipeADV
     * @throws IOException
     */
    public JxtaBidiPipeConnection(PeerGroup peerGroup, NodeInfo local, NodeInfo remote, PipeAdvertisement pipeADV, JxtaNetworkModule netModule, EventNotifier eventNotifier) throws IOException {
        this.peerGroup = peerGroup;
        this.localNodeInfo = local;
        this.remoteNodeInfo = remote;
        this.netModule = netModule;
        this.eventNotifier = eventNotifier;
        this.startedLocally = true;
        this.notifier = new Notifier();
        this.connectionHandler = new ConnectionHandler();
        //XXX in jxta 2.7 is required to specify a msg listener on creation
        this.bidiPipe = new JxtaBiDiPipe(peerGroup, pipeADV, CONNECTION_OPENING_TIMEOUT, this, true);
        this.bidiPipe.setPipeEventListener(this);

        waitPipeOpenEvent();

        nodeInfoExchange();

        init();

    }

    /**
     * Used when a remote connection is received
     * @param bipipe
     */
    public JxtaBidiPipeConnection(JxtaBiDiPipe bipipe, PeerGroup peerGroup, NodeInfo local, JxtaNetworkModule netModule, EventNotifier eventNotifier) throws IOException {
        this.bidiPipe = bipipe;
        this.peerGroup = peerGroup;
        this.localNodeInfo = local;
        this.netModule = netModule;
        this.eventNotifier = eventNotifier;
        this.startedLocally = false;
        this.notifier = new Notifier();
        this.connectionHandler = new ConnectionHandler();
        this.bidiPipe.setMessageListener(this);
        this.bidiPipe.setPipeEventListener(this);
        //FIXME needed ?
//        try {
//            LOG.log(Level.CONFIG, "Receiving remote NodeInfo");
//            remoteNodeInfo = (NodeInfo) receive(0);
//            if (remoteNodeInfo == null) {
//                throw new IOException("Error while receiving remote NodeInfo (NodeInfo is null)");
//            }
//            LOG.log(Level.CONFIG, "remote NodeInfo received\n= " + remoteNodeInfo);
//        } catch (Exception e) {
//            e.printStackTrace();
//            throw new IOException("Error while receiving remote NodeInfo", e);
//        }

//        waitPipeOpenEvent();

        nodeInfoExchange();

        init();

    }

    private void nodeInfoExchange() throws IOException {
//        if (remoteNodeInfo != null) {
        sendLocalNodeInfo();
        receiveRemoteNodeInfo();

        nodeInfoExchanged = true;
        LOG.log(Level.CONFIG, "Node info exchanged");
    }

    private void receiveRemoteNodeInfo() throws IOException {
        long remainingTime = CONNECTION_OPENING_TIMEOUT;
        synchronized (remoteLock) {
            while (receivedNodeInfo == null) {
                try {
                    LOG.log(Level.INFO, "Waiting remote node info");
                    long start = System.currentTimeMillis();
                    remoteLock.wait(remainingTime);
                    remainingTime -= (System.currentTimeMillis() - start);
                    if (receivedNodeInfo != null) {
                        LOG.log(Level.CONFIG, "Remote node info received");
                    }
                    if (remainingTime <= 0) {
                        if (receivedNodeInfo == null) {
                            throw new IOException("Remote nodeinfo not received in time");
                        }
                    }
                } catch (InterruptedException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
        }

    }

    private void sendLocalNodeInfo() throws IOException {
        //Send and receive
        LOG.log(Level.CONFIG, "Sending local NodeInfo");
        MessageEvent me = new MessageEvent(EventType.REMOTE_NODEINFO, localNodeInfo);
        send(me);
        LOG.log(Level.CONFIG, "Local NodeInfo sended");
    }

    private void init() {
        //FIXME these are null in jxta 2.7
//        System.out.println("Bidipipe remote peer adv\n" + bidiPipe.getRemotePeerAdvertisement());
//        System.out.println("Bidipipe remote pipe adv\n" + bidiPipe.getRemotePipeAdvertisement());
//        System.out.println("Adding listener to messenger");
//        PeerID remoteId = PeerID.create(URI.create(remoteNodeInfo.getID()));
//        EndpointAddress addr = new EndpointAddress(remoteId, null, null);
//        Messenger mess = peerGroup.getEndpointService().getMessenger(addr);
//        mess.addStateListener(this);
        connectionHandler.start();
        notifier.start();
    }

    public void send(Serializable o) throws IOException {
        LOG.log(Level.INFO, "Sending {0} to {1}", new Object[]{o, remoteNodeInfo == null ? "null" : remoteNodeInfo.getID()});
        Message message = new Message();
        JxtaUtil.addObjectToMessage(message, MESSAGE_NAMESPACE_NAME, MESSAGE_ELEMENT_NAME, o);
        boolean success = bidiPipe.sendMessage(message);
        if (!success) {
            LOG.log(Level.SEVERE, "Failed to send {0} to {1}", new Object[]{o, remoteNodeInfo.getID()});
            throw new IOException("Jxta failed to enqueue the message!");
        }
    }

    public void pipeMsgEvent(PipeMsgEvent event) {

        try {
            Object o = JxtaUtil.getObjectFromMessage(event.getMessage(), MESSAGE_NAMESPACE_NAME, MESSAGE_ELEMENT_NAME);

            if (o instanceof MessageEvent) {
                LOG.log(Level.INFO, "Received message event: {0} from {1}", new Object[]{o, remoteNodeInfo == null ? null : remoteNodeInfo.getID()});
                MessageEvent me = (MessageEvent) o;
                if (me.getType() == EventType.REMOTE_NODEINFO) {
                    synchronized (remoteLock) {
                        receivedNodeInfo = (NodeInfo) me.getParam(0);
                        remoteNodeInfo = receivedNodeInfo;
                        LOG.log(Level.CONFIG, "Remote node info received " + receivedNodeInfo);
                        remoteLock.notifyAll();
                    }
                } else if (!nodeInfoExchanged) {
                    /*
                     * May be that the remote node info is already received, but the thread
                     * which is handling the receiving method has not completed it's work yet.
                     * Anyway enqueue the message that will be handled when the receiving method
                     * has finished
                     */
                    LOG.log(Level.SEVERE, "No remote node info received. Notify anyway");
                    notifyEvent((EventIF) o);
                } else {
                    notifyEvent((EventIF) o);
                }
            } else if (o instanceof EchoMessage) {
                LOG.log(Level.FINE, "Received an echo message: " + o);
                connectionHandler.handle((EchoMessage) o);

            } else {

                LOG.log(Level.FINE, "Received generic object {0}", o);
                if (remoteNodeInfo == null) {
                    LOG.log(Level.SEVERE, "No remote nodeinfo received");
                }
            }
        } catch (IOException ex) {
            ex.printStackTrace();
            LOG.log(Level.SEVERE, null, ex);
        } catch (ClassNotFoundException ex) {
            ex.printStackTrace();
            LOG.log(Level.SEVERE, null, ex);
        }

    }

    public Serializable receive() throws IOException {
        return receive(RECEIVE_TIMEOUT);
    }

    private Serializable receive(int timeout) throws IOException {
        try {
            //XXX se è impostato un msgListener ritorna subito null
            Message m = bidiPipe.getMessage(timeout);
            return m == null ? null : (Serializable) JxtaUtil.getObjectFromMessage(m, MESSAGE_NAMESPACE_NAME, MESSAGE_ELEMENT_NAME);

        } catch (ClassNotFoundException ex) {
            ex.printStackTrace();
            LOG.log(Level.SEVERE, null, ex);
        } catch (InterruptedException ex) {
            ex.printStackTrace();
            LOG.log(Level.SEVERE, null, ex);
        }

        return null;
    }

    public void close() throws IOException {
        bidiPipe.close();
        disconnected = true;
    }

    public boolean isDisconnected() {
        return disconnected;
    }

    public boolean isBroken() {
        return broken;
    }

    public NodeInfo getRemoteNodeInfo() {
        return remoteNodeInfo;
    }

    public void setEventNotifier(EventNotifier eventNotifier) {
        this.eventNotifier = eventNotifier;
    }

    public void pipeEvent(int event) {
        switch (event) {
            case PipeStateListener.PIPE_OPENED_EVENT:
                /*
                 * This event is received only on the side where the connection has been started
                 */
                synchronized (openLock) {
                    LOG.log(Level.INFO, "PIPE OPENED");
                    pipeOpenEventReceived = true;
                    openLock.notifyAll();
                }
                break;
            case PipeStateListener.PIPE_CLOSED_EVENT:
                LOG.log(Level.INFO, "PIPE CLOSED");
                notifyClose(broken);
                disconnected = true;
                break;
            case PipeStateListener.PIPE_FAILED_EVENT:
                LOG.log(Level.SEVERE, "PIPE FAILED");
                break;
            default:
                throw new AssertionError();
        }
    }

    public boolean messengerStateChanged(int newState) {
//        System.out.println("MESSENGER STATE CHANGE TO " + toString(newState));
        return true;//se sono interressato ai prossimi cambiamenti
    }

    private boolean check() {
        //FIXME jxta 2.7 getRemotePeerAdvertisement() return null
        PeerID remoteId = PeerID.create(URI.create(remoteNodeInfo.getID()));
        EndpointAddress addr = new EndpointAddress(remoteId, null, null);
        Messenger mess = peerGroup.getEndpointService().getMessenger(addr);
        //TODO aggiungere resolve per vedere se arriva risp immediata
        if (mess != null) {
            mess.resolve();
        } else {
            LOG.log(Level.INFO, "Messenger to " + remoteId + " is null");
            return false;
        }
        LOG.log(Level.INFO, "Endpoint " + addr + " isClosed=" + mess.isClosed() + " state=" + mess.getState() + " " + toString(mess.getState()));
        boolean isBound = bidiPipe.isBound();
        if (!isBound) {
            LOG.log(Level.INFO, "BidiPipe to " + remoteId + " is not bound");
        }
        //XXX messo true per ritentare connessione nel caso in cui JXTA chiudi la connessione
        //dopo un certo timeout
        //a volte non funziona se si stacca il cavo
        boolean isReachable = peerGroup.getEndpointService().isReachable(remoteId, false);
        if (!isReachable) {
            LOG.log(Level.INFO, "Peer " + remoteId + " is not reachable");
        }
//        System.out.println("Bidipipe isBound= " + isBound);
//        System.out.println("isReachable= " + isReachable);
        return /*isBound && */ isReachable;
    }

    private void notifyBrokenConnection() {
        netModule.nodeFailure(remoteNodeInfo);
        InnerEvent event = new InnerEvent(EventType.CRASH_EVENT, new Object[]{this});
        notifyEvent(event);

    }

    private void notifyClose(boolean alreadyBroken) {
        netModule.connectionClosed(remoteNodeInfo);
        if (!alreadyBroken) {
            /*
             * If the connection is broken a CRASH event has already been notified
             * so there is no need to notify the closing.
             */
            InnerEvent event = new InnerEvent(EventType.CLOSE_EVENT, new Object[]{this});
            notifyEvent(event);
        }
    }

    private void notifyEvent(final EventIF event) {
        //XXX the notify of a new received event must be done by a different
        //thread to avoid JxtaBidiPipe deadlock by sending a message from a
        //pipeMsg call
        //For performance issues is better to use a single thread

//        Runnable notifier = new Runnable() {
//
//            public void run() {
//                if (eventNotifier != null) {
//                    eventNotifier.notifyEvent(event);
//                }
//            }
//        };
//
//        new Thread(notifier, "Notifier Thread " + (count++)).start();
        LOG.log(Level.CONFIG, "notifier add start");
        notifier.add(event);
        LOG.log(Level.CONFIG, "notifier add end");
    }

    private void waitPipeOpenEvent() throws IOException {
        synchronized (openLock) {
            long remainingTime = CONNECTION_OPENING_TIMEOUT;
            while (!pipeOpenEventReceived) {
                try {
                    long start = System.currentTimeMillis();
                    openLock.wait(remainingTime);
                    remainingTime -= (System.currentTimeMillis() - start);
                    if (pipeOpenEventReceived) {
                        return;
                    } else if (remainingTime <= 0) {
                        throw new IOException("Pipe open event not received in time");
                    }
                } catch (InterruptedException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
        }
    }

    /**
     * Remote PeerID toString
     * @return
     */
    @Override
    public String toString() {
        return remoteNodeInfo.getID();
    }

    private static String toString(int msgState) {
        switch (msgState) {
            case Messenger.BREAKING:
                return "BREAKING";
            case Messenger.BROKEN:
                return "BROKEN";
            case Messenger.CLOSED:
                return "CLOSED";
            case Messenger.CLOSING:
                return "CLOSING";
            case Messenger.CONNECTED:
                return "CONNECTED";
            case Messenger.DISCONNECTED:
                return "DISCONNECTED";
            case Messenger.DISCONNECTING:
                return "DISCONNECTING";
//            case Messenger.IDLE:
//                return "IDLE";
            case Messenger.RECONCLOSING:
                return "RECONCLOSING";
            case Messenger.RECONNECTING:
                return "RECONNECTING";
            case Messenger.RECONSATURATED:
                return "RECONSATURATED";
            case Messenger.RESOLCLOSING:
                return "RESOLCLOSING";
            case Messenger.RESOLPENDING:
                return "RESOLPENDING";
            case Messenger.RESOLSATURATED:
                return "RESOLSATURATED";
//            case Messenger.RESOLVED:
//                return "RESOLVED";
            case Messenger.RESOLVING:
                return "RESOLVING";
//            case Messenger.SATURATED:
//                return "SATURATED";
            case Messenger.SENDING:
                return "SENDING";
            case Messenger.SENDINGSATURATED:
                return "SENDINGSATURATED";
//            case Messenger.TERMINAL:
//                return "TERMINAL";
            case Messenger.UNRESOLVABLE:
                return "UNRESOLVABLE";
            case Messenger.UNRESOLVED:
                return "UNRESOLVED";
            case Messenger.UNRESOLVING:
                return "UNRESOLVING";
//            case Messenger.USABLE:
//                return "USABLE";
            default:
                throw new AssertionError();
        }
    }

    @Override
    protected void finalize() throws Throwable {
        if (!disconnected) {
            LOG.log(Level.SEVERE, "FINALIZE A JXTA CONNECTION WHILE NOT DISCONNECTED");
        }
        super.finalize();
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Inner classes
    ///////////////////////////////////////////////////////////////////////////
    /**
     * An asynchronous event notifier.
     * Every events generated by this connection (like received message or close
     * and broken events) are added in a buffer and notified
     * to the {@link #eventNotifier} asynchronously by another thread,
     * in the same order they were added.
     *
     */
    private class Notifier extends Thread {

        LinkedList<EventIF> toNotify = new LinkedList<EventIF>();
        Lock lock = new ReentrantLock(true);
        Condition pieno = lock.newCondition();
        Condition vuoto = lock.newCondition();
        boolean terminated = false;

        public void add(EventIF event) {
            lock.lock();
            try {
                if (terminated) {
                    LOG.log(Level.SEVERE, "Connection towards {0} is disconnected.No message will be notified", new Object[]{remoteNodeInfo.getID()});
                }
                toNotify.add(event);
                LOG.log(Level.CONFIG, "To notify:\n" + LoggingUtil.elementByLine(toNotify));
                pieno.signalAll();
            } finally {
                lock.unlock();
            }

        }

        public void broken() {
            lock.lock();
            try {
                while (toNotify.size() > 0) {
                    try {
                        LOG.log(Level.INFO, "Wait for message delivering: " + toNotify.size());
                        vuoto.await();

                    } catch (InterruptedException ex) {
                        LOG.log(Level.SEVERE, null, ex);
                    }
                }
                LOG.log(Level.INFO, "All messages delivered.Notify broken connection");
                broken = true;
                disconnected = true;

            } finally {
                lock.unlock();
            }
            /*
             * Nel lock potrebbe causare deadlock (?)
             */
            notifyBrokenConnection();
            try {
                close();
            } catch (IOException ex) {
                LOG.log(Level.SEVERE, null, ex);
            }
        }

        @Override
        public void run() {
            setName("Notifier " + remoteNodeInfo.getID());
            while (!disconnected) {
                lock.lock();
                try {
                    while (toNotify.isEmpty()) {

                        try {
                            LOG.log(Level.CONFIG, "Notifier waiting");
                            pieno.await();
                            LOG.log(Level.CONFIG, "Notifier notified");
                        } catch (InterruptedException ex) {
                            ex.printStackTrace();
                            LOG.log(Level.SEVERE, null, ex);
                        }
                    }

                    //FIXME notification must be done out from lock
                    //by coping 'toNotify'
                    LOG.log(Level.CONFIG, "Notifier starts");
                    while (!toNotify.isEmpty()) {
                        EventIF event = toNotify.removeFirst();
                        LOG.log(Level.CONFIG, "Notifier notify " + event);
                        eventNotifier.notifyEvent(event);
                        LOG.log(Level.CONFIG, "Notifier notified " + event);
                    }
                    LOG.log(Level.CONFIG, "Notify empty state");
                    vuoto.signalAll();
                    terminated = disconnected;
                } catch (Throwable t) {
                    t.printStackTrace();
                    LOG.log(Level.INFO, "Throwable " + t.getLocalizedMessage(), t);
                } finally {
                    lock.unlock();
                }

            }
        }
    }

    private class ConnectionHandler extends Thread {

        /*
         * TODO optimization: the connection handler must assure that
         * in every period almost a message must me sent and received.
         * If the user sends o receives some application message in the period
         * there is no need to send echo message.
         * It could be done by using two timer
         */
        /**
         * Maximum response delay
         * TODO too expensive. At most 3sec
         */
        private static final long TOLERANCE_TIME = 30 * 1000;
        private Lock lock = new ReentrantLock(true);
        /**
         * An object used as lock to wait for echo request
         */
        private Condition echoRequestCond = lock.newCondition();
        /**
         * An object used as lock to wait for echo reply
         */
        private Condition echoReplyCond = lock.newCondition();
        /**
         * Indicates if the waited message is received
         */
        boolean received;

        @Override
        public void run() {

            while (!disconnected) {
                try {
                    if (startedLocally) {
                        try {
                            //FIXME Thread.sleep() may be inaccurate or
                            //may not starts the thread immediately
                            Thread.sleep(CHECK_TIMEOUT);
                        } catch (InterruptedException ex) {
                            LOG.log(Level.SEVERE, null, ex);
                        }
                        LOG.log(Level.FINE, "Sending echo request");
                        sendEchoRequest();
                        LOG.log(Level.FINE, "Waiting echo replay");
                        boolean rec = waitEchoReply(TOLERANCE_TIME);
                        if (!rec) {
                            LOG.log(Level.INFO, "Echo reply not received in time");
                            notifier.broken();
                            return;
                        }

                    } else {
                        LOG.log(Level.FINE, "Waiting echo request");
                        boolean rec = waitEchoRequest(CHECK_TIMEOUT + TOLERANCE_TIME);
                        if (!rec) {
                            LOG.log(Level.INFO, "Echo request not received in time");
                            notifier.broken();
                            return;
                        }
                        LOG.log(Level.FINE, "Sending echo reply");
                        sendEchoReply();

                    }
                } catch (IOException ex) {
                    //XXX va bene notificare fallimento in seguito ad ioex ??
                    LOG.log(Level.SEVERE, ex.getLocalizedMessage(), ex);
                    notifier.broken();
                    return;
                }
            }
        }

        private void sendEchoRequest() throws IOException {
            send(EchoMessage.ECHO_REQUEST);
        }

        private void sendEchoReply() throws IOException {
            send(EchoMessage.ECHO_REPLY);
        }

        private boolean waitEchoReply(long waitTime) {
            long timeRemaining = waitTime * 1000000;//must be nanos
            lock.lock();
            try {
                while (!received) {
                    if (timeRemaining > 0) {
                        try {
                            timeRemaining = echoReplyCond.awaitNanos(timeRemaining);
                        } catch (InterruptedException ex) {
                            LOG.log(Level.SEVERE, null, ex);
                        }
                    } else {
                        return false;
                    }
                }

                received = false;
                return true;
            } finally {
                lock.unlock();
            }
        }

        private boolean waitEchoRequest(long waitTime) {
            long timeRemaining = waitTime * 1000000;//must be nanos
            lock.lock();
            try {
                while (!received) {
                    if (timeRemaining > 0) {
                        try {
                            timeRemaining = echoRequestCond.awaitNanos(timeRemaining);
                        } catch (InterruptedException ex) {
                            LOG.log(Level.SEVERE, null, ex);
                        }
                    } else {
                        return false;
                    }
                }
                received = false;
                return true;
            } finally {
                lock.unlock();
            }
        }

        private void handle(EchoMessage echoMessage) {
            lock.lock();
            try {
                switch (echoMessage.type) {
                    case EchoMessage.REQUEST:
                        received = true;
                        echoRequestCond.signalAll();
                        break;
                    case EchoMessage.REPLY:
                        received = true;
                        echoReplyCond.signalAll();
                        break;
                    default:
                        throw new AssertionError();
                }
            } finally {
                lock.unlock();
            }

        }
    }

    /**
     * An object that represents an echo message
     *
     * Implements {@link Externalizable} for more efficient serialization
     */
    private static class EchoMessage implements Externalizable {

        static final EchoMessage ECHO_REQUEST = new EchoMessage(EchoMessage.REQUEST);
        static final EchoMessage ECHO_REPLY = new EchoMessage(EchoMessage.REPLY);
        static final int REQUEST = 0;
        static final int REPLY = 1;
        int type;

        public EchoMessage() {
        }

        public EchoMessage(int type) {
            this.type = type;
        }

        public int getType() {
            return type;
        }

        public void setType(int type) {
            this.type = type;
        }

        public void writeExternal(ObjectOutput out) throws IOException {
            out.writeInt(type);
        }

        public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
            type = in.readInt();
        }

        @Override
        public String toString() {
            return type == REQUEST ? "ECHO REQUEST" : "ECHO REPLY";
        }
    }
}
