/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.impl.networkModule.jxta;

import java.io.IOException;
import java.io.Serializable;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.jxta.discovery.DiscoveryEvent;
import net.jxta.discovery.DiscoveryListener;
import net.jxta.discovery.DiscoveryService;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.peer.PeerID;
import net.jxta.peergroup.PeerGroup;
import net.jxta.platform.NetworkManager;
import net.jxta.platform.NetworkManager.ConfigMode;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.rendezvous.RendezvousEvent;
import net.jxta.rendezvous.RendezvousListener;
import p2pMapReduce.ftp.FTPManager;
import p2pMapReduce.ftp.FTPManagerFactory;
import p2pMapReduce.networkModule.ConnectionIF;
import p2pMapReduce.networkModule.NetworkConfiguration;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.nodeModule.Constants;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.nodeModule.event.EventListener;
import p2pMapReduce.nodeModule.event.EventNotifier;
import p2pMapReduce.nodeModule.NodeModuleIF;
import p2pMapReduce.util.P2PMapReduceConfig;

/**
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
//TODO perchè non implementare eventlistener in modo da raccogliere tutti gli eventi
// di basso livello e se necessario ri-notificarli ad un altro eventListener ???
public class JxtaNetworkModule implements NetworkModuleIF, RendezvousListener {

    static {
        /* FROM JXTA documentation:
         * The following step is required and only need to be done once,
         * without this step the AdvertisementFactory has no means of
         * associating an advertisement name space with the proper obect
         */
        AdvertisementFactory.registerAdvertisementInstance(NodeAdvertisement.getAdvertisementType(),
                new NodeAdvertisement.Instantiator());
    }
    private static final transient Logger LOG = Logger.getLogger(JxtaNetworkModule.class.getName());
    /**
     * Pre-set P2PMapReduce group ID
     */
    public static final String GROUP_ID = "urn:jxta:cbid-3777FC6EFA9645CAB3F7D0119086FC5B59616261646162614E5047205032503302";
    /**
     * P2PMapReduce group name
     */
    public static final String P2P_MAPREDUCE_GROUP_NAME = "P2PMapReduce Group";
    /**
     * Default peer name
     */
    public static final String DEFAULT_PEER_NAME = "P2PMapReduce Peer";
    /**
     * Key for JXTA home directory attribute in {@link NetworkConfiguration}
     */
    public static final String JXTA_HOME_ATTR = "jxta_home";
    /**
     * Key for JXTA peer mode attribute in {@link NetworkConfiguration}
     */
    public static final String JXTA_PEER_MODE_ATTR = "jxta_peer_mode";
    /**
     * Key for {@link #GROUP_FIND_WAITING_TIME}
     */
    public static final String GROUP_FIND_WAITING_TIME_ATTR = "group.find.waiting.time";
    /**
     * Time to wait for group discovery
     */
    public static final long GROUP_FIND_WAITING_TIME;
    /**
     * Key for {@link #ADV_LIFE_TIME}
     */
    public static final String ADV_LIFETIME_ATTR = "adv.lifetime";
    /**
     * Limited adv life time in local cache to automatically remove older
     * version.
     * Note:The {@link #PUBLISH_TIMEOUT } must be less than this value to allow
     * others node to discovery and update ADV before they expired
     */
    public static final long ADV_LIFE_TIME;
    /**
     * Key for {@link #ADV_LIFE_TIME}
     */
    public static final String ADV_LIFE_TIME_ATTR = "adv.life.time";
    /**
     * Key for {@link #ADV_EXPIRATION_TIME}
     */
    public static final String ADV_EXPIRATION_TIME_ATTR = "adv.expiration.time";
    /**
     * Limited adv expiration time to automatically remove older version on
     * remote node cache
     *
     * Note:The {@link #PUBLISH_TIMEOUT } must be less than this value to allow
     * others node to discovery and update ADV before they expired
     */
    public static final long ADV_EXPIRATION_TIME;
    /**
     * Key for {@link #PUBLISH_TIMEOUT}
     */
    public static final String PERIODIC_PUBLISH_TIMEOUT_ATTR = "peridoc.publish.timeout";
    /**
     * The time period to publish a new ADV for this node
     */
    public static final long PERIODIC_PUBLISH_TIMEOUT;
    /**
     * Key for {@link #PERIODIC_FIND_TIMEOUT}
     * @deprecated 
     */
    public static final String PERIODIC_FIND_TIMEOUT_ATTR = "peridoc.find.timeout";
    /**
     * Time period to discover ADV on the net
     * @deprecated 
     */
    public static final long PERIODIC_FIND_TIMEOUT;
    /**
     * Key for {@link #DIRECT_FIND_WAITING_TIME}
     */
    public static final String DIRECT_FIND_WAITING_TIME_ATTR = "direct.find.waiting.time";
    /**
     * Time to wait before terminate a direct ADV discovery
     */
    public static final long DIRECT_FIND_WAITING_TIME;
    /**
     * Key for {@link #RDV_WAITING_TIME}
     */
    public static final String RDV_WAITING_TIME_ATTR = "rdv.waiting.time";
    /**
     * Time to wait for RDV connection
     */
    public static final long RDV_WAITING_TIME;
    ///////////////////////////////////////////////////////////////////////////
    //  Single peer configuration attributes
    ///////////////////////////////////////////////////////////////////////////
    /**
     * Key for {@code is.gateway} attribute of the {@link NetworkConfiguration}
     */
    public static final String IS_GATEWAY_ATTR = "is.gateway";
    /**
     * Key for {@code rdv.seeds} attribute of the {@link NetworkConfiguration}
     */
    public static final String RDV_SEEDS_ATTR = "rdv.seeds";
    /**
     * Key for {@code rdv.autoStart} attribute of the {@link NetworkConfiguration}
     */
    public static final String RDV_AUTO_START ="rdv.autoStart";

    static {

        Properties p = P2PMapReduceConfig.getConfiguration();
        GROUP_FIND_WAITING_TIME = 1000 * Integer.parseInt(p.getProperty(GROUP_FIND_WAITING_TIME_ATTR, "5"));
        ADV_LIFE_TIME = 1000 * Integer.parseInt(p.getProperty(ADV_LIFE_TIME_ATTR, "120"));
        //MOD era 120
        ADV_EXPIRATION_TIME = 1000 * Integer.parseInt(p.getProperty(ADV_EXPIRATION_TIME_ATTR, "120"));
        PERIODIC_PUBLISH_TIMEOUT = 1000 * Integer.parseInt(p.getProperty(PERIODIC_PUBLISH_TIMEOUT_ATTR, "60"));
        PERIODIC_FIND_TIMEOUT = 1000 * Integer.parseInt(p.getProperty(PERIODIC_FIND_TIMEOUT_ATTR, "60"));
        DIRECT_FIND_WAITING_TIME = 1000 * Integer.parseInt(p.getProperty(DIRECT_FIND_WAITING_TIME_ATTR, "3"));
        RDV_WAITING_TIME = 1000 * Integer.parseInt(p.getProperty(RDV_WAITING_TIME_ATTR, "75"));

    }
    private NetworkManager jxtaNetworkManager;
    private PeerGroup netPeerGroup;
    private PeerGroup p2pMapReduceGroup;
    /**
     * Passed network configuration for this node
     */
    private NetworkConfiguration networkConfiguration;
    /**
     * Must be considered like a final variable
     */
    private JxtaNodeInfo nodeInfo;
    /**
     * Keep track of the current version of the last published NodeAdvertisement
     * of this node
     */
    private long currentNodeADVVersion = NodeAdvertisementUtil.INITIAL_VERSION_VALUE;
    /**
     * The key is the node id of the remote connection endpoint
     */
    private HashMap<String, JxtaBidiPipeConnection> connectionCache;
    private JxtaBidiPipeServerThread serverThread;
    private Thread nodeInfoControllerThread;
    /**
     * @deprecated not used. will be removed
     */
    private NodeInfoController nodeInfoController;
    private NodeModuleIF nodeModule;
    private EventNotifier eventNotifier;
    private FTPManager ftpManager;
    private Lock lock;

    public JxtaNetworkModule(NetworkConfiguration networkConfiguration) {
        this.networkConfiguration = networkConfiguration;
        this.connectionCache = new HashMap<String, JxtaBidiPipeConnection>();
        this.eventNotifier = new EventNotifier();
        this.lock = new ReentrantLock(true);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  NetworkModuleIF methods
    ///////////////////////////////////////////////////////////////////////////
    public NodeInfo joinNetwork() {
        LOG.log(Level.INFO, "Starting joining network");

        ConfigMode cm = null;

        boolean isGateway = (Boolean) networkConfiguration.get(IS_GATEWAY_ATTR);
        if (isGateway) {
            cm = ConfigMode.RENDEZVOUS_RELAY;
            //overwrite default
            networkConfiguration.set(JXTA_PEER_MODE_ATTR, cm);
        } else {
            cm = (ConfigMode) networkConfiguration.get(JXTA_PEER_MODE_ATTR);
        }

        String[] rdvSeeds = (String[]) networkConfiguration.get(RDV_SEEDS_ATTR);

        jxtaNetworkManager = JxtaUtil.startDefaultNetPeerGroup(DEFAULT_PEER_NAME, cm, rdvSeeds, networkConfiguration.get(JXTA_HOME_ATTR).toString(), true);
        netPeerGroup = jxtaNetworkManager.getNetPeerGroup();

//        waitForRendezvous(netPeerGroup);

        //Try to find the group
        p2pMapReduceGroup = JxtaUtil.findAndCreateGroup(netPeerGroup, P2P_MAPREDUCE_GROUP_NAME, GROUP_FIND_WAITING_TIME);

        if (p2pMapReduceGroup == null) {
            //Group not found. Create it
            LOG.log(Level.INFO, P2P_MAPREDUCE_GROUP_NAME + " not found. Try to create it.");
            p2pMapReduceGroup = JxtaUtil.createNewGroup(netPeerGroup, GROUP_ID, P2P_MAPREDUCE_GROUP_NAME, P2P_MAPREDUCE_GROUP_NAME);
            LOG.log(Level.INFO, "{0} created: {1}", new Object[]{P2P_MAPREDUCE_GROUP_NAME, p2pMapReduceGroup});
            // no group founded so become RDV
//            startRendezVous();
        } else {
            LOG.log(Level.INFO, P2P_MAPREDUCE_GROUP_NAME + " founded.");
        }

        LOG.log(Level.INFO, "P2PMapReduceGroup ADV:\n{0}", p2pMapReduceGroup.getPeerGroupAdvertisement());

        //Joining  group
        LOG.log(Level.INFO, "Try to join " + P2P_MAPREDUCE_GROUP_NAME);
        boolean joined = JxtaUtil.join(p2pMapReduceGroup);
        if (joined) {
            LOG.log(Level.INFO, "Successfully joined!");
        } else {
            LOG.log(Level.INFO, "Join failed");
            return null;
        }


        boolean userNode = (Boolean) networkConfiguration.get(NetworkConfiguration.USER_NODE);

        if (!userNode && !isGateway) {
            boolean rdvAutoStart = (Boolean)networkConfiguration.get(RDV_AUTO_START);
            if (rdvAutoStart) {
                int period = (100 + new Random().nextInt(120)) * 1000;
                p2pMapReduceGroup.getRendezVousService().setAutoStart(true, period);
                waitForRendezvous(p2pMapReduceGroup);
            }
        }

        NodeAdvertisement nodeADV = NodeAdvertisementUtil.newNodeAdvertisement(p2pMapReduceGroup);

        if (userNode) {
            nodeADV.setRole(Constants.USER);
        }

        nodeInfo = new JxtaNodeInfo(nodeADV);

        LOG.log(Level.INFO, "Starting ftp server...");
        ftpManager = FTPManagerFactory.createFTPManager();
        try {
            ftpManager.startLocalServer();
            nodeADV.setIpAddress(ftpManager.getIpAddress());
            nodeADV.setFtpPort(ftpManager.getFtpPort());
            LOG.log(Level.INFO, "Ftp server started");
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, "Error during ftp server starting:" + ex.getLocalizedMessage(), ex);
            return null;
        }

        //Lsiterner to remove all node adv from other peers from local cache
//        addRemoveListener();

        if (userNode) {
            startRemotePublish();
            startListeningServer(NodeAdvertisementUtil.createPipeAdvertisement(nodeADV));

        } else {
            //XXX
            startRemotePublish();
            startListeningServer(NodeAdvertisementUtil.createPipeAdvertisement(nodeADV));
//            startCacheController();
        }

        return nodeInfo;
    }

    public FTPManager getFtpManager() {
        return ftpManager;
    }

    public void becomeSlave() {
//        stopRendezVous();
    }

    public void becomeMaster() {
//        startRendezVous();
    }

    public void publish() {
        currentNodeADVVersion++;
        publish(currentNodeADVVersion);
    }

    public NodeInfo getNodeInfo() {
        return nodeInfo;
    }

    public NodeModuleIF getNodeModule() {
        return nodeModule;
    }

    public void setNodeModule(NodeModuleIF nodeModule) {

        this.nodeModule = nodeModule;
    }

    public void addEventListener(EventListener eventListener) {
        eventNotifier.addEventListener(eventListener);
    }

    /**
     * Search the network to find the NodeAdvertisement
     * of the node identified by {@code nodeID}
     *
     * @param nodeID
     * @return The most recently founded NodeInfo of the specified node
     *          or null if no NodeAdvertisement have been founded
     */
    public NodeInfo find(String nodeID) {
        Set<Advertisement> advs = JxtaUtil.blockingFindAdv(p2pMapReduceGroup.getDiscoveryService(), DiscoveryService.ADV, NodeAdvertisement.PEER_ID_TAG, nodeID, DIRECT_FIND_WAITING_TIME);
        return getAvailableNodeFromLocalCache(nodeID);
    }

    /**
     * Indicate if a node is alive.
     * Check if a node info is available directly from the local cache
     *
     * //FIXME it has to return true if the node is really available or if
     * //the node adv is founded?
     *
     * @param nodeID
     * @return
     */
    public boolean isAlive(String nodeID) {
        return find(nodeID) != null;
    }

    public int getNumMasters() {
        return getAllMasters().size();
    }

    public int getNumSlaves() {
        return getAllSlaves().size();
    }

    public int getNumNodes() {
        return getAllNodes().size();
    }

    //TODO optimization: provide search methods that returns a map id->nodeinfo
    //(the local search already implements this function)
    
    //FIXME nei risultati potrebbe nn esserci questo nodo ?? per ora no
    //perchè si guarda dalla cache locale

    public Set<NodeInfo> getAllSlaves() {
        Set<Advertisement> advs = JxtaUtil.blockingFindAdv(p2pMapReduceGroup.getDiscoveryService(), DiscoveryService.ADV, NodeAdvertisement.ROLE_TAG, "" + Constants.SLAVE, DIRECT_FIND_WAITING_TIME);
//        return getAllAvailableNodes(advs);
        return getAllAvailableNodesFromLocalCache(Constants.SLAVE);
    }

    public Set<NodeInfo> getAllMasters() {
        Set<Advertisement> advs = JxtaUtil.blockingFindAdv(p2pMapReduceGroup.getDiscoveryService(), DiscoveryService.ADV, NodeAdvertisement.ROLE_TAG, "" + Constants.MASTER, DIRECT_FIND_WAITING_TIME);
//        return getAllAvailableNodes(advs);
        return getAllAvailableNodesFromLocalCache(Constants.MASTER);
    }

    /**
     * Retrieves all the node adv on the network and for each peer keeps only
     * the node adv with higher version
     *
     * //XXX for now we see in the local cache becouse if this is a rdv 
     * @return
     */
    public Set<NodeInfo> getAllNodes() {
//        return getAllSlaves();
        Set<Advertisement> advs = JxtaUtil.blockingFindAdv(
                p2pMapReduceGroup.getDiscoveryService(),
                DiscoveryService.ADV,
                NodeAdvertisement.TYPE_TAG,
                NodeAdvertisement.getAdvertisementType(),
                DIRECT_FIND_WAITING_TIME);

//        return getAllAvailableNodes(advs);
        return getAllAvailableNodesFromLocalCache();
    }

    /*
     * FIXME the implementation of getOtherSlaves and getOtherMaster methods
     * is strictly implementation-dependent for some reasons:
     * - if the search is restricted to those advertisement of slave type
     *   i can receive an obsolete node adv where the role is slave but now
     *   the node has become a master.
     *   This can cause a lot of problem specially for job manager that in
     *   worst case could choose as backup master some nodes that have become
     *   slaves, leaving the job to have effectively no backup
     *   Two strategy are possible: get all adv or introduce a check mechanism
     *   in all relations between nodess
     *
     */
    @Override
    public Set<NodeInfo> getOtherSlaves() {
        Set<NodeInfo> ris = getAllSlaves();

        for (Iterator<NodeInfo> it = ris.iterator(); it.hasNext();) {
            NodeInfo nodeInf = it.next();
            if (nodeInf.getID().equals(nodeInfo.getID())) {
                it.remove();
            }
        }

        return ris;
    }

    public Set<NodeInfo> getOtherMasters() {
        Set<NodeInfo> ris = getAllMasters();

        for (Iterator<NodeInfo> it = ris.iterator(); it.hasNext();) {
            NodeInfo nodeInf = it.next();
            if (nodeInf.getID().equals(nodeInfo.getID())) {
                it.remove();
            }
        }

        return ris;
    }

    ///////////////////////////////////////////////////////////////////////////
    //  
    ///////////////////////////////////////////////////////////////////////////
    public NetworkManager getNetworkManager() {
        return jxtaNetworkManager;
    }

    public PeerGroup getNetPeerGroup() {
        return netPeerGroup;
    }

    public PeerGroup getP2pMapReduceGroup() {
        return p2pMapReduceGroup;
    }

    public Collection<JxtaBidiPipeConnection> getActiveConnection(){
        return connectionCache.values();
    }

    private JxtaBidiPipeConnection createConnection(NodeInfo connectionEndPoint) throws IOException {
        PipeAdvertisement pipeADV = NodeAdvertisementUtil.createPipeAdvertisement(connectionEndPoint);
        JxtaBidiPipeConnection jxtaConn = new JxtaBidiPipeConnection(p2pMapReduceGroup, nodeInfo, connectionEndPoint, pipeADV, this, eventNotifier);
        connectionCache.put(connectionEndPoint.getID(), jxtaConn);
        return jxtaConn;
    }
    ///////////////////////////////////////////////////////////////////////////
    //  Non public methods
    ///////////////////////////////////////////////////////////////////////////

    private void flush(JxtaNodeInfo failedNodeInfo) {
        try {
            String peerID = failedNodeInfo.getID();
            //Flush directly the last version
            LOG.log(Level.INFO, "Flushing " + failedNodeInfo.getNodeADV());
            p2pMapReduceGroup.getDiscoveryService().flushAdvertisement(failedNodeInfo.getNodeADV());
            //Search and flush old versions
            LinkedList<Advertisement> list = JxtaUtil.getLocalADV(p2pMapReduceGroup.getDiscoveryService(), DiscoveryService.ADV, NodeAdvertisement.PEER_ID_TAG, peerID);
            for (Advertisement advertisement : list) {
                NodeAdvertisement nodeADV = (NodeAdvertisement) advertisement;
                LOG.log(Level.INFO, "Flushing " + nodeADV);
                p2pMapReduceGroup.getDiscoveryService().flushAdvertisement(nodeADV);
            }

        } catch (IOException ex) {
            Logger.getLogger(JxtaNetworkModule.class.getName()).log(Level.SEVERE, null, ex);
        }
        if (nodeInfoController != null) {
            nodeInfoController.flush(failedNodeInfo);
        }
    }

    private void addRemoveListener() {
        p2pMapReduceGroup.getDiscoveryService().addDiscoveryListener(new DiscoveryListener() {

            public void discoveryEvent(DiscoveryEvent event) {
                Enumeration<Advertisement> ris = event.getResponse().getAdvertisements();
                while (ris.hasMoreElements()) {
                    try {
                        Advertisement nextElement = ris.nextElement();
                        if (!(nextElement instanceof NodeAdvertisement)) {
                            return;
                        }
                        p2pMapReduceGroup.getDiscoveryService().flushAdvertisement(nextElement);
                    } catch (IOException ex) {
                        LOG.log(Level.SEVERE, null, ex);
                    }
                }
            }
        });
    }

    private void waitForRendezvous(PeerGroup pg) {
        //TODO comment
        pg.getRendezVousService().addListener(this);
        boolean connected = pg.getRendezVousService().isConnectedToRendezVous();
        boolean isRDV = pg.getRendezVousService().isRendezVous();
        if (!connected && !isRDV) {
            LOG.log(Level.INFO, "Waiting for RDV connection for " + pg.getPeerGroupName());
//            int period = (30 + new Random().nextInt(30)) * 1000;
//            pg.getRendezVousService().setAutoStart(true, period);
//            p2pMapReduceGroup.getRendezVousService().startRendezVous();
            waitForRendezvousConnection(pg, 1 * 60 * 1000);
            connected = pg.getRendezVousService().isConnectedToRendezVous();
            LOG.log(Level.INFO, "Connected to RDV? " + connected);
        }
    }

    private void startRendezVous() {
        LOG.log(Level.INFO, "Starting RDV");
        p2pMapReduceGroup.getRendezVousService().startRendezVous();
        LOG.log(Level.INFO, "RDV started? " + p2pMapReduceGroup.getRendezVousService().isRendezVous());
    }

    private void stopRendezVous() {
        LOG.log(Level.INFO, "Stopping RDV");
        p2pMapReduceGroup.getRendezVousService().stopRendezVous();
        LOG.log(Level.INFO, "RDV started? " + p2pMapReduceGroup.getRendezVousService().isRendezVous());
    }

    /**
     * @deprecated use {@link #getAllAvailableNodesFromLocalCache() } instead
     * @return
     */
    protected Set<NodeInfo> getAllNodesFromLocalCache() {

        HashMap<String, JxtaNodeInfo> nodeInfoMap = new HashMap<String, JxtaNodeInfo>();

        //TODO provare direttamente con role = slave
        LinkedList<Advertisement> list = JxtaUtil.getLocalADV(p2pMapReduceGroup.getDiscoveryService(), DiscoveryService.ADV, NodeAdvertisement.TYPE_TAG, NodeAdvertisement.getAdvertisementType());

        for (Advertisement adv : list) {
            //tra quelli riferiti allo stesso peer seleziona quello a versione maggiore
            try {
                NodeAdvertisement nodeADV = (NodeAdvertisement) adv;
                JxtaNodeInfo previousNodeInfo = nodeInfoMap.get(nodeADV.getPeerID());
                if (previousNodeInfo == null) {
                    nodeInfoMap.put(nodeADV.getPeerID(), new JxtaNodeInfo(nodeADV));
                } else if (nodeADV.getVersion() > previousNodeInfo.getVersion()) {
                    previousNodeInfo.setNodeADV(nodeADV);
                }

            } catch (ClassCastException cce) {
                System.out.println("CLASS CAST EXCEPTION " + adv);
                cce.printStackTrace();
            }

        }



        return new HashSet<NodeInfo>(nodeInfoMap.values());
    }

    /**
     * TODO public for testing purpose. to make not public
     * @return
     */
    public Set<NodeInfo> getAllAvailableNodesFromLocalCache() {

        HashMap<String, JxtaNodeInfo> nodeInfoMap = new HashMap<String, JxtaNodeInfo>();
        Set<String> toRemove = new HashSet<String>();

        LinkedList<Advertisement> list = JxtaUtil.getLocalADV(p2pMapReduceGroup.getDiscoveryService(), DiscoveryService.ADV, NodeAdvertisement.TYPE_TAG, NodeAdvertisement.getAdvertisementType());

        for (Advertisement adv : list) {
            /*
             * Among all those NodeAdv referred to some node choose the most
             * recent one
             */
            try {
                NodeAdvertisement nodeADV = (NodeAdvertisement) adv;
                JxtaNodeInfo previousNodeInfo = nodeInfoMap.get(nodeADV.getPeerID());
                if (previousNodeInfo == null) {
                    nodeInfoMap.put(nodeADV.getPeerID(), new JxtaNodeInfo(nodeADV));
                } else if (nodeADV.getVersion() > previousNodeInfo.getVersion()) {
                    previousNodeInfo.setNodeADV(nodeADV);
                }
                if (!nodeADV.isAvailable()) {
                    toRemove.add(nodeADV.getPeerID());
                }
            } catch (ClassCastException cce) {
                System.out.println("CLASS CAST EXCEPTION " + adv);
                cce.printStackTrace();
            }

        }

        for (String nodeID : toRemove) {
            nodeInfoMap.remove(nodeID);
        }

//        Set<NodeInfo> ris = new TreeSet<NodeInfo>(new NodeComparator());
        Set<NodeInfo> ris = new HashSet<NodeInfo>();
        ris.addAll(nodeInfoMap.values());

        return ris;
    }

    /**
     * Returns the set of all available node info for those nodes which role
     * is {@code nodeType}.
     *
     * XXX In order to avoid the possibility of collect obsolete node info, a
     * preliminary search of all node info is done to collect the most recent
     * node info for a node.
     * After only the node info of the specified type are returned
     * @param nodeType
     * @return
     */
    public Set<NodeInfo> getAllAvailableNodesFromLocalCache(int nodeType) {
        Set<NodeInfo> allAvailableNodesFromLocalCache = getAllAvailableNodesFromLocalCache();
        for (Iterator<NodeInfo> it = allAvailableNodesFromLocalCache.iterator(); it.hasNext();) {
            NodeInfo ni = it.next();
            if (ni.getRole() != nodeType) {
                it.remove();
            }
        }
        return allAvailableNodesFromLocalCache;
    }

    protected Set<NodeInfo> getAllAvailableNodes(Set<Advertisement> resp) {

        HashMap<String, JxtaNodeInfo> nodeInfoMap = new HashMap<String, JxtaNodeInfo>();
        Set<String> toRemove = new HashSet<String>();

        for (Advertisement adv : resp) {
            /*
             * Among all those NodeAdv referred to some node choose the most
             * recent one
             */
            try {
                NodeAdvertisement nodeADV = (NodeAdvertisement) adv;
                JxtaNodeInfo previousNodeInfo = nodeInfoMap.get(nodeADV.getPeerID());
                if (previousNodeInfo == null) {
                    nodeInfoMap.put(nodeADV.getPeerID(), new JxtaNodeInfo(nodeADV));
                    if (!nodeADV.isAvailable()) {
                        toRemove.add(nodeADV.getPeerID());
                    }
                } else if (nodeADV.getVersion() > previousNodeInfo.getVersion()) {
                    previousNodeInfo.setNodeADV(nodeADV);
                    if (!nodeADV.isAvailable()) {
                        toRemove.add(nodeADV.getPeerID());
                    }
                }

            } catch (ClassCastException cce) {
                System.out.println("CLASS CAST EXCEPTION " + adv);
                cce.printStackTrace();
            }

        }

        for (String nodeID : toRemove) {
            nodeInfoMap.remove(nodeID);
        }

        Set<NodeInfo> ris = new HashSet<NodeInfo>();
        ris.addAll(nodeInfoMap.values());

        return ris;
    }

    /**
     * Try to find the most recent available version of the node info
     * for the node which id is {@code nodeID}.
     * 
     * @param nodeID
     * @return null if the node is not available or the most recent node info
     * for that node
     */
    public JxtaNodeInfo getAvailableNodeFromLocalCache(String nodeID) {

        LinkedList<Advertisement> list = JxtaUtil.getLocalADV(
                p2pMapReduceGroup.getDiscoveryService(),
                DiscoveryService.ADV,
                NodeAdvertisement.PEER_ID_TAG,
                nodeID);

        NodeAdvertisement mostRecent = null;
        for (Advertisement adv : list) {

            try {
                NodeAdvertisement nodeADV = (NodeAdvertisement) adv;
                if (mostRecent == null) {
                    mostRecent = nodeADV;
                } else if (mostRecent.getVersion() < nodeADV.getVersion()) {
                    mostRecent = nodeADV;
                }
            } catch (ClassCastException cce) {
                LOG.log(Level.SEVERE, "CLASS CAST EXCEPTION " + adv);
            }

        }

        if (mostRecent != null && mostRecent.isAvailable()) {
            return new JxtaNodeInfo(mostRecent);
        } else {
            return null;
        }
    }

    /**
     * In the case of user node only the pipe adv are been published
     */
    private void publish(long newVersion) {
        try {
            //TODO refactoring
            boolean userNode = (Boolean) networkConfiguration.get(NetworkConfiguration.USER_NODE);

            NodeAdvertisement nodeADV = new NodeAdvertisement(nodeInfo.getNodeADV());

            if (!userNode) {
                nodeADV.setID(NodeAdvertisementUtil.createNodeAdvertisementID(newVersion, p2pMapReduceGroup));
                nodeADV.setVersion(newVersion);
                nodeInfo.setNodeADV(nodeADV);
                LOG.log(Level.CONFIG, "Version {0} NodeADV ID is {1}", new Object[]{newVersion, nodeADV.getID()});
            }

            PipeAdvertisement pipeADV = NodeAdvertisementUtil.createPipeAdvertisement(nodeADV);

            if (!userNode) {
                p2pMapReduceGroup.getDiscoveryService().publish(nodeADV, ADV_LIFE_TIME, ADV_EXPIRATION_TIME);
            }
            p2pMapReduceGroup.getDiscoveryService().publish(pipeADV, ADV_LIFE_TIME, ADV_EXPIRATION_TIME);

//            List<PeerID> rendezvous = p2pMapReduceGroup.getRendezVousService().getLocalRendezVousView();
//            for (PeerID rdv : rendezvous) {
//
//                p2pMapReduceGroup.getDiscoveryService().remotePublish(rdv.toString(), nodeADV, ADV_EXPIRATION_TIME);
//                p2pMapReduceGroup.getDiscoveryService().remotePublish(rdv.toString(), pipeADV, ADV_EXPIRATION_TIME);
//            }

            if (!userNode) {
                p2pMapReduceGroup.getDiscoveryService().remotePublish(nodeADV, ADV_EXPIRATION_TIME);
            }

            p2pMapReduceGroup.getDiscoveryService().remotePublish(pipeADV, ADV_EXPIRATION_TIME);

            //XXX Flush all the old adv of this node from the local cache
            //for now limit the life time
//            LinkedList<Advertisement> local = JxtaUtil.getLocalADV(p2pMapReduceGroup.getDiscoveryService(), DiscoveryService.ADV, NodeAdvertisement.PEER_ID_TAG, nodeADV.getPeerID());
//            for (Advertisement advertisement : local) {
//                NodeAdvertisement oldVersion = (NodeAdvertisement) advertisement;
//                if (oldVersion.getVersion() < newVersion) {
//                    p2pMapReduceGroup.getDiscoveryService().flushAdvertisement(oldVersion);
//                }
//            }
        } catch (IOException ex) {
            ex.printStackTrace();
            LOG.log(Level.SEVERE, null, ex);
        }
    }

    public void send(NodeInfo endPoint, Serializable o) throws IOException {
        ConnectionIF conn = getConnection(endPoint);
        conn.send(o);
    }

    public void send(String senderID, Serializable msg) throws IOException {
        ConnectionIF conn = getConnection(senderID);
        conn.send(msg);
    }

    public ConnectionIF getConnection(NodeInfo connectionEndPoint) throws IOException {
        lock.lock();
        try {
            if (connectionEndPoint.equals(nodeInfo)) {
                LOG.log(Level.SEVERE, "Getting a new connection toward local node");
            }
            //First check in the cache
            JxtaBidiPipeConnection jxtaConn = connectionCache.get(connectionEndPoint.getID());

            if (jxtaConn == null) {
                LOG.log(Level.FINE, "Creating a connection toward " + connectionEndPoint.getID());
                jxtaConn = createConnection(connectionEndPoint);
            } else if (jxtaConn.isBroken()) {
                LOG.log(Level.SEVERE, "Found a broken connection toward " + connectionEndPoint.getID());
                connectionCache.remove(connectionEndPoint.getID());
                throw new IOException("Broken connection towards " + connectionEndPoint.getID());
            } else if (jxtaConn.isDisconnected()) {
                LOG.log(Level.SEVERE, "Found a disconnected connection toward " + connectionEndPoint.getID() + ". Try to open a new one");
                jxtaConn = createConnection(connectionEndPoint);
            }

            return jxtaConn;
        } finally {
            lock.unlock();
        }

    }

    public ConnectionIF getConnection(String nodeID) throws IOException {
        lock.lock();
        try {
            //First check in the cache
            JxtaBidiPipeConnection jxtaConn = connectionCache.get(nodeID);
            if (jxtaConn == null) {
                NodeInfo ni = find(nodeID);
                if (ni != null) {
                    return getConnection(ni);
                }

                throw new IOException("Unable to find the NodeInfo for node " + nodeID + " to create a connection.");

            } else {
                //Connection already exists
                return jxtaConn;
            }
        } finally {
            lock.unlock();
        }

    }

    protected void putConnection(String nodeID, JxtaBidiPipeConnection conn) {
        lock.lock();
        try {
            JxtaBidiPipeConnection old = connectionCache.remove(nodeID);
            if (old != null) {
                LOG.log(Level.WARNING, "Received a new connection from " + nodeID + " for which a connection is already available. Using the new one");
                /*
                 * Nel caso peggiore i due endpoint potrebbero aprire in parallelo la connessione
                 * e ricevere la notifica in ordine inverso, il che provocherebbe la chiusura di entrambe.
                 * Meglio lasciare la connessione aperta in quanto al piu saranno 2 e jxta le dovrebbe
                 * gestire come un unica connessione TCP
                 */
//                try {
//                    old.close();
//                } catch (IOException ex) {
//                    LOG.log(Level.SEVERE, "IOException during close: " + ex.getMessage(), ex);
//                }
                connectionCache.put(nodeID, conn);
            } else {
                connectionCache.put(nodeID, conn);
            }
        } finally {
            lock.unlock();
        }

    }

    protected JxtaBidiPipeConnection removeConnection(String nodeID) {
        lock.lock();
        try {
            return connectionCache.remove(nodeID);
        } finally {
            lock.unlock();
        }
    }

    //FIXME per ora notificato direttamente da JxtaConnection prima di notificare
    //node module. Meglio se chiamato da nodemodule dopo notifica ??? mi sa di no
    protected void nodeFailure(NodeInfo failedNodeInfo) {
        LOG.log(Level.INFO, "Handling node failure\n{0}", new Object[]{failedNodeInfo});
//        publishFailedADV(failedNodeInfo);
        removeConnection(failedNodeInfo.getID());
    }

    private void publishFailedADV(NodeInfo failedNodeInfo) {
        //        flush((JxtaNodeInfo) failedNodeInfo);
        JxtaNodeInfo currentNodeInfo = (JxtaNodeInfo) find(failedNodeInfo.getID());
        LOG.log(Level.INFO, "Current node info " + currentNodeInfo);
        JxtaNodeInfo previousNodeInfo = (JxtaNodeInfo) failedNodeInfo;
        NodeAdvertisement nodeADV = null;
        if (currentNodeInfo != null && currentNodeInfo.getVersion() >= previousNodeInfo.getVersion()) {
            nodeADV = currentNodeInfo.getNodeADV();
        } else {
            LOG.log(Level.WARNING, "Founded an obsolete version of nodeinfo for {0} current:\n{1} previous:\n{2}", new Object[]{failedNodeInfo.getID(), currentNodeInfo, previousNodeInfo});
            nodeADV = previousNodeInfo.getNodeADV();
        }
        if (nodeADV.getRole() != Constants.USER) {
            NodeAdvertisement failedNodeADV = NodeAdvertisementUtil.createNodeAdvertisement(nodeADV.getVersion() + NodeAdvertisementUtil.FAILED_VERSION_INCREMENT, p2pMapReduceGroup, nodeADV);
            failedNodeADV.setAvailable(false);
            try {
                LOG.log(Level.INFO, "Publishing failed node adv\n" + failedNodeADV);
                p2pMapReduceGroup.getDiscoveryService().publish(failedNodeADV, ADV_LIFE_TIME, ADV_EXPIRATION_TIME);
                //            p2pMapReduceGroup.getDiscoveryService().remotePublish(nodeADV, ADV_EXPIRATION_TIME);
            } catch (IOException ex) {
                LOG.log(Level.SEVERE, null, ex);
            }
            if (nodeInfoController != null) {
                nodeInfoController.flush(failedNodeInfo);
            }
        }
    }

    protected void connectionClosed(NodeInfo connectionEndPoint) {
        LOG.log(Level.INFO, "Handling connection closing towards " + connectionEndPoint.getID());
        removeConnection(connectionEndPoint.getID());
    }

    /**
     * Start a server thread that listen for incoming {@code JxtaBidiPipe}
     * 
     * @param pipeADV
     */
    private void startListeningServer(PipeAdvertisement pipeADV) {
        serverThread = new JxtaBidiPipeServerThread(this, eventNotifier, p2pMapReduceGroup, pipeADV);
        serverThread.start();
    }

    private void startRemotePublish() {
        new Thread() {

            @Override
            public void run() {

                while (true) {
                    try {
                        publish();
                        Thread.sleep(PERIODIC_PUBLISH_TIMEOUT);
                    } catch (InterruptedException ex) {
                        LOG.log(Level.SEVERE, null, ex);
                    }
                }
            }
        }.start();
    }

    /**
     * @deprecated 
     */
    private void startCacheController() {
        nodeInfoController = new NodeInfoController(this);
        nodeInfoControllerThread = new Thread(nodeInfoController);
        nodeInfoControllerThread.start();
    }

    public void rendezvousEvent(RendezvousEvent event) {
        String type = "";
        switch (event.getType()) {
            case RendezvousEvent.BECAMEEDGE:
                type = "BECAMEEDGE";
                break;
            case RendezvousEvent.BECAMERDV:
                type = "BECAMERDV";
                synchronized (this) {
                    notifyAll();
                }
                break;
            case RendezvousEvent.CLIENTCONNECT:
                type = "CLIENTCONNECT";
                break;
            case RendezvousEvent.CLIENTDISCONNECT:
                type = "CLIENTDISCONNECT";
                break;
            case RendezvousEvent.CLIENTFAILED:
                type = "CLIENTFAILED";
                break;
            case RendezvousEvent.CLIENTRECONNECT:
                type = "CLIENTRECONNECT";
                break;
            case RendezvousEvent.RDVCONNECT:
                type = "RDVCONNECT";
                synchronized (this) {
                    notifyAll();
                }
                break;
            case RendezvousEvent.RDVDISCONNECT:
                type = "RDVDISCONNECT";
                break;
            case RendezvousEvent.RDVFAILED:
                type = "RDVFAILED";
                break;
            case RendezvousEvent.RDVRECONNECT:
                type = "RDVRECONNECT";
                break;

            default:
                throw new AssertionError();
        }
        LOG.log(Level.INFO, "New RendezvousEvent " + type + " from " + event.getPeer());
    }

    private void waitForRendezvousConnection(PeerGroup pg, int timeout) {

        synchronized (this) {
            try {
                while (!connectedOrRDV(pg)) {
                    wait(timeout);
                    LOG.log(Level.INFO, "WAKE UP");
                }
            } catch (InterruptedException ex) {
                LOG.log(Level.SEVERE, null, ex);
            }
        }
    }

    public boolean connectedOrRDV(PeerGroup pg) {
        boolean connected = pg.getRendezVousService().isConnectedToRendezVous();
//        connected |= p2pMapReduceGroup.getRendezVousService().isRendezVous();
        connected |= !pg.getRendezVousService().getLocalRendezVousView().isEmpty();
        return connected;
    }

    /**
     * Send a remote discovery query and wait for 'timeout' ms
     * 
     * @param timeout
     */
    private void remoteBlockingDiscovery(long timeout) {
        String peerID = null;
        List<PeerID> redezvous = p2pMapReduceGroup.getRendezVousService().getLocalRendezVousView();
        if (!redezvous.isEmpty()) {
            for (PeerID rdvID : redezvous) {
                JxtaUtil.oneResponseFindAdv(p2pMapReduceGroup.getDiscoveryService(), rdvID.toString(), DiscoveryService.ADV, NodeAdvertisement.TYPE_TAG, NodeAdvertisement.getAdvertisementType(), timeout);
            }
            try {
                Thread.sleep(timeout);
            } catch (InterruptedException ex) {
                LOG.log(Level.SEVERE, null, ex);
            }
        } else {
            Set<Advertisement> adv = JxtaUtil.blockingFindAdv(p2pMapReduceGroup.getDiscoveryService(), peerID, DiscoveryService.ADV, NodeAdvertisement.TYPE_TAG, NodeAdvertisement.getAdvertisementType(), timeout);

        }
        //FIXME questa chiamata genera deadlock o per GUI o per NodeFSM
//        InnerEvent e = new InnerEvent(EventType.REMOTE_DISCOVERY, new Object[]{adv});
//        eventNotifier.notifyEvent(e);
    }
}
