/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.impl.networkModule.jxta;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.URI;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import net.jxta.credential.AuthenticationCredential;
import net.jxta.discovery.DiscoveryEvent;
import net.jxta.discovery.DiscoveryListener;
import net.jxta.discovery.DiscoveryService;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.MimeMediaType;
import net.jxta.document.StructuredDocument;
import net.jxta.document.XMLElement;
import net.jxta.endpoint.ByteArrayMessageElement;
import net.jxta.endpoint.Message;
import net.jxta.endpoint.MessageElement;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.ProtocolNotSupportedException;
import net.jxta.id.IDFactory;
import net.jxta.impl.content.ContentServiceImpl;
import net.jxta.impl.membership.none.NoneMembershipService;
import net.jxta.impl.peergroup.CompatibilityUtils;
import net.jxta.impl.peergroup.StdPeerGroup;
import net.jxta.impl.peergroup.StdPeerGroupParamAdv;
import net.jxta.impl.protocol.DiscoveryConfigAdv;
import net.jxta.membership.Authenticator;
import net.jxta.peergroup.PeerGroup;
import net.jxta.peergroup.PeerGroupID;
import net.jxta.platform.NetworkConfigurator;
import net.jxta.platform.NetworkManager;
import net.jxta.protocol.DiscoveryResponseMsg;
import net.jxta.protocol.ModuleImplAdvertisement;
import net.jxta.protocol.PeerAdvertisement;
import net.jxta.protocol.PeerGroupAdvertisement;
import p2pMapReduce.util.LoggingUtil;

/**
 * An utility class for JXTA
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class JxtaUtil {

    private static final transient Logger LOG = Logger.getLogger(JxtaUtil.class.getName());
    public static String DEFAULT_JXTA_HOME = "./jxta";
    private static final int DEFAULT_THRESHOLD = 1000;
    private final static MimeMediaType GZIP_MEDIA_TYPE = new MimeMediaType("application/gzip").intern();

    public static NetworkManager startDefaultNetPeerGroup(String peerName, NetworkManager.ConfigMode configMode, String[] rdvSeeds, String home, boolean clean) {
        return startDefaultNetPeerGroup(peerName, configMode, rdvSeeds, home, -1, clean);
    }

    public static NetworkManager startDefaultNetPeerGroup(String peerName, NetworkManager.ConfigMode configMode, String[] rdvSeeds, String home, int tcpPort, boolean clean) {

        PeerGroup ris = null;
        NetworkManager nManager = null;
        try {
            File homeF = new File(home);
            if (clean && homeF.exists()) {
                recursiveDelete(homeF);
            }

            nManager = new NetworkManager(configMode, peerName, homeF.toURI());
            NetworkConfigurator nc = nManager.getConfigurator();
            nc.setTcpEnabled(true);
            //MOD si abilita per testare seed
            nc.setHttpEnabled(true);
            nc.setUseMulticast(true);
            if (tcpPort != -1) {
                nc.setTcpPort(tcpPort);
            }

            if (rdvSeeds != null) {
                nc.setUseOnlyRendezvousSeeds(true);
                nc.setUseOnlyRelaySeeds(true);
                for (String seed : rdvSeeds) {
                    nc.addSeedRendezvous(URI.create(seed));
                    nc.addSeedRelay(URI.create(seed));
                }
            }


            ris = nManager.startNetwork();

        } catch (PeerGroupException ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        }

        return nManager;
    }

    ///////////////////////////////////////////////////////////////////////////
    //  
    ///////////////////////////////////////////////////////////////////////////
    /**
     * Find a peer group named as {@code peerGroupName }.
     * If a response is received in {@code timeout }ms creates and returns the
     * PeerGroup from the founded ADV
     *
     * @param netPeerGroup
     * @param peerGroupName
     * @param timeOut
     * @return
     */
    public static PeerGroup findAndCreateGroup(PeerGroup netPeerGroup, String peerGroupName, long timeOut) {
        PeerGroupAdvertisement pgADV = findGroup(netPeerGroup.getDiscoveryService(), peerGroupName, timeOut);
        if (pgADV != null) {
            return createGroup(netPeerGroup, pgADV);
        } else {
            return null;
        }
    }

    public static PeerGroupAdvertisement findGroup(DiscoveryService ds, long timeOut) {

        final LinkedList<PeerGroupAdvertisement> list = new LinkedList<PeerGroupAdvertisement>();

        final Object lock = new Object();

        ds.getRemoteAdvertisements(null, DiscoveryService.GROUP, null, null, 1, new DiscoveryListener() {

            public void discoveryEvent(DiscoveryEvent event) {
                if (event != null) {
                    LOG.log(Level.CONFIG, "{0}", new Object[]{getEventInfo(event)});
                    DiscoveryResponseMsg dRM = event.getResponse();
                    list.add((PeerGroupAdvertisement) dRM.getAdvertisements().nextElement());

                    synchronized (lock) {
                        lock.notifyAll();
                    }

                } else {
                    System.out.println("RECEIVED A NULL DISCOVERY EVENT");
                }
            }
        });

        synchronized (lock) {
            try {
                lock.wait(timeOut);
            } catch (InterruptedException ex) {
                Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);

            }
        }

        return list.getFirst();
    }

    /**
     * Try to find an advertisement for a group with name 'peerGroupName'.
     * If timeout>0 it waits for timeout ms and then return null if no DiscoveyEvents are received.
     * If timeout = 0 bloks until a DicoveryEvent is received.
     * @param ds
     * @param peerGroupName
     * @param timeOut
     * @return
     */
    public static PeerGroupAdvertisement findGroup(final DiscoveryService ds, String peerGroupName, long timeOut) {

        final LinkedList<PeerGroupAdvertisement> list = new LinkedList<PeerGroupAdvertisement>();

        final Object lock = new Object();

        //TODO threshold = 1 is good?
        ds.getRemoteAdvertisements(null, DiscoveryService.GROUP, "Name", peerGroupName, 1, new DiscoveryListener() {

            public void discoveryEvent(DiscoveryEvent event) {
                if (event != null) {
                    LOG.log(Level.INFO, "{0}", new Object[]{getEventInfo(event)});
//                    System.err.println(getEventInfo(event));
                    DiscoveryResponseMsg dRM = event.getResponse();
                    list.add((PeerGroupAdvertisement) dRM.getAdvertisements().nextElement());
                    ds.removeDiscoveryListener(this);
                    synchronized (lock) {

                        lock.notifyAll();

                    }

                }
            }
        });

        synchronized (lock) {
            try {
                lock.wait(timeOut);
            } catch (InterruptedException ex) {
                Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
            }
        }

        return list.size() > 0 ? list.getFirst() : null;
    }

    /**
     * Creates a new group by group name and description.
     * The typical use of this routine is creating a whole new group based
     * on a newly created and possibly unpublished implementation adv.
     * 
     * @param parentGroup
     * @param peerGroupName
     * @param peerGroupDesc
     * @return
     */
    public static PeerGroup createNewGroup(PeerGroup parentGroup, String peerGroupName, String peerGroupDesc) {

        PeerGroup ris = null;
        try {

            //FIX for jxta 2.5
//            ModuleImplAdvertisement miadv = parentGroup.getAllPurposePeerGroupImplAdvertisement();
//            ris = parentGroup.newGroup(null, miadv, peerGroupName, peerGroupDesc);
            //FIX for jxta 2.7
            PeerGroupID id = IDFactory.newPeerGroupID(parentGroup.getPeerGroupID());
            ModuleImplAdvertisement miADV = createAllPurposePeerGroupImplAdv(peerGroupDesc);
            PeerGroupAdvertisement pgADV = (PeerGroupAdvertisement) AdvertisementFactory.newAdvertisement(PeerGroupAdvertisement.getAdvertisementType());
            pgADV.setName(peerGroupName);
            pgADV.setDescription(peerGroupDesc);
            pgADV.setPeerGroupID(id);
            pgADV.setModuleSpecID(PeerGroup.allPurposePeerGroupSpecID);
            ris = parentGroup.newGroup(pgADV);
//            ris = parentGroup.newGroup(id, miADV, peerGroupName, peerGroupDesc, true);
            System.out.println("SHOW NEW GROUP ADV");
            showAllLocalADV(ris.getDiscoveryService(), DiscoveryService.GROUP);
            showAllLocalADV(ris.getDiscoveryService(), DiscoveryService.PEER);

        } catch (Exception ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        }

        return ris;

    }

    /**
     * Create a new group by specifing group name and description.
     * The typical use of this routine is creating a whole new group based
     * on a newly created and possibly unpublished implementation adv.
     * @param parentGroup
     * @param peerGroupName
     * @param peerGroupDesc
     * @return
     */
    public static PeerGroup createNewGroup(PeerGroup parentGroup, String peerGroupId, String peerGroupName, String peerGroupDesc) {

        PeerGroup ris = null;
        try {

            //FIX for jxta 2.5
//            ModuleImplAdvertisement miadv = parentGroup.getAllPurposePeerGroupImplAdvertisement();
//            ris = parentGroup.newGroup(null, miadv, peerGroupName, peerGroupDesc);
            //FIX for jxta 2.7
            PeerGroupID id = PeerGroupID.create(URI.create(peerGroupId));
//            ModuleImplAdvertisement miADV = createAllPurposePeerGroupImplAdv(peerGroupDesc);
            PeerGroupAdvertisement pgADV = (PeerGroupAdvertisement) AdvertisementFactory.newAdvertisement(PeerGroupAdvertisement.getAdvertisementType());
            pgADV.setName(peerGroupName);
            pgADV.setDescription(peerGroupDesc);
            pgADV.setPeerGroupID(id);
            pgADV.setModuleSpecID(PeerGroup.allPurposePeerGroupSpecID);

            //MOD added
            DiscoveryConfigAdv discoveryConfig = (DiscoveryConfigAdv) AdvertisementFactory.newAdvertisement(DiscoveryConfigAdv.getAdvertisementType());
            discoveryConfig.setForwardAlwaysReplica(true);
            discoveryConfig.setForwardBelowTreshold(true);
            discoveryConfig.setLocalOnly(false);
            StructuredDocument serviceParam = pgADV.getServiceParam(PeerGroup.discoveryClassID);
            System.out.println("serviceParam = " + serviceParam);
            pgADV.putServiceParam(PeerGroup.discoveryClassID, (StructuredDocument) discoveryConfig.getDocument(MimeMediaType.XMLUTF8));
            //added

            ris = parentGroup.newGroup(pgADV);
//            ris = parentGroup.newGroup(id, miADV, peerGroupName, peerGroupDesc, true);
//            System.out.println("SHOW NEW GROUP ADV");
//            showAllLocalADV(ris.getDiscoveryService(), DiscoveryService.GROUP);
//            showAllLocalADV(ris.getDiscoveryService(), DiscoveryService.PEER);

        } catch (Exception ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        }

        return ris;

    }

    public static ModuleImplAdvertisement createAllPurposePeerGroupImplAdv(String peerGroupDesc) {

        ModuleImplAdvertisement implAdv = CompatibilityUtils.createModuleImplAdvertisement(
                PeerGroup.allPurposePeerGroupSpecID, StdPeerGroup.class.getName(),
                peerGroupDesc);

        // Create the service list for the group.
        StdPeerGroupParamAdv paramAdv = new StdPeerGroupParamAdv();

        // set the services
        paramAdv.addService(PeerGroup.endpointClassID, PeerGroup.refEndpointSpecID);
        paramAdv.addService(PeerGroup.resolverClassID, PeerGroup.refResolverSpecID);
        paramAdv.addService(PeerGroup.membershipClassID, PeerGroup.refMembershipSpecID);
        paramAdv.addService(PeerGroup.accessClassID, PeerGroup.refAccessSpecID);

        // standard services
        paramAdv.addService(PeerGroup.discoveryClassID, PeerGroup.refDiscoverySpecID);

        paramAdv.addService(PeerGroup.rendezvousClassID, PeerGroup.refRendezvousSpecID);
        paramAdv.addService(PeerGroup.pipeClassID, PeerGroup.refPipeSpecID);
        paramAdv.addService(PeerGroup.peerinfoClassID, PeerGroup.refPeerinfoSpecID);

        paramAdv.addService(PeerGroup.contentClassID, ContentServiceImpl.MODULE_SPEC_ID);

        // Insert the newParamAdv in implAdv
        XMLElement paramElement = (XMLElement) paramAdv.getDocument(MimeMediaType.XMLUTF8);
        implAdv.setParam(paramElement);

        return implAdv;

    }

    public static PeerGroup createGroup(PeerGroup netPG, PeerGroupAdvertisement pgAdv) {
        try {
            return netPG.newGroup(pgAdv.getPeerGroupID());
        } catch (PeerGroupException ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        }

        return null;
    }

    /**
     * Used to join a group
     * 
     * @param pg
     * @return
     */
    public static boolean join(PeerGroup pg) {
        boolean success = false;
        try {
            StructuredDocument myCredentials = null;
            AuthenticationCredential myAuthenticationCredential = new AuthenticationCredential(pg, null, myCredentials);
            //FIXME jxta 2.5
//            MembershipService myMembershipService = pg.getMembershipService();
//            Authenticator myAuthenticator = myMembershipService.apply(myAuthenticationCredential);
            //FIXME jxta 2.7
            NoneMembershipService myMembershipService = new NoneMembershipService();
            Authenticator myAuthenticator = myMembershipService.apply(myAuthenticationCredential);
            if (!myAuthenticator.isReadyForJoin()) {
                System.out.println("Filed to join group");
                return false;
            }


            //TODO credential info not needed, insert only join
//            Credential c = myMembershipService.join(myAuthenticator);
//
//            StructuredTextDocument std = (StructuredTextDocument) c.getDocument(new MimeMediaType("text/plain"));
//            StringWriter out = new StringWriter();
//            std.sendToWriter(out);
//            out.close();
//            System.out.println("Credential doc " + out);

            success = true;
//        } catch (IOException ex) {
//            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        } catch (PeerGroupException ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        } catch (ProtocolNotSupportedException ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        } catch (Exception ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        }

        return success;
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Utility methods to find ADV
    ///////////////////////////////////////////////////////////////////////////
    public static LinkedList<Advertisement> getLocalADV(DiscoveryService ds, int advType, String attName, String attValue) {
        try {
            Enumeration<Advertisement> list = ds.getLocalAdvertisements(advType, attName, attValue);
            LinkedList<Advertisement> ris = new LinkedList<Advertisement>();
            while (list.hasMoreElements()) {
                ris.add(list.nextElement());
            }
            return ris;

        } catch (IOException ex) {
            ex.printStackTrace();
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        }

        return null;


    }

    public static void findAdv(DiscoveryService ds, int advType, String attName, String attValue) {

        ds.getRemoteAdvertisements(null, advType, attName, attValue, Integer.MAX_VALUE);

    }

    /**
     * Sends a remote discovery and waits for {@code timeout}ms
     * 
     * @param ds
     * @param advType
     * @param attName
     * @param attValue
     * @param timeout
     * @return The set of the adv received
     */
    public static Set<Advertisement> blockingFindAdv(final DiscoveryService ds, int advType, String attName, String attValue, long timeout) {
        return blockingFindAdv(ds, null, advType, attName, attValue, DEFAULT_THRESHOLD, timeout);
    }

    /**
     * Blocca per timeout ms. Allo scadere ritorna la lista con tutti
     * gli ADV ricevuti
     * @param ds
     * @param advType
     * @param attName
     * @param attValue
     * @param timeout
     * @return
     */
    public static Set<Advertisement> blockingFindAdv(final DiscoveryService ds, String peerID, int advType, String attName, String attValue, long timeout) {
        return blockingFindAdv(ds, peerID, advType, attName, attValue, DEFAULT_THRESHOLD, timeout);
    }

    /**
     * Sends a remote discovery and waits for {@code timeout}ms collecting
     * results.
     *
     * @param ds
     * @param peerID
     * @param advType
     * @param attName
     * @param attValue
     * @param threshold
     * @param timeout
     * @return the set of all advertisement received in the specified time
     */
    public static Set<Advertisement> blockingFindAdv(final DiscoveryService ds, String peerID, int advType, String attName, String attValue, int threshold, long timeout) {
        final Set<Advertisement> ris = new HashSet<Advertisement>();
        //TODO da eliminare. solo per testing
        final HashMap<String, LinkedList<String>> responders = new HashMap<String, LinkedList<String>>();

        DiscoveryListener dl = new DiscoveryListener() {

            public void discoveryEvent(DiscoveryEvent event) {

                LOG.log(Level.FINE, "{0}", new Object[]{getEventInfo(event)});
                LinkedList<String> peerIDs = new LinkedList<String>();
                DiscoveryResponseMsg dRM = event.getResponse();

                synchronized (ris) {
                    for (Enumeration<Advertisement> e = dRM.getAdvertisements(); e.hasMoreElements();) {
                        Advertisement a = e.nextElement();
                        ris.add(a);
                        peerIDs.add(((NodeAdvertisement) a).getPeerID());
                    }
                    LinkedList<String> put = responders.put(event.getSource().toString(), peerIDs);
                    if (put != null && put.size() > 0) {
                        LOG.log(Level.FINE, event.getSource() + " has already responded ");
                        peerIDs.addAll(put);
                    }
                    LOG.log(Level.FINE, "Responders are:" + responders.size() + "\n{0}", LoggingUtil.elementByLine(responders));
                }



            }
        };

        int queryID = ds.getRemoteAdvertisements(peerID, advType, attName, attValue, threshold, dl);

        synchronized (ris) {
            try {
                ris.wait(timeout);
                //FIXME per il testing si mantiene il dl ma va rimosso
//                ds.removeDiscoveryListener(dl);

            } catch (InterruptedException ex) {
                LOG.log(Level.SEVERE, null, ex);
            }

        }

        LOG.log(Level.FINE, "Responders are:\n{0}", LoggingUtil.elementByLine(responders));

        return ris;
    }

    /**
     * Non bloccante
     * Rimuove il listener dopo aver ricevuto la prima risp
     *
     * @param ds
     * @param advType
     * @param attName
     * @param attValue
     * @param timeout
     * @return
     */
    public static void oneResponseFindAdv(final DiscoveryService ds, String peerID, int advType, String attName, String attValue, long timeout) {

        DiscoveryListener dl = new OneResponseDL(ds);

        int queryID = ds.getRemoteAdvertisements(peerID, advType, attName, attValue, Integer.MAX_VALUE, dl);

    }

    public static Set<Advertisement> findAdv(final DiscoveryService ds, String peerID, int advType, String attName, String attValue, long timeout) {
        //TODO utilizzare solo ris tanto il listener viene rimosso
        final Set<Advertisement> set = new HashSet<Advertisement>();
        final boolean flag[] = new boolean[1];

        DiscoveryListener dl = new DiscoveryListener() {

            public void discoveryEvent(DiscoveryEvent event) {
                synchronized (set) {
                    if (event != null && !flag[0]) {
                        LOG.log(Level.CONFIG, "{0}", new Object[]{getEventInfo(event)});
                        //FIXME
                        //Forzo la publ locale del mio adv che potrebbe essere stato sovrascritto da questo evento

                        DiscoveryResponseMsg dRM = event.getResponse();
                        Enumeration<Long> esp = dRM.getExpirations();

                        for (Enumeration<Advertisement> e = dRM.getAdvertisements(); e.hasMoreElements();) {
                            set.add(e.nextElement());
                        }
                        flag[0] = true;
                        set.notifyAll();

                    } else {
                        System.out.println("RECEIVED A NULL DISCOVERY EVENT");
                    }
                }
            }
        };

        int queryID = ds.getRemoteAdvertisements(peerID, advType, attName, attValue, Integer.MAX_VALUE, dl);

        synchronized (set) {
            try {

                while (!flag[0]) {
                    set.wait();
                }
                //Response received
                ds.removeDiscoveryListener(dl);

            } catch (InterruptedException ex) {
                Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
            }

        }


        return set;
    }

    /**
     * Adds an Object to message within the specified name space and with the specified element name
     *
     * @param message   the message to add the object to
     * @param nameSpace the name space to add the object under
     * @param elemName  the given element name
     * @param object    the object
     * @throws IOException if an io error occurs
     */
    public static void addObjectToMessage(Message message, String nameSpace, String elemName, Object object) throws IOException {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        ObjectOutputStream oos = new ObjectOutputStream(bos);
        oos.writeObject(object);
        oos.close();
        bos.close();
        //TODO use compression
        addByteArrayToMessage(message, nameSpace, elemName, bos.toByteArray(), false);
    }

    /**
     * Adds an byte array to a message
     *
     * @param message   The message to add to
     * @param nameSpace The namespace of the element to add. a null value assumes default namespace.
     * @param elemName  Name of the Element.
     * @param data      the byte array
     * @param compress  indicates whether to use GZIP compression
     * @throws IOException if an io error occurs
     */
    public static void addByteArrayToMessage(Message message, String nameSpace, String elemName, byte[] data, boolean compress) throws IOException {
        byte[] buffer = data;
        MimeMediaType mimeType = MimeMediaType.AOS;

        if (compress) {
            ByteArrayOutputStream outStream = new ByteArrayOutputStream();
            GZIPOutputStream gos = new GZIPOutputStream(outStream);

            gos.write(data, 0, data.length);
            gos.finish();
            gos.close();
            buffer = outStream.toByteArray();
            mimeType = GZIP_MEDIA_TYPE;
        }
        message.addMessageElement(nameSpace, new ByteArrayMessageElement(elemName, mimeType, buffer, null));
    }

    /**
     * Reads a single Java Object from a Message.
     *
     * @param message   The message containing the object.
     * @param nameSpace The name space of the element containing the object.
     * @param elemName  The name of the element containing the object.
     * @return The Object or {@code null} if the Message contained no such element.
     * @throws IOException if an io error occurs
     * @throws ClassNotFoundException if an object could not constructed from the message element
     */
    public static Object getObjectFromMessage(Message message, String nameSpace, String elemName) throws IOException, ClassNotFoundException {
        InputStream is = getInputStreamFromMessage(message, nameSpace, elemName);

        if (null == is) {
            return null;
        }
        ObjectInputStream ois = new ObjectInputStream(is);

        return ois.readObject();
    }

    /**
     * Returns an InputStream for a byte array
     *
     * @param message   The message to retrieve from
     * @param nameSpace The namespace of the element to get.
     * @param elemName  Name of the Element.
     * @return The {@code InputStream} or {@code null} if the message has no such element, String elemName) throws IOException {
     * @throws IOException if an io error occurs
     */
    public static InputStream getInputStreamFromMessage(Message message, String nameSpace, String elemName) throws IOException {
        InputStream result = null;
        MessageElement element = message.getMessageElement(nameSpace, elemName);

        if (null == element) {
            return null;
        }

        if (element.getMimeType().equals(GZIP_MEDIA_TYPE)) {
            result = new GZIPInputStream(element.getStream());
        } else if (element.getMimeType().equals(MimeMediaType.AOS)) {
            result = element.getStream();
        }
        return result;
    }

    private static Object getEventInfo(final DiscoveryEvent event) {
        Object o = new Object() {

            @Override
            public String toString() {
                return getEventDesc(event).toString();
            }
        };

        return o;
    }

    private static StringBuilder getEventDesc(DiscoveryEvent event) {
        DiscoveryResponseMsg dRM = event.getResponse();
        PeerAdvertisement respPeerAdv = dRM.getPeerAdvertisement();
        StringBuilder sb = new StringBuilder();
        sb.append("\n**** RECEIVED A DISCOVERY EVENT FROM ").append(event.getSource()).append(" ****\n");
        sb.append("RESPONDER ADV=\n").append(respPeerAdv).append("\n");
        sb.append("RESPONSES COUNT= ").append(dRM.getResponseCount()).append("\n");
        sb.append("RESPONSES:\n");
        sb.append(getADVInfo(dRM.getAdvertisements(), dRM.getExpirations()));
        sb.append("**** END ****");
        return sb;
    }

    /**
     * Show on stout all the local adv of the specified type
     * @param advType
     */
    public static void showAllLocalADV(DiscoveryService ds, int advType) {
        try {
            System.out.println("---- SHOW ALL LOCAL ADV OF TYPE " + advType(advType) + " ----");
            Enumeration<Advertisement> en = ds.getLocalAdvertisements(advType, null, null);

            showADV(en);

            System.out.println("---- END ---- ");
        } catch (IOException ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public static void showAllLocalADV(DiscoveryService ds, int advType, String attName, String attValue) {
        try {
            System.out.println("---- SHOW ALL LOCAL ADV OF TYPE " + advType(advType) + " att:" + attName + " val:" + attValue + "----");
            Enumeration<Advertisement> en = ds.getLocalAdvertisements(advType, attName, attValue);

            showADV(ds, en);

            System.out.println("---- END ---- ");
        } catch (IOException ex) {
            Logger.getLogger(JxtaUtil.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Return a string with the name of the adv type specified
     * by {@code advType } (ADV,GROUP or PEER)
     * @param advType
     * @return
     */
    public static String advType(int advType) {

        switch (advType) {
            case DiscoveryService.ADV:
                return "ADV";
            case DiscoveryService.GROUP:
                return "GROUP";
            case DiscoveryService.PEER:
                return "PEER";
            default:
                throw new IllegalArgumentException("ADV type not valid: " + advType);
        }
    }

    public static void showADV(Enumeration<Advertisement> advertisements) {

        if (advertisements == null) {
            return;
        }
        int count = 0;
        while (advertisements.hasMoreElements()) {
            Advertisement adv = advertisements.nextElement();
            System.out.println("ADV " + count);
            System.out.println(adv);
            count++;
        }
    }

    public static void showADV(DiscoveryService ds, Enumeration<Advertisement> advertisements) {

        if (advertisements == null) {
            return;
        }
        int count = 0;
        while (advertisements.hasMoreElements()) {
            Advertisement adv = advertisements.nextElement();
            System.out.println("ADV " + count + " lt: " + ds.getAdvLifeTime(adv) + " exp: " + ds.getAdvExpirationTime(adv));
            System.out.println(adv);
            count++;
        }
    }

    public static String getADVInfo(Enumeration<Advertisement> advertisements) {

        if (advertisements == null) {
            return "";
        }
        StringBuilder sb = new StringBuilder();
        int count = 0;
        while (advertisements.hasMoreElements()) {
            Advertisement adv = advertisements.nextElement();
            sb.append("ADV ").append(count).append("\n");
            sb.append(adv);
            count++;
        }

        return sb.toString();
    }

    public static StringBuilder getADVInfo(Enumeration<Advertisement> advertisements, Enumeration<Long> expirations) {

        if (advertisements == null) {
            return new StringBuilder();
        }
        StringBuilder sb = new StringBuilder();
        int count = 0;
        while (advertisements.hasMoreElements()) {
            Advertisement adv = advertisements.nextElement();
            sb.append("ADV ").append(count).append(" ").append(expirations.nextElement()).append("\n");
            sb.append(adv);
            count++;
        }

        return sb;
    }

    public static void showEnumeration(Enumeration enumeration) {

        if (enumeration == null) {
            return;
        }
        int count = 0;
        while (enumeration.hasMoreElements()) {
            Object o = enumeration.nextElement();
            System.out.println("Element " + count);
            System.out.println(o);
            count++;
        }
    }

    /**
     * Recursively deletes the directory specified by the File parameter
     * @param directory
     */
    public static void recursiveDelete(File directory) {

        File[] subFiles = directory.listFiles();

        if (subFiles != null) {
            for (int i = 0; i < subFiles.length; i++) {
                if (subFiles[i].isDirectory()) {
                    recursiveDelete(subFiles[i]);
                }
                subFiles[i].delete();
            }
            directory.delete();
        }

    }

    private static class OneResponseDL implements DiscoveryListener {

        private final DiscoveryService ds;

        public OneResponseDL(DiscoveryService ds) {
            this.ds = ds;
        }
        boolean done;

        public synchronized void discoveryEvent(DiscoveryEvent event) {
            if (event != null && !done) {
                LOG.log(Level.CONFIG, "{0}", new Object[]{getEventInfo(event)});
                //FIXME Forzare la publ locale del mio adv che potrebbe essere stato sovrascritto da questo evento
                DiscoveryResponseMsg dRM = event.getResponse();
                Enumeration<Long> esp = dRM.getExpirations();
                done = true;
                ds.removeDiscoveryListener(this);
            }
        }
    }
}
