/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.impl.networkModule.jxta;

import java.io.IOException;
import java.io.InputStream;
import net.jxta.id.ID;
import net.jxta.id.IDFactory;

import java.io.Serializable;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Enumeration;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.jxta.document.Advertisement;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.Attributable;
import net.jxta.document.Document;
import net.jxta.document.Element;
import net.jxta.document.MimeMediaType;
import net.jxta.document.StructuredDocument;
import net.jxta.document.StructuredDocumentFactory;
import net.jxta.document.StructuredTextDocument;
import net.jxta.document.TextElement;
import p2pMapReduce.nodeModule.Constants;

/**
 * Create a advertisement totally describing a node
 *
 * //TODO if possible make this Advertisement to handle generic properties like
 * a Properties object, and implementing {@link java.io.Externalizable } to
 * serialize all key,value pairs
 */
public class NodeAdvertisement extends Advertisement implements Comparable<NodeAdvertisement>, Cloneable, Serializable {

    public static final String ADV_TYPE = "jxta:NodeAdvertisement";
    public static final String PREFIX = "NodeAdvertisement-";
    public static final String ID_TAG = PREFIX + "ID";
    public static final String VERSION_TAG = PREFIX + "version";
    public static final String AVAILABLE_TAG = PREFIX + "available";
    public static final String TYPE_TAG = PREFIX + "type";
    public static final String ROLE_TAG = PREFIX + "role";
    public static final String LOAD_TAG = PREFIX + "load";
    public static final String PEER_ID_TAG = PREFIX + "peer_id";
    public static final String PIPE_ID_TAG = PREFIX + "pipe_id";
    public static final String IP_ADDRESS_TAG = PREFIX + "ip_address";
    public static final String FTP_PORT_TAG = PREFIX + "ftp_port";
    /**
     * Indexable fields.  Advertisements must define the indexables, in order
     * to properly index and retrieve these advertisements locally and on the
     * network
     *
     * //TODO
     * lasciando indicizzabile la versione si permette lo stesso adv con versioni diverse
     * 
     */
    private static final String[] INDEXABLE_FILEDS = {ID_TAG, VERSION_TAG, TYPE_TAG, ROLE_TAG,/* LOAD_TAG,*/ PEER_ID_TAG/*, PIPE_ID_TAG*/};

    static {
        // The following step is required and only need to be done once,
        // without this step the AdvertisementFactory has no means of
        // associating an advertisement name space with the proper obect
        // in this cast the AdvertisementTutorial
        AdvertisementFactory.registerAdvertisementInstance(NodeAdvertisement.getAdvertisementType(),
                new NodeAdvertisement.Instantiator());
    }
    /**
     * Advertisement ID
     */
    private ID id;
    /**
     * Version of this NodeAdvertisement
     */
    private long version;
    /**
     * Indicate if this node is available at the moment
     */
    private boolean available;
    /**
     * Node role: {@link Constants#SLAVE} or {@link Constants#MASTER}
     */
    private int role;
    /**
     * Node load
     */
    private double load;
    /**
     * //TODO useful to use a PeerID instance directly ??
     * The string that represents the PeerID of the peer associated to this
     * node
     */
    private String peerID;
    /**
     * The string that represents the PipeID of the pipe associated to this
     * node
     */
    private String pipeID;
    /**
     * The string contains the ip address of this node
     */
    private String ipAddress;
    /**
     * The ftp port where the ftp server of this node is listening
     */
    private int ftpPort;

    /**
     * Default Constructor
     */
    public NodeAdvertisement() {
        id = ID.nullID;
    }

    public NodeAdvertisement(NodeAdvertisement nodeADV) {
        this.id = nodeADV.id;
        this.version = nodeADV.version;
        this.available = nodeADV.available;
        this.role = nodeADV.role;
        this.load = nodeADV.load;
        this.peerID = nodeADV.peerID;
        this.pipeID = nodeADV.pipeID;
        this.ipAddress = nodeADV.ipAddress;
        this.ftpPort = nodeADV.ftpPort;
    }

    public NodeAdvertisement(ID id) {
        this.id = id;
    }

    /**
     * Construct from a StructuredDocument
     *
     * @param root Root element
     */
    public NodeAdvertisement(Element root) {
        TextElement doc = (TextElement) root;

        if (!getAdvertisementType().equals(doc.getName())) {
            throw new IllegalArgumentException(
                    "Could not construct : " + getClass().getName() + "from doc containing a " + doc.getName());
        }
        initialize(doc);
    }

    /**
     * Construct a doc from InputStream
     *
     * @param stream the underlying input stream.
     * @throws IOException if an I/O error occurs.
     */
    public NodeAdvertisement(InputStream stream) throws IOException {
        StructuredTextDocument doc = (StructuredTextDocument) StructuredDocumentFactory.newStructuredDocument(MimeMediaType.XMLUTF8, stream);
        initialize(doc);
    }

    /**
     * {@inheritDoc}
     *
     * @param asMimeType Document encoding
     * @return The document value
     */
    @Override
    public Document getDocument(MimeMediaType asMimeType) {
        StructuredDocument adv = StructuredDocumentFactory.newStructuredDocument(asMimeType, getAdvertisementType());

        if (adv instanceof Attributable) {
            ((Attributable) adv).addAttribute("xmlns:jxta", "http://jxta.org");
        }
        Element e;

        e = adv.createElement(ID_TAG, getID().toString());
        adv.appendChild(e);
        e = adv.createElement(VERSION_TAG, "" + version);
        adv.appendChild(e);
        e = adv.createElement(AVAILABLE_TAG, "" + available);
        adv.appendChild(e);
        e = adv.createElement(TYPE_TAG, getAdvertisementType());
        adv.appendChild(e);
        e = adv.createElement(ROLE_TAG, "" + role);
        adv.appendChild(e);
        e = adv.createElement(LOAD_TAG, "" + load);
        adv.appendChild(e);
        e = adv.createElement(PEER_ID_TAG, peerID);
        adv.appendChild(e);
        e = adv.createElement(PIPE_ID_TAG, pipeID);
        adv.appendChild(e);
        e = adv.createElement(IP_ADDRESS_TAG, ipAddress);
        adv.appendChild(e);
        e = adv.createElement(FTP_PORT_TAG, "" + ftpPort);
        adv.appendChild(e);

        return adv;
    }

    /**
     * returns the id of the device
     *
     * @return ID the device id
     */
    @Override
    public ID getID() {
        return id;
    }

    /**
     * sets the unique id
     *
     * @param id The id
     */
    public void setID(ID id) {
        this.id = (id == null ? null : id);
    }

    public long getVersion() {
        return version;
    }

    public void setVersion(long version) {
        this.version = version;
    }

    public boolean isAvailable() {
        return available;
    }

    public void setAvailable(boolean available) {
        this.available = available;
    }

    public double getLoad() {
        return load;
    }

    public void setLoad(double load) {
        this.load = load;
    }

    public int getRole() {
        return role;
    }

    public void setRole(int role) {
        this.role = role;
    }

    public String getPeerID() {
        return peerID;
    }

    public void setPeerID(String peerID) {
        this.peerID = peerID;
    }

    public String getPipeID() {
        return pipeID;
    }

    public void setPipeID(String pipeID) {
        this.pipeID = pipeID;
    }

    public int getFtpPort() {
        return ftpPort;
    }

    public void setFtpPort(int ftpPort) {
        this.ftpPort = ftpPort;
    }

    public String getIpAddress() {
        return ipAddress;
    }

    public void setIpAddress(String ipAddress) {
        this.ipAddress = ipAddress;
    }

    /**
     * Process an individual element from the document.
     *
     * @param elem the element to be processed.
     * @return true if the element was recognized, otherwise false.
     */
    protected boolean handleElement(TextElement elem) {
        final String name = elem.getName();
        if (name.equals(ID_TAG)) {
            try {
                URI idURI = new URI(elem.getTextValue());

                setID(IDFactory.fromURI(idURI));
            } catch (URISyntaxException badID) {
                throw new IllegalArgumentException("unknown ID format in advertisement: " + elem.getTextValue());
            } catch (ClassCastException badID) {
                throw new IllegalArgumentException("Id is not a known id type: " + elem.getTextValue());
            }
            return true;
        }
        if (name.equals(VERSION_TAG)) {
            setVersion(Long.parseLong(elem.getTextValue()));
            return true;
        }
        if (name.equals(AVAILABLE_TAG)) {
            setAvailable(Boolean.parseBoolean(elem.getTextValue()));
            return true;
        }
        if (name.equals(TYPE_TAG)) {
            return true;
        }
        if (name.equals(ROLE_TAG)) {
            setRole(Integer.parseInt(elem.getTextValue()));
            return true;
        }
        if (name.equals(LOAD_TAG)) {
            setLoad(Double.parseDouble(elem.getTextValue()));
            return true;
        }
        if (name.equals(PEER_ID_TAG)) {
            setPeerID(elem.getTextValue());
            return true;
        }
        if (name.equals(PIPE_ID_TAG)) {
            setPipeID(elem.getTextValue());
            return true;
        }
        if (name.equals(IP_ADDRESS_TAG)) {
            setIpAddress(elem.getTextValue());
            return true;
        }
        if (name.equals(FTP_PORT_TAG)) {
            setFtpPort(Integer.parseInt(elem.getTextValue()));
            return true;
        }

        // element was not handled
        return false;
    }

    /**
     * Initialize a System advertisement from a portion of a structured document.
     *
     * @param root document root
     */
    protected void initialize(Element root) {
        if (!TextElement.class.isInstance(root)) {
            throw new IllegalArgumentException(getClass().getName() + " only supports TextElement");
        }
        TextElement doc = (TextElement) root;

        if (!doc.getName().equals(getAdvertisementType())) {
            throw new IllegalArgumentException(
                    "Could not construct : " + getClass().getName() + "from doc containing a " + doc.getName());
        }
        Enumeration elements = doc.getChildren();

        while (elements.hasMoreElements()) {
            TextElement elem = (TextElement) elements.nextElement();

            if (!handleElement(elem)) {
                Logger.getLogger(NodeAdvertisement.class.getName()).log(Level.SEVERE, "Unhandleded element \'" + elem.getName() + "\' in " + doc.getName());
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final String[] getIndexFields() {
        return INDEXABLE_FILEDS;
    }

    /**
     * Equals if same ID and same version
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj instanceof NodeAdvertisement) {
            NodeAdvertisement adv = (NodeAdvertisement) obj;

            return getID().equals(adv.getID());
        }

        return false;
    }

    @Override
    public int hashCode() {
        int hash = 3;
        hash = 37 * hash + (this.id != null ? this.id.toString().hashCode() : 0);
        /*hash = 37 * hash + new Long(this.version).hashCode()*/ ;
        return hash;
    }

    public int compareTo(NodeAdvertisement other) {
        if (this.equals(other)) {
            return 0;
        } else {
            if (this.getPeerID().equals(other.getPeerID())) {
                //If they refer to same peer order them by decreasing version values
                //Same versions value is not possible because they are not equal
                //so 0 will be never returned for different NodeADV
                return (int) (other.version - version);
            } else {
                //They refer to different peer
                //order them by id value
                //Surely 0 will be never returned because the id string are different
                return getID().toString().compareTo(other.getID().toString());

            }
        }
    }

    @Override
    public NodeAdvertisement clone() throws CloneNotSupportedException {
        NodeAdvertisement clone = (NodeAdvertisement) super.clone();
        clone.setID(getID());
        clone.setVersion(getVersion());
        clone.setAvailable(clone.isAvailable());
        clone.setRole(getRole());
        clone.setLoad(getLoad());
        clone.setPeerID(getPeerID());
        clone.setPipeID(getPipeID());
        clone.setIpAddress(getIpAddress());
        clone.setFtpPort(getFtpPort());

        return clone;
    }

    /**
     * All messages have a type (in xml this is &#0033;doctype) which
     * identifies the message
     *
     * @return String "jxta:NodeAdvertisement"
     */
    public static String getAdvertisementType() {
        return ADV_TYPE;
    }

    @Override
    public String getAdvType() {
        return getAdvertisementType();
    }

    /**
     * Utility method to increment version number
     */
    public void incrVersion() {
        version++;
    }

    /**
     * Instantiator
     */
    public static class Instantiator implements AdvertisementFactory.Instantiator {

        /**
         * Returns the identifying type of this Advertisement.
         *
         * @return String the type of advertisement
         */
        public String getAdvertisementType() {
            return NodeAdvertisement.getAdvertisementType();
        }

        /**
         * Constructs an instance of <CODE>Advertisement</CODE> matching the
         * type specified by the <CODE>advertisementType</CODE> parameter.
         *
         * @return The instance of <CODE>Advertisement</CODE> or null if it
         *         could not be created.
         */
        public Advertisement newInstance() {
            return new NodeAdvertisement();
        }

        /**
         * Constructs an instance of <CODE>Advertisement</CODE> matching the
         * type specified by the <CODE>advertisementType</CODE> parameter.
         *
         * @param root Specifies a portion of a StructuredDocument which will
         *             be converted into an Advertisement.
         * @return The instance of <CODE>Advertisement</CODE> or null if it
         *         could not be created.
         */
        public Advertisement newInstance(net.jxta.document.Element root) {
            return new NodeAdvertisement(root);
        }
    }

}
