/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.impl.nodeModule;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.fs.FileSystemView;
import p2pMapReduce.networkModule.NetworkConfiguration;
import p2pMapReduce.networkModule.NetworkModuleFactory;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.entity.BackupJob;
import p2pMapReduce.nodeModule.nodeFsm.NodeFSM;
import p2pMapReduce.nodeModule.NodeModuleIF;
import p2pMapReduce.nodeModule.TaskManager;
import p2pMapReduce.entity.PrimaryJob;
import p2pMapReduce.entity.SlaveTask;
import p2pMapReduce.ftp.FTPManager;
import p2pMapReduce.nodeModule.event.EventIF;
import p2pMapReduce.nodeModule.nodeFsm.NodeState.StateName;

/**
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class NodeModuleImpl implements NodeModuleIF {

    private static final transient Logger LOG = Logger.getLogger(NodeModuleImpl.class.getName());
    private NetworkModuleIF networkModule;
    private FileSystemView fsView;
    private NodeFSM nodeFSM;
    private NodeInfo nodeInfo;
    private LinkedList<PrimaryJob> primaryJobList;
    private LinkedList<BackupJob> backupJobList;
    private HashMap<String, SlaveTask> slaveTaskMap;
    private HashMap<String, TaskManager> taskManagerMap;
    private String coordinatorID;

    public NodeModuleImpl(NetworkConfiguration nc) throws IOException {

        primaryJobList = new LinkedList<PrimaryJob>();
        backupJobList = new LinkedList<BackupJob>();
        slaveTaskMap = new HashMap<String, SlaveTask>();
        taskManagerMap = new HashMap<String, TaskManager>();
        
        networkModule = NetworkModuleFactory.newNetworkModule(nc);
        networkModule.setNodeModule(this);
        networkModule.addEventListener(this);
        nodeInfo = networkModule.joinNetwork();

        //Must be create after the network module
        fsView = new FileSystemView(getFtpManager().getBaseDir());
        nodeFSM = new NodeFSM(this);
    }

    public NodeInfo getNodeInfo() {
        return nodeInfo;
    }

    public NodeFSM getNodeFSM() {
        return nodeFSM;
    }

    public String getNodeID() {
        return nodeInfo.getID();
    }

    public String getCoordinatorID() {
        return coordinatorID;
    }

    public void setCoordinatorID(String coordinatorID) {
        this.coordinatorID = coordinatorID;
    }

    public boolean isCoordinator() {
        return nodeInfo.getID().equals(coordinatorID);
    }

    public NetworkModuleIF getNetworkModule() {
        return networkModule;
    }

    public FTPManager getFtpManager() {
        return networkModule.getFtpManager();
    }

    public FileSystemView getFileSystemView() {
        return fsView;
    }

    public LinkedList<BackupJob> getBackupJobList() {
        return backupJobList;
    }

    public LinkedList<PrimaryJob> getPrimaryJobList() {
        return primaryJobList;
    }

    public HashMap<String, SlaveTask> getSlaveTaskMap() {
        return slaveTaskMap;
    }

    public void addPrimaryJob(PrimaryJob pj) {
        primaryJobList.add(pj);
    }

    public PrimaryJob removePrimaryJob(String jobID) {
        for (Iterator<PrimaryJob> it = primaryJobList.iterator(); it.hasNext();) {
            PrimaryJob pj = it.next();
            if (pj.getJobID().equals(jobID)) {
                it.remove();
                return pj;
            }
        }

        return null;
    }

    public PrimaryJob getPrimaryJob(String jobID) {
        for (PrimaryJob pj : primaryJobList) {
            if (pj.getJobID().equals(jobID)) {
                return pj;
            }
        }

        return null;
    }

    public void addBackupJob(BackupJob bj) {
        backupJobList.add(bj);
    }

    public BackupJob removeBackupJob(String jobID) {
        for (Iterator<BackupJob> it = backupJobList.iterator(); it.hasNext();) {
            BackupJob bj = it.next();
            if (bj.getJobID().equals(jobID)) {
                it.remove();
                return bj;
            }

        }

        return null;
    }

    public BackupJob getBackupJob(String jobID) {
        for (BackupJob bj : backupJobList) {
            if (bj.getJobID().equals(jobID)) {
                return bj;
            }
        }

        return null;
    }

    public void addSlaveTask(SlaveTask st) {
        slaveTaskMap.put(st.getTask().getTaskAttIDAsString(), st);
    }

    public SlaveTask removeSlaveTask(String taskID) {
        return slaveTaskMap.remove(taskID);
    }

    public SlaveTask getSlaveTask(String taskID) {
        return slaveTaskMap.get(taskID);
    }

    public void addTaskManager(String taskID, TaskManager tm) {
        taskManagerMap.put(taskID, tm);
    }

    public TaskManager getTaskManager(String taskID) {
        return taskManagerMap.get(taskID);
    }

    public TaskManager removeTaskManager(String taskID) {
        return taskManagerMap.remove(taskID);
    }

    public void handleEvent(EventIF event) {
        LOG.log(Level.CONFIG, "Received new event: " + event);
        switch (event.getType()) {
            default:
                nodeFSM.handleEvent(event);
        }
    }

    /**
     * Dangerous forcing of state transition.
     * Only for testing purpose, will be removed
     * @param stateName
     */
    public void forceTransition(StateName stateName) {
        nodeFSM.forceTransition(stateName);

    }
}
