/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 *  This code has been partially modified as part of P2P-MapReduce.
 *
 *  Copyright (C) 2011 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.mapreduceModule;

import java.io.Serializable;

/**************************************************
 * Describes the current status of a job.
 **************************************************/
public class JobStatus implements Serializable, Cloneable {

    /**
     * Current state of the job
     */
    public static enum State {

        RUNNING(1),
        SUCCEEDED(2),
        FAILED(3),
        PREP(4),
        KILLED(5);
        int value;

        State(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }
    };
    private JobID jobid;
    private float mapProgress;
    private float reduceProgress;
    private float cleanupProgress;
    private float setupProgress;
    private State runState;
    private long startTime;
    private String user;
    private String schedulingInfo = "NA";
    private String jobName;
    private String jobFile;
    private long finishTime;
    private boolean isRetired;
    private String historyFile = "";
    private String trackingUrl = "";

    /**
     */
    public JobStatus() {
    }

    /**
     * Create a job status object for a given jobid.
     * @param jobid The jobid of the job
     * @param setupProgress The progress made on the setup
     * @param mapProgress The progress made on the maps
     * @param reduceProgress The progress made on the reduces
     * @param cleanupProgress The progress made on the cleanup
     * @param runState The current state of the job
     * @param jp Priority of the job.
     * @param user userid of the person who submitted the job.
     * @param jobName user-specified job name.
     * @param jobFile job configuration file.
     * @param trackingUrl link to the web-ui for details of the job.
     */
    public JobStatus(JobID jobid, float setupProgress, float mapProgress,
            float reduceProgress, float cleanupProgress,
            State runState, String user, String jobName,
            String jobFile, String trackingUrl) {
        this.jobid = jobid;
        this.setupProgress = setupProgress;
        this.mapProgress = mapProgress;
        this.reduceProgress = reduceProgress;
        this.cleanupProgress = cleanupProgress;
        this.runState = runState;
        this.user = user;
        this.jobName = jobName;
        this.jobFile = jobFile;
        this.trackingUrl = trackingUrl;
    }

    /**
     * Sets the map progress of this job
     * @param p The value of map progress to set to
     */
    public synchronized void setMapProgress(float p) {
        this.mapProgress = (float) Math.min(1.0, Math.max(0.0, p));
    }

    /**
     * Sets the cleanup progress of this job
     * @param p The value of cleanup progress to set to
     */
    protected synchronized void setCleanupProgress(float p) {
        this.cleanupProgress = (float) Math.min(1.0, Math.max(0.0, p));
    }

    /**
     * Sets the setup progress of this job
     * @param p The value of setup progress to set to
     */
    protected synchronized void setSetupProgress(float p) {
        this.setupProgress = (float) Math.min(1.0, Math.max(0.0, p));
    }

    /**
     * Sets the reduce progress of this Job
     * @param p The value of reduce progress to set to
     */
    public synchronized void setReduceProgress(float p) {
        this.reduceProgress = (float) Math.min(1.0, Math.max(0.0, p));
    }

    /**
     * Set the finish time of the job
     * @param finishTime The finishTime of the job
     */
    protected synchronized void setFinishTime(long finishTime) {
        this.finishTime = finishTime;
    }

    /**
     * Set the job history file url for a completed job
     */
    protected synchronized void setHistoryFile(String historyFile) {
        this.historyFile = historyFile;
    }

    /**
     * Set the link to the web-ui for details of the job.
     */
    protected synchronized void setTrackingUrl(String trackingUrl) {
        this.trackingUrl = trackingUrl;
    }

    /**
     * Set the job retire flag to true.
     */
    protected synchronized void setRetired() {
        this.isRetired = true;
    }

    /**
     * Change the current run state of the job.
     */
    protected synchronized void setState(State state) {
        this.runState = state;
    }

    /**
     * Set the start time of the job
     * @param startTime The startTime of the job
     */
    protected synchronized void setStartTime(long startTime) {
        this.startTime = startTime;
    }

    /**
     * @param userName The username of the job
     */
    protected synchronized void setUsername(String userName) {
        this.user = userName;
    }

    /**
     * Used to set the scheduling information associated to a particular Job.
     *
     * @param schedulingInfo Scheduling information of the job
     */
    protected synchronized void setSchedulingInfo(String schedulingInfo) {
        this.schedulingInfo = schedulingInfo;
    }

    /**
     * @return Percentage of progress in maps
     */
    public synchronized float getMapProgress() {
        return mapProgress;
    }

    /**
     * @return Percentage of progress in cleanup
     */
    public synchronized float getCleanupProgress() {
        return cleanupProgress;
    }

    /**
     * @return Percentage of progress in setup
     */
    public synchronized float getSetupProgress() {
        return setupProgress;
    }

    /**
     * @return Percentage of progress in reduce
     */
    public synchronized float getReduceProgress() {
        return reduceProgress;
    }

    /**
     * @return running state of the job
     */
    public synchronized State getState() {
        return runState;
    }

    /**
     * @return start time of the job
     */
    synchronized public long getStartTime() {
        return startTime;
    }

    @Override
    public Object clone() {
        try {
            return super.clone();
        } catch (CloneNotSupportedException cnse) {
            // Shouldn't happen since we do implement Clonable
            throw new InternalError(cnse.toString());
        }
    }

    /**
     * @return The jobid of the Job
     */
    public JobID getJobID() {
        return jobid;
    }

    /**
     * @return the username of the job
     */
    public synchronized String getUsername() {
        return this.user;
    }

    /**
     * Gets the Scheduling information associated to a particular Job.
     * @return the scheduling information of the job
     */
    public synchronized String getSchedulingInfo() {
        return schedulingInfo;
    }

    /**
     * Returns true if the status is for a completed job.
     */
    public synchronized boolean isJobComplete() {
        return (runState == JobStatus.State.SUCCEEDED
                || runState == JobStatus.State.FAILED
                || runState == JobStatus.State.KILLED);
    }

    /**
     * Get the user-specified job name.
     */
    public String getJobName() {
        return jobName;
    }

    /**
     * Get the configuration file for the job.
     */
    public String getJobFile() {
        return jobFile;
    }

    /**
     * Get the link to the web-ui for details of the job.
     */
    public synchronized String getTrackingUrl() {
        return trackingUrl;
    }

    /**
     * Get the finish time of the job.
     */
    public synchronized long getFinishTime() {
        return finishTime;
    }

    /**
     * Check whether the job has retired.
     */
    public synchronized boolean isRetired() {
        return isRetired;
    }

    /**
     * @return the job history file name for a completed job. If job is not
     * completed or history file not available then return null.
     */
    public synchronized String getHistoryFile() {
        return historyFile;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("job-id : ").append(jobid);
        sb.append("map-progress : ").append(mapProgress);
        sb.append("reduce-progress : ").append(reduceProgress);
        sb.append("cleanup-progress : ").append(cleanupProgress);
        sb.append("setup-progress : ").append(setupProgress);
        sb.append("runstate : ").append(runState);
        sb.append("start-time : ").append(startTime);
        sb.append("user-name : ").append(user);
        sb.append("scheduling-info : ").append(schedulingInfo);
        return sb.toString();
    }
}
