/*
 * P2P-MapReduce
 * 
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.mapreduceModule;

import java.util.Iterator;
import p2pMapReduce.mapreduceModule.conf.Configuration;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.NoSuchElementException;
import p2pMapReduce.mapreduceModule.io.serializer.Deserializer;
import p2pMapReduce.mapreduceModule.io.serializer.SerializationFactory;
import p2pMapReduce.mapreduceModule.output.RecordWriter;

/**
 * NOTE: the format of intermediate result depends on {@link MapContext#write(K,V)}
 * method that refer to {@link RecordWriter}
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class ReduceContextImpl<KEYIN, VALUEIN, KEYOUT, VALUEOUT>
        extends TaskInputOutputContextImpl<KEYIN, VALUEIN, KEYOUT, VALUEOUT>
        implements ReduceContext<KEYIN, VALUEIN, KEYOUT, VALUEOUT> {

    private String partitionFile;
//    private BufferedReader br;
    private InputStream input;
    private KEYIN key;                                  // current key
    private VALUEIN value;                              // current value
    private boolean firstValue = false;                 // first value in key
    private boolean nextKeyIsSame = false;              // more w/ this key
    private boolean hasMore;
    private Deserializer<KEYIN> keyDeserializer;
    private Deserializer<VALUEIN> valueDeserializer;
    private Iterable<VALUEIN> iterable;
    private KeyValueInput keyValueInput;

    public ReduceContextImpl(Configuration conf, TaskAttemptID taskid, String partitionFile, RecordWriter<KEYOUT, VALUEOUT> output, OutputCommitter committer, StatusReporter reporter, Class<KEYIN> keyClass, Class<VALUEIN> valueClass) throws FileNotFoundException, IOException {
        super(conf, taskid, output, committer, reporter);
        this.partitionFile = partitionFile;
        input = new DataInputStream(new FileInputStream(partitionFile));
        SerializationFactory sf = new SerializationFactory(conf);
        keyDeserializer = sf.getDeserializer(keyClass);
        keyDeserializer.open(input);
        valueDeserializer = sf.getDeserializer(valueClass);
        valueDeserializer.open(input);
        keyValueInput = new KeyValueInput();
        hasMore = keyValueInput.next();
        iterable = new ValueIterable();
    }

    public boolean nextKey() throws IOException, InterruptedException {
        while (hasMore && nextKeyIsSame) {
            nextKeyValue();
        }
        if (hasMore) {
//            if (inputKeyCounter != null) {
//                inputKeyCounter.increment(1);
//            }
            return nextKeyValue();
        } else {
            return false;
        }
    }

    /**
     * Advance to the next key/value pair.
     */
    @Override
    public boolean nextKeyValue() throws IOException, InterruptedException {
        if (!hasMore) {
            key = null;
            value = null;
            return false;
        }
        firstValue = !nextKeyIsSame;

        key = keyValueInput.getCurrentkey();
        value = keyValueInput.getCurrentValue();

//        if (isMarked) {
//            backupStore.write(nextKey, nextVal);
//        }

        hasMore = keyValueInput.next();
        if (hasMore) {
            KEYIN nextKey = keyValueInput.getCurrentkey();
            nextKeyIsSame = key.equals(nextKey);
        } else {
            nextKeyIsSame = false;
        }
//        inputValueCounter.increment(1);
        return true;
    }

    public KEYIN getCurrentKey() {
        return key;
    }

    @Override
    public VALUEIN getCurrentValue() {
        return value;
    }

    protected class ValueIterator implements ReduceContext.ValueIterator<VALUEIN> {

        private boolean inReset = false;
        private boolean clearMarkFlag = false;

        @Override
        public boolean hasNext() {
//            try {
//                if (inReset && backupStore.hasNext()) {
//                    return true;
//                }
//            } catch (Exception e) {
//                e.printStackTrace();
//                throw new RuntimeException("hasNext failed", e);
//            }
            return firstValue || nextKeyIsSame;
        }

        @Override
        public VALUEIN next() {
//            if (inReset) {
//                try {
//                    if (backupStore.hasNext()) {
//                        backupStore.next();
//                        DataInputBuffer next = backupStore.nextValue();
//                        buffer.reset(next.getData(), next.getPosition(), next.getLength());
//                        value = valueDeserializer.deserialize(value);
//                        return value;
//                    } else {
//                        inReset = false;
//                        backupStore.exitResetMode();
//                        if (clearMarkFlag) {
//                            clearMarkFlag = false;
//                            isMarked = false;
//                        }
//                    }
//                } catch (IOException e) {
//                    e.printStackTrace();
//                    throw new RuntimeException("next value iterator failed", e);
//                }
//            }

            // if this is the first record, we don't need to advance
            if (firstValue) {
                firstValue = false;
                return value;
            }
            // if this isn't the first record and the next key is different, they
            // can't advance it here.
            if (!nextKeyIsSame) {
                throw new NoSuchElementException("iterate past last value");
            }
            // otherwise, go to the next key/value pair
            try {
                nextKeyValue();
                return value;
            } catch (IOException ie) {
                throw new RuntimeException("next value iterator failed", ie);
            } catch (InterruptedException ie) {
                // this is bad, but we can't modify the exception list of java.util
                throw new RuntimeException("next value iterator interrupted", ie);
            }
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("remove not implemented");
        }

        @Override
        public void mark() throws IOException {
//            if (backupStore == null) {
//                backupStore = new BackupStore<KEYIN, VALUEIN>(conf, taskid);
//            }
//            isMarked = true;
//            if (!inReset) {
//                backupStore.reinitialize();
//                if (currentKeyLength == -1) {
//                    // The user has not called next() for this iterator yet, so
//                    // there is no current record to mark and copy to backup store.
//                    return;
//                }
//                assert (currentValueLength != -1);
//                int requestedSize = currentKeyLength + currentValueLength
//                        + WritableUtils.getVIntSize(currentKeyLength)
//                        + WritableUtils.getVIntSize(currentValueLength);
//                DataOutputStream out = backupStore.getOutputStream(requestedSize);
//                writeFirstKeyValueBytes(out);
//                backupStore.updateCounters(requestedSize);
//            } else {
//                backupStore.mark();
//            }
        }

        @Override
        public void reset() throws IOException {
            // We reached the end of an iteration and user calls a
            // reset, but a clearMark was called before, just throw
            // an exception
//            if (clearMarkFlag) {
//                clearMarkFlag = false;
//                backupStore.clearMark();
//                throw new IOException("Reset called without a previous mark");
//            }
//
//            if (!isMarked) {
//                throw new IOException("Reset called without a previous mark");
//            }
//            inReset = true;
//            backupStore.reset();
        }

        @Override
        public void clearMark() throws IOException {
//            if (backupStore == null) {
//                return;
//            }
//            if (inReset) {
//                clearMarkFlag = true;
//                backupStore.clearMark();
//            } else {
//                inReset = isMarked = false;
//                backupStore.reinitialize();
//            }
        }

        /**
         * This method is called when the reducer moves from one key to
         * another.
         * @throws IOException
         */
        public void resetBackupStore() throws IOException {
//            if (backupStore == null) {
//                return;
//            }
//            inReset = isMarked = false;
//            backupStore.reinitialize();
//            currentKeyLength = -1;
        }

        /**
         * This method is called to write the record that was most recently
         * served (before a call to the mark). Since the framework reads one
         * record in advance, to get this record, we serialize the current key
         * and value
         * @param out
         * @throws IOException
         */
        private void writeFirstKeyValueBytes(DataOutputStream out)
                throws IOException {
//            assert (getCurrentKey() != null && getCurrentValue() != null);
//            WritableUtils.writeVInt(out, currentKeyLength);
//            WritableUtils.writeVInt(out, currentValueLength);
//            Serializer<KEYIN> keySerializer =
//                    serializationFactory.getSerializer(keyClass);
//            keySerializer.open(out);
//            keySerializer.serialize(getCurrentKey());
//
//            Serializer<VALUEIN> valueSerializer =
//                    serializationFactory.getSerializer(valueClass);
//            valueSerializer.open(out);
//            valueSerializer.serialize(getCurrentValue());
        }
    }

    protected class ValueIterable implements Iterable<VALUEIN> {

        private ValueIterator iterator = new ValueIterator();

        @Override
        public Iterator<VALUEIN> iterator() {
            return iterator;
        }
    }

    /**
     * Iterate through the values for the current key, reusing the same value
     * object, which is stored in the context.
     * @return the series of values associated with the current key. All of the
     * objects returned directly and indirectly from this method are reused.
     */
    public Iterable<VALUEIN> getValues() throws IOException, InterruptedException {
        return iterable;
    }

    private class KeyValueInput {

        private KEYIN key;
        private VALUEIN value;

        public boolean next() throws IOException {
            if (input.available() > 0) {
                //Must pass null to instantiate a new object
                key = keyDeserializer.deserialize(null);
                value = valueDeserializer.deserialize(null);
                return true;
            }
            return false;
        }

        public KEYIN getCurrentkey() {
            return key;
        }

        public VALUEIN getCurrentValue() {
            return value;
        }
    }
}
