/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.mapreduceModule.input;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.JobContext;
import p2pMapReduce.mapreduceModule.MRJobConfig;
import p2pMapReduce.mapreduceModule.Path;
import p2pMapReduce.mapreduceModule.TaskAttemptContext;
import p2pMapReduce.mapreduceModule.conf.Configuration;

/**
 * Read input text files and generate the {@link  InputSplit}
 * Each InputSplit is a {@link TextFileInputSplit} and always terminate with a whole
 * text line.
 * Provide also a LineRecordReader
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class TextFileInputFormat extends InputFormat<Long, String> {

    private static final transient Logger LOG = Logger.getLogger(TextFileInputFormat.class.getName());
    //TODO per ora path locali assoluti
    public static final String LOCAL_INPUT_PATHS_ATTR = "mapreduce.local.input.input_paths";
    @Deprecated
    public static final String INPUT_PATHS_ATTR = "mapreduce.input.input_paths";
    public static final String INPUTSPLIT_SUBDIR_ATTR = "mapreduce.job.input.inputsplit.subdir";
    public static final String INPUT_PATHS_SEPARATOR = ",";
    /**
     * The size of a single split in MB
     * 
     */
    private static final int DEFAULT_SPLIT_SIZE = 1;
    private int nextSplit;

    public static void main(String[] args) throws IOException {
        aggregate("/home/francesco/Scrivania/test/input/inputSplits/", new File("/home/francesco/Scrivania/test/input/aggregate.txt"));
    }

    /**
     * Add a path to an input of the job.
     * The path refers to local File System
     *
     * @param job
     * @param path
     */
    public static void addLocalInputPath(Job job, String path) {
        Configuration conf = job.getConfiguration();
//        String dirStr = StringUtils.escapeString(path.toString());
        String dirs = conf.get(LOCAL_INPUT_PATHS_ATTR);
        conf.set(LOCAL_INPUT_PATHS_ATTR, dirs == null ? path : dirs + INPUT_PATHS_SEPARATOR + path);
    }

    public static String getLocalInputPath(Job job) {
        Configuration conf = job.getConfiguration();
//        String dirStr = StringUtils.escapeString(path.toString());
        String dirs = conf.get(LOCAL_INPUT_PATHS_ATTR);
        return dirs;
    }

    /**
     * Add a {@link Path} to the list of inputs for the map-reduce job.
     *
     * @param job The {@link Job} to modify
     * @param path {@link Path} to be added to the list of inputs for
     *            the map-reduce job.
     * @deprecated
     */
    public static void addInputPath_old(Job job, String path) {
        Configuration conf = job.getConfiguration();
//        String dirStr = StringUtils.escapeString(path.toString());
        String dirs = conf.get(INPUT_PATHS_ATTR);
        conf.set(INPUT_PATHS_ATTR, dirs == null ? path : dirs + INPUT_PATHS_SEPARATOR + path);
    }

    public static void addInputPath(Job job, String path) {
        Configuration conf = job.getConfiguration();
//        String dirStr = StringUtils.escapeString(path.toString());
        conf.set(MRJobConfig.JOB_DIR_ATTR, path);
    }

    private static void aggregate(String path, File out) throws IOException {
        File dir = new File(path);
        LinkedList<String> names = new LinkedList<String>();
        for (File file : dir.listFiles()) {
            names.add(file.getName());
        }
        Collections.sort(names);
        LinkedList<File> orderedSplit = new LinkedList<File>();

        for (String fileName : names) {
            for (File f : dir.listFiles()) {
                if (f.getName().equals(fileName)) {
                    orderedSplit.add(f);
                }
            }
        }

        System.out.println("orderedSplit = " + orderedSplit);

        FileOutputStream fos = new FileOutputStream(out);
        for (File file : orderedSplit) {
            byte[] buf = new byte[1024];
            FileInputStream fis = new FileInputStream(file);
            while (fis.available() > 0) {
                int read = fis.read(buf);
                fos.write(buf, 0, read);
            }
            fis.close();
        }
        fos.close();

    }

    @Override
    public RecordReader<Long, String> createRecordReader(InputSplit split, TaskAttemptContext context) throws IOException, InterruptedException {
        return new LineRecordReader();
    }

    public List<InputSplit> getSplits(JobContext jobContext) throws IOException, InterruptedException {
        LinkedList<InputSplit> inputSplits = new LinkedList<InputSplit>();

        LinkedList<File> inputFiles = getInputs(jobContext);
        LOG.log(Level.FINE, "inputFiles:\n" + inputFiles);
        String workingDir = jobContext.getWorkingDirectory().toString();
        String jobInputSplitsDir = workingDir + "/" + jobContext.getConfiguration().get(MRJobConfig.JOB_INPUTSPLIT_DIR_ATTR);
        TextFileInputSplit currentFileSplit = null;
        int splitSizeInBytes = 1024 * 1024 * jobContext.getConfiguration().getInt(MRJobConfig.INPUTSPLIT_SIZE, DEFAULT_SPLIT_SIZE);
        LOG.log(Level.FINE, "Split size is :" + splitSizeInBytes);
        for (File file : inputFiles) {
            currentFileSplit = nextSplit(jobInputSplitsDir, file);

            long bytesRemaining = splitSizeInBytes - currentFileSplit.getLength();
            long fileSize = file.length();
            long bytesWritten = 0;
            BufferedReader br = new BufferedReader(new FileReader(file));
            while (true) {
                long written = currentFileSplit.append(br, bytesRemaining);
                bytesWritten += written;
                bytesRemaining = splitSizeInBytes - currentFileSplit.getLength();
                if (bytesRemaining <= 0) {
                    currentFileSplit.close();
                    inputSplits.add(currentFileSplit);
                    currentFileSplit = nextSplit(jobInputSplitsDir, file);
                    bytesRemaining = splitSizeInBytes;
                }
                if (written == 0) {
                    break;
                }
            }
            br.close();
            currentFileSplit.close();
            inputSplits.add(currentFileSplit);
            LOG.log(Level.FINE, "File size: {0} bytes written: {1} diff= {2}", new Object[]{fileSize, bytesWritten, fileSize - bytesWritten});
        }

        if (!inputSplits.getLast().equals(currentFileSplit) && currentFileSplit.getLength() > 0) {
            inputSplits.add(currentFileSplit);
        }

        return inputSplits;

    }

    /**
     * Returns the list of the all file retrievable from the input path
     * specified in the JobContext
     *
     * @param jobContext
     * @return
     */
    private LinkedList<File> getInputs(JobContext jobContext) {
        try {
            String workingDir = jobContext.getWorkingDirectory().toString();
            String inputPathsString = workingDir + "/" + jobContext.getConfiguration().get(MRJobConfig.JOB_INPUT_DIR_ATTR);
            LOG.log(Level.CONFIG, "InputPathsStrings is: " + inputPathsString);
            LinkedList<File> ris = new LinkedList<File>();
            list(new File(inputPathsString), ris);
            //        String[] inputPaths = inputPathsString.split(INPUT_PATHS_SEPARATOR);
            //        LinkedList<File> ris = new LinkedList<File>();
            //
            //        for (String inputPath : inputPaths) {
            //            File f = new File(inputPath);
            //            list(f, ris);
            //        }
            return ris;
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, null, ex);
        }
        return null;
    }

    /**

     * If the file is a dir recursively add all the file contained in it and
     * in its subdirectory.
     * Add the file directly instead
     * @param src
     * @param fileList
     */
    private void list(File src, LinkedList<File> fileList) {
        if (src.isDirectory()) {
            for (File file : src.listFiles()) {
                list(file, fileList);
            }
        } else {
            fileList.add(src);
        }
    }

    private TextFileInputSplit nextSplit(String inputSplitsDir, File file) throws FileNotFoundException, IOException {
        String filePath = inputSplitsDir + nextSplit + "_split.txt";
        nextSplit++;
        return new TextFileInputSplit(filePath, file.getAbsolutePath());
    }
}
