/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 *  This code has been partially modified as part of P2P-MapReduce.
 *
 *  Copyright (C) 2011 University of Calabria
 *  Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.mapreduceModule.io.serializer;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import p2pMapReduce.mapreduceModule.conf.Configuration;
import p2pMapReduce.mapreduceModule.util.ReflectionUtils;
import p2pMapReduce.mapreduceModule.util.StringUtils;
import p2pMapReduce.mapreduceModule.conf.Configured;

/**
 * <p>
 * A factory for {@link Serialization}s.
 * </p>
 */
public class SerializationFactory extends Configured {

    private static final String SERIALIZATIONS_ATTR = "io.serializations";
    private static final transient Logger LOG = Logger.getLogger(SerializationFactory.class.getName());
    private List<Serialization<?>> serializations = new ArrayList<Serialization<?>>();

    /**
     * <p>
     * Serializations are found by reading the <code>io.serializations</code>
     * property from <code>conf</code>, which is a comma-delimited list of
     * classnames.
     * </p>
     */
    public SerializationFactory(Configuration conf) {
        super(conf);
        for (String serializerName : conf.getStrings(SERIALIZATIONS_ATTR, new String[]{WritableSerialization.class.getName()})) {
            add(conf, serializerName);
        }
    }

    @SuppressWarnings("unchecked")
    private void add(Configuration conf, String serializationName) {
        try {
            Class<? extends Serialization> serializionClass =
                    (Class<? extends Serialization>) conf.getClassByName(serializationName);
            serializations.add((Serialization) ReflectionUtils.newInstance(serializionClass, getConf()));
        } catch (ClassNotFoundException e) {
            LOG.warning("Serialization class not found: "
                    + StringUtils.stringifyException(e));
        }
    }

    public <T> Serializer<T> getSerializer(Class<T> c) {
        return getSerialization(c).getSerializer(c);
    }

    public <T> Deserializer<T> getDeserializer(Class<T> c) {
        return getSerialization(c).getDeserializer(c);
    }

    @SuppressWarnings("unchecked")
    public <T> Serialization<T> getSerialization(Class<T> c) {
        for (Serialization serialization : serializations) {
            if (serialization.accept(c)) {
                return (Serialization<T>) serialization;
            }
        }
        return null;
    }
}
