/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.mapreduceModule.output;

import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;
import p2pMapReduce.mapreduceModule.JobContext;
import p2pMapReduce.mapreduceModule.MRJobConfig;
import p2pMapReduce.mapreduceModule.Partitioner;
import p2pMapReduce.mapreduceModule.util.ReflectionUtils;
import p2pMapReduce.mapreduceModule.TaskAttemptContext;
import p2pMapReduce.mapreduceModule.TaskAttemptID;
import p2pMapReduce.mapreduceModule.conf.Configuration;
import p2pMapReduce.mapreduceModule.io.serializer.SerializationFactory;
import p2pMapReduce.mapreduceModule.io.serializer.Serializer;

/**
 * Writes map temp output
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class MapRecordWriter<K, V> extends RecordWriter<K, V> {

    private Partitioner<K, V> partitioner;
    private Serializer<K> keySerializer;
    private Serializer<V> valueSerializer;
    private HashMap<Integer, PartitionFile> paritionFilesMap;
    /**
     * Represent a temp map output dir
     */
    private String partitionDir;
    private int numPartitions;
    private int mapTaskNumber;

    public MapRecordWriter(TaskAttemptID taskAttID, JobContext jobContext, Class<K> keyClass, Class<V> valueClass) throws ClassNotFoundException, IOException {
        partitioner = (Partitioner<K, V>) ReflectionUtils.newInstance(jobContext.getPartitionerClass(), jobContext.getConfiguration());
        Configuration conf = jobContext.getConfiguration();
        SerializationFactory sf = new SerializationFactory(conf);
        keySerializer = sf.getSerializer(keyClass);
        valueSerializer = sf.getSerializer(valueClass);
        paritionFilesMap = new HashMap<Integer, PartitionFile>();
        //TODO uguagliare schema di costruzione path
        // / prima o dopo?
        partitionDir = jobContext.getWorkingDirectory() + "/" + conf.get(MRJobConfig.TASK_TEMP_DIR) + "/map/" + taskAttID.toString() + "/";
//        System.out.println("partitionDir = " + partitionDir);
        numPartitions = jobContext.getNumReduceTasks();
//        System.out.println("numPartitions = " + numPartitions);
        mapTaskNumber = taskAttID.getTaskID().getId();
    }

    @Override
    public void write(K key, V value) throws IOException, InterruptedException {
        int partition = partitioner.getPartition(key, value, numPartitions);
        PartitionFile pf = getPartitionFile(partition);
        DataOutputStream dos = pf.getDataOutputStream();
        keySerializer.open(dos);
        keySerializer.serialize(key);
        valueSerializer.open(dos);
        valueSerializer.serialize(value);
    }

    @Override
    public void close(TaskAttemptContext context) throws IOException, InterruptedException {
        for (PartitionFile partitionFile : paritionFilesMap.values()) {
            partitionFile.close();
        }
    }

    private PartitionFile getPartitionFile(int partition) throws FileNotFoundException, IOException {
        PartitionFile pf = paritionFilesMap.get(partition);
        if (pf == null) {
            pf = createPartitionFile(partition);
            paritionFilesMap.put(partition, pf);
        }
        return pf;
    }

    private PartitionFile createPartitionFile(int partition) throws FileNotFoundException, IOException {
        String path = partitionDir + "partition_" + partition + "_" + mapTaskNumber+".tmp";
        return new PartitionFile(path);
    }
}
