/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.networkModule;

import java.io.IOException;
import java.io.Serializable;
import java.util.Set;
import p2pMapReduce.ftp.FTPManager;
import p2pMapReduce.nodeModule.event.EventListener;
import p2pMapReduce.nodeModule.NodeModuleIF;
import p2pMapReduce.nodeModule.event.MessageEvent;

/**
 * A NetworkModuleIF is the manager of all the network interaction of the node
 * in the P2PMapReduce framework.
 * It gives a way to:
 * <br>
 *  - join the P2PMapReduce network
 * <br>
 *  - search the network for some nodes
 * <br>
 *  - obtain connections between nodes
 * <br>
 *  - obtain an {@link FTPManager}
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public interface NetworkModuleIF {

    /**
     * Call this method once to enter the P2PMapReduce network.
     * The method returns an instance of NodeInfo containing all the information
     * about the node associated to this network module.
     * A reference to this NodeInfo instance will be maintained from this module
     * and can be retrieved in the future by calling {@link #getNodeInfo()}.
     * All the updates to the node information will affect this NodeInfo instance.
     *
     * It also:
     *  - publishes this NodeInfo by calling publish()
     *  - make this NetworkModule capable to receive remote connection from other nodes.
     *  - starts an FTP server
     * 
     * @return
     */
    NodeInfo joinNetwork();

    /**
     * Return the same instance of NodeInfo returned by the joinNetwork() method.
     * @return the same instance of NodeInfo returned by joinNetwork() or null if
     * joinNetwork() has never be executed with success.
     */
    NodeInfo getNodeInfo();

    /**
     * Return an FTPManager for this node
     * @return
     */
    FTPManager getFtpManager();

    /**
     * Used to send the information of the NodeInfo instance associated
     * to this network module across the P2PMapReduce network.
     * It must be called after every update to the NodeInfo of this network module
     * in order to make all node known about .
     *
     */
    void publish();

    /**
     * Invoked by the node module to notify that the node has changed its role
     * from Master to Slave
     */
    void becomeSlave();

    /**
     * Invoked by the node module to notify that the node has changed its role
     * from Slave to Master
     */
    void becomeMaster();

    /**
     * Returns the number of all the visible master nodes from this network module
     * at the moment
     * @return
     */
    int getNumMasters();

    /**
     * Returns the number of all the visible slave nodes from this network module
     * at the moment
     * @return
     */
    int getNumSlaves();

    /**
     * Returns the number of all the visible nodes from this network module
     * at the moment
     * @return
     */
    int getNumNodes();

    /**
     * Perform a search in the network for the NodeInfo of the node identified
     * by {@code nodeID}
     * 
     * @param nodeID
     * @return
     */
    NodeInfo find(String nodeID);

    /**
     * Indicate if the node identified by 'nodeID' is alive.
     *
     * @param nodeID
     * @return true if a NodeInfo is founded by calling {@link #find(java.lang.String)}
     * @see #find(String)
     */
    boolean isAlive(String nodeID);

    /**
     * Return all the visible nodes in the network by this network module
     * at the moment
     * @return
     */
    Set<NodeInfo> getAllNodes();

    /**
     * Return all the visible slave nodes in the network by this network module
     * at the moment
     * @return
     */
    Set<NodeInfo> getAllSlaves();

    /**
     * Return all the visible master nodes in the network by this network module
     * at the moment
     * @return
     */
    Set<NodeInfo> getAllMasters();

    /**
     * Return all the visible slaves nodes in the network by this network module
     * at the moment, except this node if it is a slave
     * @return
     */
    Set<NodeInfo> getOtherSlaves();

    /**
     * Return all the visible slaves nodes in the network by this network module
     * at the moment, except this node if it is a master
     * @return
     */
    Set<NodeInfo> getOtherMasters();

    /**
     * Send the object {@code msg} to the node specified by {@code endPoint}
     * If no connection is already available to this node, try to create a new
     * one using the provided NodeInfo
     * 
     * @param endPoint
     * @param msg
     * @throws IOException
     */
    void send(NodeInfo endPoint, Serializable msg) throws IOException;

    /**
     * Send the object {@code msg} to the node identified by {@code destID}.
     * If no connection is already available to this node, try to retrieve the
     * NodeInfo for this node and to create a new connection.
     * 
     * @param destID
     * @param msg
     * @throws IOException
     */
    void send(String destID, Serializable msg) throws IOException;

    /**
     * Returns a connection between this network module and the network module
     * relative to the node described by the NodeInfo {@code connectionEndPoint}
     *
     * @param connectionEndPoint
     * @return
     * @throws IOException
     */
    ConnectionIF getConnection(NodeInfo connectionEndPoint) throws IOException;

    /**
     * Returns a connection between this network module and the network module
     * relative to the node identified by {@code nodeID}.
     * If no connection is already available to this node try to retrieve the
     * NodeInfo for this node and to create a new connection.
     * IOException is thrown instead or in consequence of errors.
     *
     * @param connectionEndPoint
     * @return
     * @throws IOException
     */
    ConnectionIF getConnection(String nodeID) throws IOException;

    /**
     * Returns the {@link NodeModuleIF} associated to this network module
     * @return
     */
    NodeModuleIF getNodeModule();

    void setNodeModule(NodeModuleIF nodeModule);

    /**
     * Used to add a {@link EventListener} to which forward all the
     * {@link MessageEvent} received by the network connections
     * 
     * @param eventListener
     */
    void addEventListener(EventListener eventListener);
}
