/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule;

import p2pMapReduce.networkModule.NodeInfo;
import java.io.IOException;
import java.util.LinkedList;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.fs.FileSystemView;
import p2pMapReduce.ftp.FTPManager;
import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.MRJobConfig;
import p2pMapReduce.mapreduceModule.Path;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.entity.BackupJob;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.InnerEvent;
import p2pMapReduce.nodeModule.event.MessageEvent;
import p2pMapReduce.nodeModule.nodeFsm.CoordinationFSM;
import p2pMapReduce.nodeModule.nodeFsm.FSM;
import p2pMapReduce.nodeModule.nodeFsm.NodeFSM;
import p2pMapReduce.nodeModule.nodeFsm.NodeState;

/**
 * Acts as a backup of a job.
 * Receives the job update from the primary and updates the backup state.
 * When the primary seems failed starts the election algorithm with others
 * backup job manager.
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class BackupJobManager extends FSM implements Runnable {

    private static final transient Logger LOG = Logger.getLogger(BackupJobManager.class.getName());
    private String LOG_PREFIX;
    private static final int PRIMARY_ELECTION_TIMEOUT = 30 * 1000;
    private final BackupJobManagerState CHECK_PRIMARY;
    private final BackupJobManagerState ELECTING_PRIMARY;
    private final BackupJobManagerState WAITING_PRIMARY;
    private final BackupJobManagerState NEW_PRIMARY;
    private final BackupJob managedBackupJob;
    private NodeFSM nodeFSM;
    private Thread executorThread;
    /**
     * The primary update message to use during the election algorithm
     */
    private MessageEvent primaryUpdateMsg;
    /**
     * The election message to use during the election algorithm
     */
    private MessageEvent electionMsg;
    Timer timer;
    TimeoutTask timeoutTask = new TimeoutTask();

    public BackupJobManager(NodeFSM nodeFSM, BackupJob managedBackupJob) {

        this.nodeFSM = nodeFSM;
        this.managedBackupJob = managedBackupJob;
        LOG_PREFIX = "BackupJobManager for " + managedBackupJob.getJobID() + ": ";
        this.timer = new Timer(LOG_PREFIX + "Timer", true);
        CHECK_PRIMARY = new CheckPrimary();
        ELECTING_PRIMARY = new ElectingPrimary();
        WAITING_PRIMARY = new WaitingPrimary();
        NEW_PRIMARY = new NewPrimary();
    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM methods
    ///////////////////////////////////////////////////////////////////////////
    public synchronized void primaryMasterFailure(String primaryID) {
        ((BackupJobManagerState) currentState()).primaryMasterFailure(primaryID);
    }

    public synchronized void primaryElectionMessage(String jobID, String senderID, double senderLoad) {
        ((BackupJobManagerState) currentState()).primaryElectionMessage(jobID, senderID, senderLoad);
    }

    public synchronized void primaryUpdate(String jobID, String updatedPrimaryID, double updatedPrimaryLoad) {
        ((BackupJobManagerState) currentState()).primaryUpdate(jobID, updatedPrimaryID, updatedPrimaryLoad);
    }

    private synchronized void timeout(Object param) {
        ((BackupJobManagerState) currentState()).timeout(param);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Public methods
    ///////////////////////////////////////////////////////////////////////////
    //FIXME come il run di JobManager dovrebbe essere sync?? forse si perchè
    //potrebbero arrivare messaggi di job update
    public void run() {
        setupEnv();
        retrieveInput();
        transition(CHECK_PRIMARY);
    }

    public void start() {
        executorThread = new Thread(this);
        executorThread.start();
    }

    public void stop() {
        LOG.log(Level.INFO, "BackupJobManager for job {0} stopped", managedBackupJob.getJobID());
    }

    public String getPrimaryID() {
        return managedBackupJob.getPrimaryID();
    }

    public void setPrimaryID(String primaryNodeID) {
        managedBackupJob.setPrimaryID(primaryNodeID);
    }

    /**
     * Compares this node data with the specified data.
     *
     * @param otherID
     * @return
     */
    private boolean loss(String otherID, double otherLoadValue) {
        NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();
        double nodeLoadValue = netModule.getNodeInfo().getLoad();
        String nodeID = netModule.getNodeInfo().getID();
        if (nodeLoadValue < otherLoadValue) {
            return false;
        } else if (nodeLoadValue == otherLoadValue) {
            return nodeID.compareTo(otherID) < 0 ? false : true;
        } else {
            return true;
        }
    }

    private void setupEnv() {
        LOG.log(Level.INFO, LOG_PREFIX + "Setup environment");

        FileSystemView fsV = nodeFSM.getNodeModule().getFileSystemView();
        Job job = managedBackupJob.getJob();
        job.setWorkingDirectory(new Path(fsV.getRootDir()));
        String jobDir = managedBackupJob.getJobID().toString();
        fsV.mkdir(jobDir);
    }

    private void retrieveInput() {
        LOG.log(Level.INFO, LOG_PREFIX + "Retrieving input");
        NetworkModuleIF netModule = nodeFSM.getNetworkModule();
        //XXX the input must be retrieved from user node instead from primary
        //The user node is considered reliable and have no download request
        //from the slave node like the primary
        NodeInfo userNode = managedBackupJob.getUserNodeInfo();
        Job job = managedBackupJob.getJob();

        FTPManager ftpM = nodeFSM.getNodeModule().getFtpManager();

        String inputDirPath = job.getConfiguration().get(MRJobConfig.JOB_INPUT_DIR_ATTR);
        boolean success = ftpM.downloadDir(userNode, inputDirPath);
        if (!success) {
            throw new RuntimeException("Unable to retrive input file from " + userNode.getIpAddress() + " at " + userNode.getFtpPort() + " on " + managedBackupJob.getPrimaryID());
        }

        String jarsDirPath = job.getConfiguration().get(MRJobConfig.JOB_JAR_DIR_ATTR);
        success = ftpM.downloadDir(userNode, jarsDirPath);
        if (!success) {
            throw new RuntimeException("Unable to retrive input jar file from " + userNode.getIpAddress() + " at " + userNode.getFtpPort() + " on " + managedBackupJob.getPrimaryID());
        }
        LOG.log(Level.INFO, "{0} input retrieved", new Object[]{LOG_PREFIX});
    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM state classes
    ///////////////////////////////////////////////////////////////////////////
    private class BackupJobManagerState extends NodeState {
    }

    private class CheckPrimary extends BackupJobManagerState {

        @Override
        public void primaryMasterFailure(String primaryID) {
            LOG.log(Level.INFO, "Primary " + primaryID + " failed.Transition to BACKUP_JOB_MANAGER.ELECTING_PRIMARY");
            transition(ELECTING_PRIMARY);
        }

        @Override
        public void primaryElectionMessage(String jobID, String senderID, double senderLoad) {

            boolean loss = loss(senderID, senderLoad);
            if (loss) {
                //An election message is received from another master with
                //an higher priority
                timeoutTask.cancel();
                LOG.log(Level.INFO, "Received a response from higher priority (" + senderID + ").Transition to WAITING_COORDINATOR");
                transition(WAITING_PRIMARY);
            } else {
                //An election message is received from another master with
                //a lower priority
                LOG.log(Level.INFO, "Received a response from lower priority (" + senderID + "). Sending response.");
                NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();
                try {
                    netModule.send(senderID, electionMsg);
                } catch (IOException ex) {
                    ex.printStackTrace();
                    LOG.log(Level.SEVERE, null, ex);
                }
                transition(ELECTING_PRIMARY);

            }
        }

        @Override
        public void primaryUpdate(String jobID, String updatedPrimaryID, double updatedPrimaryLoad) {
            LOG.log(Level.INFO, "Received new primary id: " + updatedPrimaryID + ".Transition to BACKUP_JOB_MANAGER.CHECK_PRIMARY");
            managedBackupJob.setPrimaryID(updatedPrimaryID);
        }
    }

    /**
     * The election algorithm is executed between only the master nodes that are
     * backup node for the job.
     * Every node sends to others an election message containing its id and load
     * value.
     * The node which lower load value become the new primary. On same load value
     * the node with higher id become the new primary.
     *
     */
    private class ElectingPrimary extends BackupJobManagerState {

        @Override
        protected void entryAction() {
            //Send election message to all master with higher identifier
            LOG.log(Level.INFO, "Entering BACKUP_JOB_MANAGER.ELECTING_PRIMARY: sending election messages to higher masters");
            NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();
            NodeInfo nodeInfo = netModule.getNodeInfo();
            LinkedList<String> backupMasters = managedBackupJob.getBackupMasterList();
            LOG.log(Level.INFO, "Available backupMasters for primary election algorithm: " + backupMasters.size());
            primaryUpdateMsg = new MessageEvent(EventType.PRIMARY_UPDATE, managedBackupJob.getJobID(), nodeInfo.getID(), nodeInfo.getLoad());
            electionMsg = new MessageEvent(EventType.PRIMARY_ELECTION_MESSAGE, managedBackupJob.getJobID(), nodeInfo.getID(), nodeInfo.getLoad());
            for (String backupMaster : backupMasters) {
                if (!backupMaster.equals(nodeInfo.getID())) {
                    //To all BackupMaster except this
                    try {
                        LOG.log(Level.INFO, "Sending to " + backupMaster);
                        netModule.send(backupMaster, electionMsg);
                        LOG.log(Level.INFO, "Sent");
                    } catch (IOException ex) {
                        ex.printStackTrace();
                        LOG.log(Level.SEVERE, null, ex);
                    }
                }
            }

            timeoutTask = new TimeoutTask();
            timer.schedule(timeoutTask, PRIMARY_ELECTION_TIMEOUT);
            LOG.log(Level.INFO, "Messages sent. Waitinig for " + PRIMARY_ELECTION_TIMEOUT + " secs");

        }

        @Override
        public void primaryElectionMessage(String jobID, String senderID, double senderLoad) {

            boolean loss = loss(senderID, senderLoad);
            if (loss) {
                //An election message is received from another master with
                //an higher priority
                timeoutTask.cancel();
                LOG.log(Level.INFO, "Received a response from higher priority (" + senderID + ").Transition to WAITING_COORDINATOR");
                transition(WAITING_PRIMARY);
            } else {
                //An election message is received from another master with
                //a lower priority
                LOG.log(Level.INFO, "Received a response from lower priority (" + senderID + "). Sending response.");
                NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();
                try {
                    netModule.send(senderID, electionMsg);
                } catch (IOException ex) {
                    ex.printStackTrace();
                    LOG.log(Level.SEVERE, null, ex);
                }

            }
        }

        @Override
        public void primaryUpdate(String jobID, String updatedPrimaryID, double updatedPrimaryLoad) {
            timeoutTask.cancel();
            LOG.log(Level.INFO, "Received new primary id: " + updatedPrimaryID + ".Transition to BACKUP_JOB_MANAGER.CHECK_PRIMARY");
            managedBackupJob.setPrimaryID(updatedPrimaryID);
            transition(CHECK_PRIMARY);
        }

        @Override
        public void timeout(Object param) {
            //NO respose are received.
            //Becoming new primary
            LOG.log(Level.INFO, "Becoming the new primary");
            transition(NEW_PRIMARY);
        }
    }

    private class WaitingPrimary extends BackupJobManagerState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, "Entering BACKUP_JOB_MANAGER.WAITING_PRIMARY");
            timeoutTask = new TimeoutTask();
            timer.schedule(timeoutTask, PRIMARY_ELECTION_TIMEOUT);
        }

        //FIXME manage primary election??

        @Override
        public void primaryUpdate(String jobID, String updatedPrimaryID, double updatedPrimaryLoad) {
            timeoutTask.cancel();
            LOG.log(Level.INFO, "Primary update for " + jobID + " with new primary " + updatedPrimaryID + ".Transition to BACKUP_JOB_MANAGER.CHECK_PRIMARY");
            managedBackupJob.setPrimaryID(updatedPrimaryID);
            transition(CHECK_PRIMARY);
        }

        @Override
        public void timeout(Object param) {
            LOG.log(Level.INFO, "Primary update not received in time.Transition to ELECTING_PRIMARY");
            transition(ELECTING_PRIMARY);
        }
    }

    private class NewPrimary extends BackupJobManagerState {

        @Override
        protected void entryAction() {

            String nodeID = nodeFSM.getNodeModule().getNodeID();
            //Removing this node as a backup node
            managedBackupJob.getBackupMasterList().remove(nodeID);
            NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();

            //notify 'other' backup master that this node is the new primary
            LOG.log(Level.INFO, "Sending PrimaryUpdate message to BackupMaster nodes");
            for (String backupMaster : managedBackupJob.getBackupMasterList()) {
                try {
                    netModule.send(backupMaster, primaryUpdateMsg);
                } catch (IOException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
            //XXX This notification will be done by jobReassigned in Management
//            //notify the interested slaves that this is the new primary
//            for (JobTask jobTask : managedBackupJob.getJobTaskList()) {
//                try {
//                    netModule.send(jobTask.getSlaveID(), primaryUpdateMsg);
//                } catch (IOException ex) {
//                    Logger.getLogger(BackupJobManager.class.getName()).log(Level.SEVERE, null, ex);
//                }
//            }

            InnerEvent backupJobCompletedMsg = new InnerEvent(EventType.BACKUP_JOB_COMPLETED, managedBackupJob.getJobID());
            //send to NODE.MASTER.RECOVERY
            nodeFSM.handleEvent(backupJobCompletedMsg);

            InnerEvent jobReassMsg = new InnerEvent(EventType.JOB_REASSIGNED, managedBackupJob);
            //send to to NODE.MASTER.MANAGEMENT
            nodeFSM.handleEvent(jobReassMsg);

        }

        @Override
        public void primaryElectionMessage(String jobID, String senderID, double senderLoadValue) {
            try {
                //Election message received while i'm already the new primary
                //Send direcly primary update msg
                LOG.log(Level.INFO, "Received an election message from: " + senderID + ". Sending primary updated msg");
                nodeFSM.getNodeModule().getNetworkModule().send(senderID, primaryUpdateMsg);
            } catch (IOException ex) {
                ex.printStackTrace();
                Logger.getLogger(CoordinationFSM.class.getName()).log(Level.SEVERE, null, ex);
            }


        }

        @Override
        public void primaryUpdate(String jobID, String updatedPrimaryID, double updatedPrimaryLoad) {
            //New coordinator id received while i'm already the new primary
            //To assure algorithm convergence leave the higher priority backupMaster to continue as new primary
            boolean loss = loss(updatedPrimaryID, updatedPrimaryLoad);
            if (loss) {
                timeoutTask.cancel();
                LOG.log(Level.INFO, "Received new higher priority primary id: " + updatedPrimaryID + ".Transition to BACKUP_JOB_MANAGER.CHECK_PRIMARY");
                //TODO controllare: il nuovo primary avrà inviato a tutti il mess di primay update?
                managedBackupJob.setPrimaryID(updatedPrimaryID);
                transition(CHECK_PRIMARY);
            } else {
                LOG.log(Level.INFO, "Received new lower priority primary id: " + updatedPrimaryID + ".Sending primary update message.");
                try {
                    nodeFSM.getNodeModule().getNetworkModule().send(updatedPrimaryID, primaryUpdateMsg);
                } catch (IOException ex) {
                    ex.printStackTrace();
                    Logger.getLogger(CoordinationFSM.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
    }

    private class TimeoutTask extends TimerTask {

        @Override
        public void run() {
            timeout(null);
        }
    }
}
