/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule;

import p2pMapReduce.networkModule.NodeInfo;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.PriorityQueue;
import java.util.Properties;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.fs.FileSystemView;
import p2pMapReduce.ftp.FTPManager;
import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.JobStatus;
import p2pMapReduce.mapreduceModule.MRJobConfig;
import p2pMapReduce.mapreduceModule.Path;
import p2pMapReduce.mapreduceModule.util.ReflectionUtils;
import p2pMapReduce.mapreduceModule.TaskAttemptID;
import p2pMapReduce.mapreduceModule.TaskID;
import p2pMapReduce.mapreduceModule.TaskType;
import p2pMapReduce.mapreduceModule.conf.Configuration;
import p2pMapReduce.mapreduceModule.input.InputFormat;
import p2pMapReduce.mapreduceModule.input.InputSplit;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.entity.JobTask;
import p2pMapReduce.entity.MapTask;
import p2pMapReduce.entity.PrimaryJob;
import p2pMapReduce.entity.ReduceTask;
import p2pMapReduce.entity.Task;
import p2pMapReduce.entity.Task.TaskStatus;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.InnerEvent;
import p2pMapReduce.nodeModule.event.MessageEvent;
import p2pMapReduce.nodeModule.nodeFsm.FSM;
import p2pMapReduce.nodeModule.nodeFsm.NodeFSM;
import p2pMapReduce.nodeModule.nodeFsm.NodeState;
import p2pMapReduce.util.P2PMapReduceConfig;
import p2pMapReduce.util.LoggingUtil;

/**
 * The manager of a job.
 * It takes care of:
 * <br>
 * - creation of input splits
 * - map and reduce tasks scheduling
 * - coordination of map and reduce phase
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
//TODO encapsulate the state of job manager in an ad hoc data structure
public class JobManager extends FSM implements Runnable {

    private static final transient Logger LOG = Logger.getLogger(JobManager.class.getName());
    private String LOG_PREFIX;
    /**
     * Key for {@link #REQUIRED_BACKUP_MANAGER}
     */
    private static final String REQUIRED_BACKUP_MANAGER_ATTR = "job.backupmanager.required";
    /**
     * The numbers of backup manager to associate to each job.
     * All positive values are permitted, 0 included.
     * The special value '-1' indicates that all the available masters must
     * be used as backup manager
     */
    private static final int REQUIRED_BACKUP_MANAGER;
    /**
     * Time period to check the job status and take care of task failed
     */
    public static final int MONITORING_TIMEOUT = (int) (2 * 60 * 1000);

    static {
        Properties p = P2PMapReduceConfig.getConfiguration();
        REQUIRED_BACKUP_MANAGER = Integer.parseInt(p.getProperty(REQUIRED_BACKUP_MANAGER_ATTR, "2"));
    }
    private final JobManagerState ASSIGNMENT;
    private final JobManagerState MONITORING;
    /**
     * The node fsm associate to this node
     */
    private NodeFSM nodeFSM;
    /**
     * The primary job managed by this JobManager
     */
    private PrimaryJob managedPrimaryJob;
    /**
     * Thread in which this job manager is executed
     */
    private Thread executorThread;
    /**
     * 
     * For each node associates the set of the job tasks which have been
     * submitted to the node.
     * When a new slave is selected the task set is empty.
     *
     * TODO evaluate if maintain or delete completed tasks
     */
    private Map<NodeInfo, Set<JobTask>> currentTaskAssignment;
    /**
     * All the map tasks attempt submitted to some node and considered still alive.
     */
    private Map<TaskID, TreeSet<JobTask>> submittedMapTaskAttempt;
    /**
     * All the map tasks attempt submitted to some node and considered failed.
     */
    private Map<TaskID, TreeSet<JobTask>> failedMapTaskAttempt;
    /**
     * All the map tasks attempt submitted to some node and considered successfully
     * completed.
     * XXX using HashMap for serialization
     */
    private HashMap<TaskID, TreeSet<JobTask>> successfullMapTaskAttempt;
    /**
     * All the reduce tasks attempt submitted to some node and considered still alive.
     */
    private Map<TaskID, TreeSet<JobTask>> submittedReduceTaskAttempt;
    /**
     * All the reduce tasks attempt submitted to some node and considered failed.
     */
    private Map<TaskID, TreeSet<JobTask>> failedReduceTaskAttempt;
    /**
     * All the reduce tasks attempt submitted to some node and considered successfully
     * completed.
     */
    private Map<TaskID, TreeSet<JobTask>> successfullReduceTaskAttempt;
    /**
     * A scheduler to select the next tasks attempt to assign and the node to which
     * this tasks must be assigned.
     * 
     */
    private TaskScheduler taskScheduler;
    /**
     * Task from which a check message is expected
     * in the current monitoring period
     */
    private Set<JobTask> taskToCheck;
    /**
     * A black list of those nodes that seems failed.
     * A node is put in the black list if a failure is notified to this JobManager.
     * The selection of a node (backup or slave) to which submit a task takes
     * care that node contained in the black list cannot be selected.
     */
    private Set<String> blackList;
    /**
     * Id of the reduce task for which the user has still not retrieved
     * output
     */
    private Set<TaskID> reducerOutputToRetrieve;
    /**
     * Indicate if this JobManager has been started in consequence of a previous
     * failed primary.
     * 
     */
    private boolean isReassigned;
    Timer timer;
    TimeoutTask timeoutTask = new TimeoutTask();

    public JobManager(NodeFSM nodeFSM, PrimaryJob primaryJob, boolean isReassigned) {
        this.nodeFSM = nodeFSM;
        this.managedPrimaryJob = primaryJob;
        this.currentTaskAssignment = new HashMap<NodeInfo, Set<JobTask>>();
        this.taskToCheck = new HashSet<JobTask>();
        this.blackList = new HashSet<String>();
        this.isReassigned = isReassigned;
        LOG_PREFIX = "JobManager(" + primaryJob.getJobID() + ")-";
        this.timer = new Timer(LOG_PREFIX + " Timer", true);
        this.taskScheduler = new TaskScheduler();
        ASSIGNMENT = new Assignment();
        MONITORING = new Monitoring();

    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM methods
    ///////////////////////////////////////////////////////////////////////////
    private synchronized void timeout(Object param) {
        ((JobManagerState) currentState()).timeout(param);
    }

    public synchronized void backupMasterFailure(String backupID) {
        ((JobManagerState) currentState()).backupMasterFailure(backupID);
    }

    public synchronized void slaveCheck(String taskID) {
        ((JobManagerState) currentState()).slaveCheck(taskID);
    }

    public synchronized void slaveFailure(String failedSlaveID) {
        ((JobManagerState) currentState()).slaveFailure(failedSlaveID);
    }

    public synchronized void taskCompleted(String completedTaskID, Task.TaskStatus completedTaskStatus) {
        ((JobManagerState) currentState()).taskCompleted(completedTaskID, completedTaskStatus);
    }

    public synchronized void outputRetrieved(JobTask reduceJT, boolean completionState) {
        ((JobManagerState) currentState()).outputRetrieved(reduceJT, completionState);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Public methods 
    ///////////////////////////////////////////////////////////////////////////
    public void start() {
        LOG.log(Level.INFO, "{0} Starting JobManager for job {1}", new Object[]{LOG_PREFIX, managedPrimaryJob.getJobID()});
        executorThread = new Thread(this, LOG_PREFIX);
        /**
         *
         * XXX se si utilizza run() si blocca il nodo fino alla fine dell
         * downl degli input e alla creazione degli input split
         * che può essere un tempo troppo lungo
         * provo ad usare start.
         *
         * Se si usa start mentre il thread inizializza il job nel caso di un
         * reassigned job potrebbero arrivare i messaggi dagli slave che hanno
         * ricevuto il primary update e trovare lo stato corrente a null
         * poichè non è ancora terminato il metodo init()
         */
        executorThread.start();
    }

    /**
     * see start()
     */
    public synchronized void run() {

        LOG.log(Level.INFO, "{0} Running JobManager for {1}", new Object[]{LOG_PREFIX, managedPrimaryJob});
        try {

            init();

            transition(ASSIGNMENT);

            managedPrimaryJob.setJobState(JobStatus.State.RUNNING);

        } catch (Throwable t) {
            LOG.log(Level.SEVERE, LOG_PREFIX + " throwable during init " + t.getLocalizedMessage(), t);
            handleJobCompletion(JobStatus.State.FAILED);
        }

    }

    public void stop() {
        LOG.log(Level.INFO, "{0} stopped", new Object[]{LOG_PREFIX});
        timeoutTask.cancel();
    }

    /**
     * Indicates if the node which id is {@code nodeID} was blacklisted by
     * this JobManager
     * 
     * @param nodeID
     * @return True if the node was blacklisted by this JobManager
     */
    public boolean isBlackListed(String nodeID) {
        return blackList.contains(nodeID);
    }
    ///////////////////////////////////////////////////////////////////////////
    //  Private methods
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Initiated the job environment and all the data structures
     */
    private void init() throws Exception {

        setup(isReassigned);

        //FIXME  backupManagerAssignment must be done after inputSplit generation
        //in order to make backup master known of map number
        backupManagerAssignment();
        //FIXME slaveSelection must be done after inputSplit generation??
        slaveSelection();

        initTasks();

        if (isReassigned) {
            loadJobStatus();
            if (isJobCompleted()) {
                if (reducerOutputToRetrieve.isEmpty()) {
                    LOG.log(Level.INFO, "{0} job reassigned is totally completed", new Object[]{LOG_PREFIX});
                    handleJobCompletion(JobStatus.State.SUCCEEDED);
                } else {
                    LOG.log(Level.INFO, "{0} job reassigned is completed. Sending reduce output to retrive", new Object[]{LOG_PREFIX});
                    sendReduceCompletion();
                }
            }
        }

    }

    /**
     * Setup job configuration and retrieve input files from user node
     * @param isReassigned True if the job has been reassigned to this job
     * manager in consequence of a primary failure
     */
    private void setup(boolean isReassigned) {
        Job job = managedPrimaryJob.getJob();

        updateWorkingDir();

        if (!isReassigned) {
            //if it is a reassigned job the defaults are alredy set
            setDefaults(job);
        }

        //Transfer all input from user temp dir into job input dir
        //Note: if it is a reassinged job all the input may be already retrieved
        retrieveInput();
    }

    /**
     * Sets the job working directory as {@link FTPManager#getHomeDir() }
     * @param job
     */
    private void updateWorkingDir() {
        FileSystemView fsv = nodeFSM.getNodeModule().getFileSystemView();
        String workingDir = fsv.getRootDir();
        managedPrimaryJob.getJob().setWorkingDirectory(new Path(workingDir));
    }

    private void setDefaults(Job job) {

        Configuration conf = job.getConfiguration();

        loadProperties(conf);

    }

    private void retrieveInput() {
        LOG.log(Level.INFO, "{0} retrieving input", new Object[]{LOG_PREFIX});
        Job job = managedPrimaryJob.getJob();
        Configuration c = job.getConfiguration();
        FileSystemView fsv = nodeFSM.getNodeModule().getFileSystemView();
        FTPManager ftpM = nodeFSM.getNodeModule().getFtpManager();

        String jobWorkDirPath = job.getJobID().toString() + "/";
        fsv.mkdirs(jobWorkDirPath);

//        String userTmpDir = job.getConfiguration().get(MRJobConfig.JOB_DIR_ATTR);
        NodeInfo userNode = managedPrimaryJob.getUserNodeInfo();

        String inputDir = c.get(MRJobConfig.JOB_INPUT_DIR_ATTR);
        fsv.mkdirs(inputDir);
        boolean success = ftpM.downloadDir(userNode, inputDir);
        if (!success) {
            throw new RuntimeException(LOG_PREFIX + " failed to retrieve input files");
        }

        String jarsDir = c.get(MRJobConfig.JOB_JAR_DIR_ATTR);
        fsv.mkdirs(jarsDir);
        success = ftpM.downloadDir(userNode, jarsDir);
        if (!success) {
            throw new RuntimeException(LOG_PREFIX + " failed to retrieve jar files");
        }
        LOG.log(Level.INFO, "{0} input retrieved", new Object[]{LOG_PREFIX});
    }

    /**
     * Initialize the data structure used to manage the task scheduling
     *
     * @throws Exception
     */
    private void initTasks() throws Exception {

        Job job = managedPrimaryJob.getJob();
        //TODO if it is a reassigned job could be no need to
        //generate input splits if all map tasks are successfull
        int numMapper = generateInputSplits();

        submittedMapTaskAttempt = new HashMap<TaskID, TreeSet<JobTask>>();
        successfullMapTaskAttempt = new HashMap<TaskID, TreeSet<JobTask>>();
        failedMapTaskAttempt = new HashMap<TaskID, TreeSet<JobTask>>();
        for (int id = 0; id < numMapper; id++) {
            TaskID taskID = new TaskID(job.getJobID(), TaskType.MAP, id);
            submittedMapTaskAttempt.put(taskID, new TreeSet<JobTask>(new JobTaskComparator()));
            successfullMapTaskAttempt.put(taskID, new TreeSet<JobTask>(new JobTaskComparator()));
            failedMapTaskAttempt.put(taskID, new TreeSet<JobTask>(new JobTaskComparator()));
        }

        submittedReduceTaskAttempt = new HashMap<TaskID, TreeSet<JobTask>>();
        successfullReduceTaskAttempt = new HashMap<TaskID, TreeSet<JobTask>>();
        failedReduceTaskAttempt = new HashMap<TaskID, TreeSet<JobTask>>();
        reducerOutputToRetrieve = new HashSet<TaskID>();
        for (int id = 0; id < job.getNumReduceTasks(); id++) {
            TaskID taskID = new TaskID(job.getJobID(), TaskType.REDUCE, id);
            submittedReduceTaskAttempt.put(taskID, new TreeSet<JobTask>(new JobTaskComparator()));
            successfullReduceTaskAttempt.put(taskID, new TreeSet<JobTask>(new JobTaskComparator()));
            failedReduceTaskAttempt.put(taskID, new TreeSet<JobTask>(new JobTaskComparator()));
            reducerOutputToRetrieve.add(taskID);
        }

    }

    /**
     * Load all the data structures to make the state consistent with the
     * {@link  PrimaryJob} status.
     * Reconstruct the state from {@link PrimaryJob#getJobTaskList() }
     * Used when the job is reassigned to a new master in consequence of a
     * primary failure.
     */
    private void loadJobStatus() {
        //TODO handle update of blacklisted nodes
        LOG.log(Level.INFO, "{0} loading job status", new Object[]{LOG_PREFIX});
        NetworkModuleIF netModule = nodeFSM.getNetworkModule();
        Set<NodeInfo> allSlaves = netModule.getAllSlaves();
        for (JobTask jobTask : managedPrimaryJob.getJobTaskList()) {
            TaskID taskID = jobTask.getTask().getTaskID();
            NodeInfo slaveNode = find(allSlaves, jobTask.getSlaveID());
            if (slaveNode != null) {
                Set<JobTask> current = currentTaskAssignment.get(slaveNode);
                if (current == null) {
                    current = new HashSet<JobTask>();
                    currentTaskAssignment.put(slaveNode, current);
                }
                current.add(jobTask);
            }
            if (jobTask.isMap()) {
                //MAP task
                switch (jobTask.getTaskStatus()) {
                    case ASSIGNED:
                        submittedMapTaskAttempt.get(taskID).add(jobTask);
                        break;
                    case FAILED:
                        failedMapTaskAttempt.get(taskID).add(jobTask);
                        break;
                    case SUCCESS:
                        successfullMapTaskAttempt.get(taskID).add(jobTask);
                        break;
                    default:
                        throw new AssertionError();
                }
            } else {
                //REDUCE TASK
                switch (jobTask.getTaskStatus()) {
                    case ASSIGNED:
                        submittedReduceTaskAttempt.get(taskID).add(jobTask);
                        break;
                    case FAILED:
                        failedReduceTaskAttempt.get(taskID).add(jobTask);
                        break;
                    case SUCCESS:
                        successfullReduceTaskAttempt.get(taskID).add(jobTask);
                        if (jobTask.isOutputRetrieved()) {
                            reducerOutputToRetrieve.remove(taskID);
                        }
                        break;
                    default:
                        throw new AssertionError();
                }
            }
        }
    }

    private int generateInputSplits() throws InterruptedException, ClassNotFoundException, IOException {
        LOG.log(Level.INFO, "{0} generating input splits", new Object[]{LOG_PREFIX});
        Job job = managedPrimaryJob.getJob();
        //Create input splits
        InputFormat<?, ?> inputFormat = ReflectionUtils.newInstance(job.getInputFormatClass(), job.getConfiguration());
        List<InputSplit> inputSplits = inputFormat.getSplits(job);
        Class<?> inputSplitClass = inputSplits.get(0).getClass();
        job.getConfiguration().setClass(MRJobConfig.INPUTSPLIT_CLASS_ATTR, inputSplitClass, InputSplit.class);
        int numMapper = inputSplits.size();
        job.getConfiguration().setInt(MRJobConfig.NUM_MAPS, numMapper);
        LOG.log(Level.INFO, "{0} {1} input splits generated", new Object[]{LOG_PREFIX, numMapper});
        return numMapper;
    }

    private void assignment() {

        boolean newBackupManager = backupManagerAssignment();

        slaveSelection();

        LinkedList<JobTask> scheduledTasks = taskScheduler.schedule();

        if (newBackupManager || scheduledTasks.size() > 0) {
            //A job update msg must be sended to backup masters
            MessageEvent jobUpdate = new MessageEvent(EventType.JOB_UPDATE, managedPrimaryJob.getJobID(), managedPrimaryJob.getBackupMasterList(), scheduledTasks);
            for (Iterator<String> backupMastersIt = managedPrimaryJob.getBackupMasterList().iterator(); backupMastersIt.hasNext();) {
                String backupID = backupMastersIt.next();
                try {
                    nodeFSM.getNodeModule().getNetworkModule().send(backupID, jobUpdate);
                } catch (IOException ex) {
                    ex.printStackTrace();
                    LOG.log(Level.SEVERE, null, ex);
                    if (!nodeFSM.getNodeModule().getNetworkModule().isAlive(backupID)) {
                        //FIXME The node seems to be unavailable
                        //simply remove it from the backupMasterList and wait for next
                        //assignment cycle to find a new backupMaster
                        //If no backup are available it will immediately find a new one
                        backupMastersIt.remove();
                    }
                }
            }
        }

    }

    /**
     * Search for new master node to assign them the role of backup master
     * if some new backup master are required
     * 
     */
    private boolean backupManagerAssignment() {
        //All the available master as backup master
        int requiredBackupNode = 0;
        if (REQUIRED_BACKUP_MANAGER == -1) {
            requiredBackupNode = nodeFSM.getNodeModule().getNetworkModule().getNumMasters() - 1;
        } else {
            requiredBackupNode = REQUIRED_BACKUP_MANAGER;
        }
        int B = requiredBackupNode - managedPrimaryJob.getBackupMasterList().size();
        //The list of task that required an assignment
        LOG.log(Level.INFO, LOG_PREFIX + "{0} new BackupManager required", new Object[]{B});
        if (B > 0) {
            return backupManagerAssignment(B);
        }

        return false;
    }

    private boolean backupManagerAssignment(int B) {

        boolean ris = false;
        LinkedList<NodeInfo> backupList = selectNewBackupMaster(B);
        ris = !backupList.isEmpty();

        for (NodeInfo mast : backupList) {
            //Must be added before message sending to backup masters
            managedPrimaryJob.getBackupMasterList().add(mast.getID());
        }

        LOG.log(Level.INFO, "{0} BackupMasters selected: {1}", new Object[]{LOG_PREFIX, backupList});

        MessageEvent backupJobAssMsg = new MessageEvent(EventType.BACKUP_JOB_ASSIGNED, managedPrimaryJob, nodeFSM.getNodeModule().getNodeID());
        for (NodeInfo masterNode : backupList) {
            try {
                nodeFSM.getNetworkModule().send(masterNode, backupJobAssMsg);

            } catch (IOException ex) {
                ex.printStackTrace();
                LOG.log(Level.WARNING, "{0} IOex during backup assignment to {1}. Removing it from backupMasterList", new Object[]{LOG_PREFIX, masterNode.getID()});
                //Direcly remove for assignment message failure.
                managedPrimaryJob.getBackupMasterList().remove(masterNode.getID());
            }
        }

        return ris;

    }

    private LinkedList<NodeInfo> selectNewBackupMaster(int B) {
        LinkedList<NodeInfo> backupMasterList = new LinkedList<NodeInfo>();
        int count = 0;
        for (NodeInfo manager : sortedMasters()) {
            if (!backupMasterList.contains(manager)
                    && !managedPrimaryJob.getBackupMasterList().contains(manager.getID())
                    && !blackList.contains(manager.getID())) {
                backupMasterList.add(manager);
                count++;
                if (count == B) {
                    break;
                }
            }
        }

        return backupMasterList;
    }

    /**
     * Returns a {@link TreeSet} made by all the other masters node info,
     * ordered by a {@link NodeLoadComparator } (load ascending values)
     * @return
     */
    private Set<NodeInfo> sortedMasters() {
        Set<NodeInfo> masters = nodeFSM.getNetworkModule().getOtherMasters();
        TreeSet<NodeInfo> sortedMasters = new TreeSet<NodeInfo>(new NodeLoadComparator());
        for (NodeInfo nodeInfo : masters) {
            sortedMasters.add(nodeInfo);
        }
        return sortedMasters;
    }

    private void slaveSelection() {
        //Like Hadoop MapReduce the required number of slaves node for a job should be
        //the number of the map tasks
        int requiredSlave = managedPrimaryJob.getJob().getConfiguration().getInt(MRJobConfig.NUM_MAPS, -1);
        int S = requiredSlave - currentTaskAssignment.size();
        //The list of task that required an assignment
        LOG.log(Level.INFO, LOG_PREFIX + "{0} new slave required", new Object[]{S});
        if (S > 0) {
            selectNewSlave(S);
        }
    }

    private void selectNewSlave(int S) {

        int count = 0;
        LinkedList<NodeInfo> newSelectedSlaves = new LinkedList<NodeInfo>();
        //First try to select new slave if available
        for (NodeInfo slave : sortedSlaves()) {
            if (!currentTaskAssignment.containsKey(slave)
                    && !blackList.contains(slave.getID())) {
                newSelectedSlaves.add(slave);
                currentTaskAssignment.put(slave, new HashSet<JobTask>());
            }
            count++;
            if (count == S) {
                return;
            }
        }
        LOG.log(Level.INFO, LOG_PREFIX + "New {0} slaves selected", new Object[]{newSelectedSlaves.size()});
    }

    /**
     * Returns a {@link TreeSet} made by all the other slaves node info,
     * ordered by a {@link NodeLoadComparator } (load ascending values)
     * @return
     */
    private Set<NodeInfo> sortedSlaves() {
        Set<NodeInfo> slaves = nodeFSM.getNetworkModule().getAllSlaves();
        TreeSet<NodeInfo> sortedSlaves = new TreeSet<NodeInfo>(new NodeLoadComparator());
        for (NodeInfo nodeInfo : slaves) {
            sortedSlaves.add(nodeInfo);
        }
        return sortedSlaves;
    }

    /**
     * Fills {@link #taskToCheck} with all the task attempt id
     * for those task attempt currently submitted
     *
     */
    private void nextTasksToCheck() {
        taskToCheck.clear();
        /**
         * TODO ottimizzare: si possono usare
         * submittedReduceTaskAttempt  e submittedMapTaskAttempt ??
         */
        for (JobTask jt : managedPrimaryJob.getJobTaskList()) {
            if (jt.getTaskStatus() == TaskStatus.ASSIGNED) {
                taskToCheck.add(jt);
            }
        }
    }

    /**
     * Makes failed all the {@link JobTask } from which
     * no slave check has been received.
     * @return The list of all {@link JobTask } made failed
     */
    private LinkedList<JobTask> checkTasks() {
        LinkedList<JobTask> failedJobTasks = new LinkedList<JobTask>();

        LOG.log(Level.INFO, "{0} No slave check received for these tasks {1}", new Object[]{LOG_PREFIX, taskToCheck});
        for (Iterator<JobTask> it = taskToCheck.iterator(); it.hasNext();) {
            JobTask jt = it.next();
            it.remove();
            setFailed(jt);
            failedJobTasks.add(jt);
        }

        return failedJobTasks;
    }

    private JobTask getJobTask(String taskAttID) {
        for (JobTask jt : managedPrimaryJob.getJobTaskList()) {
            if (jt.getTaskAttIDAsString().equals(taskAttID)) {
                return jt;
            }
        }
        return null;
    }

    private void setFailed(JobTask jobTask) {
        TaskID taskID = jobTask.getTask().getTaskAttID().getTaskID();

        if (jobTask.isMap()) {
            submittedMapTaskAttempt.get(taskID).remove(jobTask);
            successfullMapTaskAttempt.get(taskID).remove(jobTask);
            failedMapTaskAttempt.get(taskID).add(jobTask);
        } else {
            submittedReduceTaskAttempt.get(taskID).remove(jobTask);
            successfullReduceTaskAttempt.get(taskID).remove(jobTask);
            failedReduceTaskAttempt.get(taskID).add(jobTask);
        }

        jobTask.setTaskStatus(TaskStatus.FAILED);
    }

    private LinkedList<JobTask> setSlaveFailed(String failedSlaveID) {
        LinkedList<JobTask> failedJobTask = new LinkedList<JobTask>();

        //TODO ottimizzare: solo per task sottomessi o successfull dello slave in questione
        for (JobTask jobTask : managedPrimaryJob.getJobTaskList()) {
            if (failedSlaveID.equals(jobTask.getSlaveID()) && jobTask.getTaskStatus() != TaskStatus.FAILED) {
                setFailed(jobTask);
                failedJobTask.add(jobTask);
            }
        }
        removeSlaveNode(failedSlaveID);

        blackList.add(failedSlaveID);

        return failedJobTask;
    }

    private boolean isJobCompleted() {
        return mapPhaseCompleted() && reducePhaseCompleted();
    }

    private boolean mapPhaseCompleted() {
        for (TreeSet<JobTask> treeSet : successfullMapTaskAttempt.values()) {
            if (treeSet.isEmpty()) {
                return false;
            }
        }
        return true;
    }

    private boolean reducePhaseCompleted() {
        for (TreeSet<JobTask> treeSet : successfullReduceTaskAttempt.values()) {
            if (treeSet.isEmpty()) {
                return false;
            }
        }
        return true;
    }

    private void updateMapPhaseProgress() {
        float progress = 0F;
        for (TreeSet<JobTask> treeSet : successfullMapTaskAttempt.values()) {
            if (!treeSet.isEmpty()) {
                progress++;
            }
        }
        //Percentage of successfull completed map tasks
        progress = progress / successfullMapTaskAttempt.size();
        managedPrimaryJob.getJob().getStatus().setMapProgress(progress);
    }

    private void updateReducePhaseProgress() {
        float progress = 0F;
        for (TreeSet<JobTask> treeSet : successfullReduceTaskAttempt.values()) {
            if (!treeSet.isEmpty()) {
                progress++;
            }
        }
        //Percentage of successfull completed map tasks
        progress = progress / successfullReduceTaskAttempt.size();
        managedPrimaryJob.getJob().getStatus().setReduceProgress(progress);
    }

    /**
     * Sends a {@link EventType#JOB_UPDATE } message to all backup masters.
     * If a failure is detected during message sending because the backup node
     * is not alive, it is removed from the backup node list
     * @param updatedJobTask
     */
    private void sendJobUpdate(JobTask updatedJobTask) {
        LinkedList<JobTask> updatedJobTasks = new LinkedList<JobTask>();
        updatedJobTasks.add(updatedJobTask);
        sendJobUpdate(updatedJobTasks);
    }

    /**
     * Sends a {@link EventType#JOB_UPDATE } message to all backup masters.
     * If a failure is detected during message sending because the backup node
     * is not alive, it is removed from the backup node list
     * @param completedJobTask
     */
    private void sendJobUpdate(LinkedList<JobTask> updatedTaskList) {

        MessageEvent jobUpdateMsg = new MessageEvent(EventType.JOB_UPDATE, managedPrimaryJob.getJobID(), managedPrimaryJob.getBackupMasterList(), updatedTaskList);
        LOG.log(Level.INFO, "{0} Sending jobUpdateMsg to backup masters", new Object[]{LOG_PREFIX});
        for (Iterator<String> backupMastersIt = managedPrimaryJob.getBackupMasterList().iterator(); backupMastersIt.hasNext();) {
            String backupID = backupMastersIt.next();
            try {
                nodeFSM.getNodeModule().getNetworkModule().send(backupID, jobUpdateMsg);
            } catch (IOException ex) {
                LOG.log(Level.SEVERE, LOG_PREFIX + ex.getLocalizedMessage(), ex);
                if (!nodeFSM.getNodeModule().getNetworkModule().isAlive(backupID)) {
                    LOG.log(Level.INFO, LOG_PREFIX + "Removing backup master: " + backupID);
                    backupMastersIt.remove();
                }
            }
        }
    }

    /**
     * Sends a {@link EventType#REDUCE_COMPLETED} message to the user node
     * @param completedJobTask
     */
    private void sendReduceCompleted(JobTask completedJobTask) {
        LOG.log(Level.INFO, "{0} sending reduce completed event for {1}", new Object[]{LOG_PREFIX, completedJobTask});
        MessageEvent reduceCompletedMsg = new MessageEvent(EventType.REDUCE_COMPLETED, completedJobTask, nodeFSM.getNodeModule().getNodeInfo());
        NodeInfo userNode = managedPrimaryJob.getUserNodeInfo();
        try {
            nodeFSM.getNetworkModule().send(userNode, reduceCompletedMsg);
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, LOG_PREFIX + " " + ex.getLocalizedMessage(), ex);
        }
    }

    /**
     * Sends a {@link EventType#REDUCE_COMPLETED} message to the user node
     * for every reduce task which output has not been retrieved
     * Only the first reduce attempt are considered
     */
    private void sendReduceCompletion() {
        for (TaskID taskID : reducerOutputToRetrieve) {
            JobTask jt = successfullReduceTaskAttempt.get(taskID).first();
            sendReduceCompleted(jt);
        }
    }

    private void handleJobCompletion(JobStatus.State completionState) {
        timer.cancel();
        LOG.log(Level.INFO, "{0} job completed with status {1}", new Object[]{LOG_PREFIX, completionState});
        managedPrimaryJob.setJobState(completionState);
        InnerEvent jobCompletedMsg = new InnerEvent(EventType.JOB_COMPLETED, managedPrimaryJob.getJobID(), managedPrimaryJob.getJobState());
        //Send to NODE.MASTER.MANAGMENT
        nodeFSM.handleEvent(jobCompletedMsg);
        //NOTE:After this no more transition to ASSIGNMENT will be done
        //so the job manager will be stopped
    }

    private void removeSlaveNode(String failedSlaveID) {
        for (Iterator<NodeInfo> it = currentTaskAssignment.keySet().iterator(); it.hasNext();) {
            NodeInfo slaveNode = it.next();
            if (slaveNode.getID().equals(failedSlaveID)) {
                it.remove();
            }
        }
    }

    /**
     * Returns a {@link StringBuilder } that contains a string description
     * of the state of this JobManager
     * @return
     */
    private StringBuilder getStateDesc() {
        StringBuilder sb = new StringBuilder(managedPrimaryJob.toString());
        sb.append("\n");
        for (Iterator<Entry<NodeInfo, Set<JobTask>>> it = currentTaskAssignment.entrySet().iterator(); it.hasNext();) {
            Entry<NodeInfo, Set<JobTask>> entry = it.next();
            sb.append("Slave: ").append(entry.getKey().getID()).append(" assigned task: ").append(entry.getValue()).append("\n");
        }

        return sb;
    }

    private void loadProperties(Configuration conf) {

        Properties p = P2PMapReduceConfig.getConfiguration();
        for (Entry<Object, Object> entry : p.entrySet()) {
            conf.set(entry.getKey().toString(), entry.getValue().toString());
        }

    }

    /**
     *
     * @param allSlaves
     * @param slaveID
     * @return
     */
    private NodeInfo find(Set<NodeInfo> allSlaves, String slaveID) {
        if (slaveID != null) {
            for (NodeInfo nodeInfo : allSlaves) {
                if (nodeInfo.getID().equals(slaveID)) {
                    return nodeInfo;
                }
            }
        }

        return null;
    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM State classes
    ///////////////////////////////////////////////////////////////////////////
    private class JobManagerState extends NodeState {
    }

    private class Assignment extends JobManagerState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, "{0} Entering JOB_MANAGER.ASSIGNMENT. Managed PrimaryJob state is:\n{1}", new Object[]{LOG_PREFIX, getStateDesc()});
            assignment();
            LOG.log(Level.INFO, LOG_PREFIX + "Transition to JOB_MANAGER.MONITORING");
            transition(MONITORING);
        }
    }

    private class Monitoring extends JobManagerState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, LOG_PREFIX + "Entering JOB_MANAGER.MONITORING. Managed PrimaryJob state is:\n" + getStateDesc());
            nextTasksToCheck();
            LOG.log(Level.INFO, LOG_PREFIX + "Waiting slave check for these tasks:\n" + LoggingUtil.elementByLine(taskToCheck));
            timeoutTask = new TimeoutTask();
            timer.schedule(timeoutTask, MONITORING_TIMEOUT);
        }

        @Override
        protected void exitAction() {
            timeoutTask.cancel();
        }

        @Override
        public void taskCompleted(String completedTaskAttID, TaskStatus completedTaskStatus) {
            LOG.log(Level.INFO, "{0}Task {1} completed with status {2}", new Object[]{LOG_PREFIX, completedTaskAttID, completedTaskStatus});
            JobTask completedJobTask = getJobTask(completedTaskAttID);
            completedJobTask.setTaskStatus(completedTaskStatus);

            TaskID taskID = completedJobTask.getTask().getTaskID();
            if (taskToCheck.remove(completedJobTask)) {
                LOG.log(Level.INFO, "{0}Task {1} not checked but completed", new Object[]{LOG_PREFIX, completedTaskAttID});
            }

            /*
             * Establishes if this task is last map task attempt
             * awaited to complete map phase
             */
            boolean lastMapTaskAttempt = false;
            if (completedJobTask.isMap()) {

                submittedMapTaskAttempt.get(taskID).remove(completedJobTask);
                if (completedTaskStatus == TaskStatus.SUCCESS) {
                    boolean wasMapPhaseCompleted = mapPhaseCompleted();
                    successfullMapTaskAttempt.get(taskID).add(completedJobTask);
                    lastMapTaskAttempt = !wasMapPhaseCompleted && mapPhaseCompleted();
                } else {
                    failedMapTaskAttempt.get(taskID).add(completedJobTask);
                }
                updateMapPhaseProgress();
            } else if (completedJobTask.isReduce()) {
                submittedReduceTaskAttempt.get(taskID).remove(completedJobTask);
                if (completedTaskStatus == TaskStatus.SUCCESS) {
                    successfullReduceTaskAttempt.get(taskID).add(completedJobTask);
                    sendReduceCompleted(completedJobTask);
                } else {
                    failedReduceTaskAttempt.get(taskID).add(completedJobTask);
                }
                updateReducePhaseProgress();
            } else {
                throw new IllegalStateException("Unhandeled task type " + completedJobTask.getTask().getTaskType());
            }

            LOG.log(Level.INFO, "{0}JobTasks status:\n{1}", new Object[]{LOG_PREFIX, LoggingUtil.elementByLine(managedPrimaryJob.getJobTaskList())});

            sendJobUpdate(completedJobTask);

            if (completedTaskStatus == TaskStatus.FAILED) {
                LOG.log(Level.INFO, LOG_PREFIX + "Transition to JOB_MANAGER.ASSIGNMENT");
                transition(ASSIGNMENT);
            } else if (lastMapTaskAttempt) {
                LOG.log(Level.INFO, LOG_PREFIX + "MAP PHASE COMPLETED.Transition to JOB_MANAGER.ASSIGNMENT");
                transition(ASSIGNMENT);
            } else if (isJobCompleted()) {
                LOG.log(Level.INFO, "{0} Job {1} completed. Waiting for user output retrieval for {2}", new Object[]{LOG_PREFIX, managedPrimaryJob.getJobID(), LoggingUtil.elementByLine(reducerOutputToRetrieve)});
                //Interrupt assignment cycle
                timeoutTask.cancel();
            }

        }

        @Override
        public void slaveCheck(String taskID) {
            JobTask jt = getJobTask(taskID);
            taskToCheck.remove(jt);
            LOG.log(Level.INFO, "{0}Received a slave check for task {1}.\nRemaining tasks to check: {2}", new Object[]{LOG_PREFIX, taskID, taskToCheck});
        }

        @Override
        public void backupMasterFailure(String backupID) {
            LOG.log(Level.INFO, "{0}Backup master {1} failed.Transition to JOB_MANAGER.ASSIGNMENT", new Object[]{LOG_PREFIX, backupID});
            managedPrimaryJob.getBackupMasterList().remove(backupID);
            blackList.add(backupID);
            timeoutTask.cancel();
            transition(ASSIGNMENT);
        }

        @Override
        public void slaveFailure(String failedSlaveID) {
            /*
             * If a node is failed all the tasks relative to that node
             * have to be declared failed
             */
            LOG.log(Level.INFO, LOG_PREFIX + "Slave {0} failed.Transition to JOB_MANAGER.ASSIGNMENT", failedSlaveID);
            LinkedList<JobTask> failedJobTasks = setSlaveFailed(failedSlaveID);

            if (!failedJobTasks.isEmpty()) {
                updateMapPhaseProgress();
                updateReducePhaseProgress();
                sendJobUpdate(failedJobTasks);
            }

            timeoutTask.cancel();
            transition(ASSIGNMENT);
        }

        @Override
        public void outputRetrieved(JobTask reduceJT, boolean completetionState) {
            LOG.log(Level.INFO, "{0} output retrieval for {1} completed with status {2}", new Object[]{LOG_PREFIX, reduceJT, completetionState});
            //Find the relative JobTask
            JobTask jobTask = null;
            for (JobTask jt : managedPrimaryJob.getJobTaskList()) {
                if (jt.equals(reduceJT)) {
                    jobTask = jt;
                    break;
                }
            }

            if (completetionState) {

                jobTask.setOutputRetrieved(true);

                TaskID reducerID = jobTask.getTask().getTaskID();
                reducerOutputToRetrieve.remove(reducerID);

                sendJobUpdate(jobTask);

                if (reducerOutputToRetrieve.isEmpty()) {
                    //All the outputs have been retrieved from user.
                    //Made the job successfully completed
                    timeoutTask.cancel();
                    handleJobCompletion(JobStatus.State.SUCCEEDED);
                }
            } else {
                //Failure in output retrieval of a reduce task
                //made the reduce task attempt as failed or slave node as
                //failed if it is not alive
                if (nodeFSM.getNetworkModule().isAlive(jobTask.getSlaveID())) {

                    setFailed(jobTask);
                    sendJobUpdate(jobTask);
                    updateReducePhaseProgress();
                } else {
                    LinkedList<JobTask> failedJT = setSlaveFailed(jobTask.getSlaveID());
                    sendJobUpdate(failedJT);
                    updateMapPhaseProgress();
                    updateReducePhaseProgress();
                }

                transition(ASSIGNMENT);
            }
        }

        @Override
        public void timeout(Object param) {
            if (!timeoutTask.cancel()) {
                LOG.log(Level.INFO, LOG_PREFIX + "Timeout. Transition to JOB_MANAGER.ASSIGNMENT");
                LinkedList<JobTask> failedJobTasks = checkTasks();
                if (!failedJobTasks.isEmpty()) {
                    updateMapPhaseProgress();
                    updateReducePhaseProgress();
                    sendJobUpdate(failedJobTasks);
                }
                transition(ASSIGNMENT);
            } else {
                /*
                 * 
                 * If cancel() returns true the current timeoutTask is 
                 * not fired, so this call comes from a previously scheduled
                 * timeoutTask which was not cancelled in time (before leaving
                 * Monitoring state)
                 */
                LOG.log(Level.SEVERE, "Current TimeoutTask not fired!Probably obsolete fire. Ignore it and reschedule");
                timeoutTask = new TimeoutTask();
                timer.schedule(timeoutTask, MONITORING_TIMEOUT);
            }
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Generic inner classes
    ///////////////////////////////////////////////////////////////////////////
    private class TimeoutTask extends TimerTask {

        @Override
        public void run() {
            timeout(null);
        }
    }

    /**
     * A TaskScheduler takes care of map and reduce tasks scheduling.
     * First searches for the tasks that required to be submitted, so create new
     * task attempts for these tasks and submit them to the best suitable slave
     * nodes.
     */
    private class TaskScheduler {

        /**
         * Find the tasks that must be submitted, select a slave for each
         * task and then submit all the task to the relative slave.
         * Updates the {@link JobManager} data structure.
         *
         * @return the list of the JobTask successfully submitted to the relative slave
         */
        public LinkedList<JobTask> schedule() {

            if (!mapPhaseCompleted()) {
                return mapSchedule();
            } else {
                return reduceSchedule();
            }
        }

        private LinkedList<JobTask> mapSchedule() {
            LinkedList<JobTask> jobTaskSubmitted = new LinkedList<JobTask>();
            LinkedList<Task> mapTaskToSubmit = mapTaskToSubmit();
            if (!mapTaskToSubmit.isEmpty()) {
                Map<NodeInfo, LinkedList<JobTask>> mapJobTaskToSubmit = assign(mapTaskToSubmit);
                String nodeID = nodeFSM.getNodeModule().getNodeID();
                for (Iterator<Entry<NodeInfo, LinkedList<JobTask>>> it = mapJobTaskToSubmit.entrySet().iterator(); it.hasNext();) {
                    Entry<NodeInfo, LinkedList<JobTask>> entry = it.next();
                    NodeInfo slave = entry.getKey();
                    LinkedList<Task> taskList = constructTaskList(entry.getValue());
                    MessageEvent taskAssMsg = new MessageEvent(EventType.TASK_ASSIGNED, taskList, nodeID);
                    try {
                        LOG.log(Level.INFO, LOG_PREFIX + "Subimitting tasks:\n{0} to {1}", new Object[]{LoggingUtil.elementByLine(entry.getValue()), entry.getKey().getID()});
                        nodeFSM.getNetworkModule().send(slave, taskAssMsg);
                        //Task submission should be successfull
                        //Update all the data structure
                        Set<JobTask> current = currentTaskAssignment.get(slave);
                        if (current == null) {
                            //XXX non dovrebbe accadere
                            current = new HashSet<JobTask>();
                            currentTaskAssignment.put(slave, current);
                        }
                        for (JobTask jobTask : entry.getValue()) {
                            current.add(jobTask);
                            TaskID taskID = jobTask.getTask().getTaskAttID().getTaskID();
                            submittedMapTaskAttempt.get(taskID).add(jobTask);
                            managedPrimaryJob.getJobTaskList().add(jobTask);
                            jobTaskSubmitted.add(jobTask);
                        }
                    } catch (IOException ex) {
                        ex.printStackTrace();
                        LOG.log(Level.SEVERE, LOG_PREFIX + "IOException during submission to " + slave.getID() + ex.getLocalizedMessage());
                    }
                }
            }
            return jobTaskSubmitted;
        }

        private LinkedList<JobTask> reduceSchedule() {

            LinkedList<JobTask> jobTaskSubmitted = new LinkedList<JobTask>();
            LinkedList<Task> reduceTaskToSubmit = reduceTaskToSubmit();
            if (!reduceTaskToSubmit.isEmpty()) {
                Map<NodeInfo, LinkedList<JobTask>> reduceJobTaskToSubmit = assign(reduceTaskToSubmit);
                String nodeID = nodeFSM.getNodeModule().getNodeID();
                for (Iterator<Entry<NodeInfo, LinkedList<JobTask>>> it = reduceJobTaskToSubmit.entrySet().iterator(); it.hasNext();) {
                    Entry<NodeInfo, LinkedList<JobTask>> entry = it.next();
                    NodeInfo slave = entry.getKey();
                    LinkedList<Task> taskList = constructTaskList(entry.getValue());
                    MessageEvent taskAssMsg = new MessageEvent(EventType.TASK_ASSIGNED, taskList, nodeID);
                    try {
                        LOG.log(Level.INFO, LOG_PREFIX + "Subimitting tasks:\n{0} to {1}", new Object[]{LoggingUtil.elementByLine(entry.getValue()), entry.getKey().getID()});
                        nodeFSM.getNetworkModule().send(slave, taskAssMsg);
                        //Task submission should be successfull
                        //Update all the data structure
                        Set<JobTask> current = currentTaskAssignment.get(slave);
                        if (current == null) {
                            //XXX non dovrebbe accadere
                            current = new HashSet<JobTask>();
                            currentTaskAssignment.put(slave, current);
                        }
                        for (JobTask jobTask : entry.getValue()) {
                            current.add(jobTask);
                            TaskID taskID = jobTask.getTask().getTaskAttID().getTaskID();
                            submittedReduceTaskAttempt.get(taskID).add(jobTask);
                            managedPrimaryJob.getJobTaskList().add(jobTask);
                            jobTaskSubmitted.add(jobTask);
                        }
                    } catch (IOException ex) {
                        //TODO manage the exception. note: the data structure will
                        //be not updated if an ioex occur
                        ex.printStackTrace();
                        LOG.log(Level.SEVERE, LOG_PREFIX + "IOException during submission to " + slave.getID() + ex.getLocalizedMessage());
                    }
                }

            }
            return jobTaskSubmitted;
        }

        /**
         * Find all the map task that must be submitted.
         * A task must be submitted because no attempt exists
         * that are already successfully completed or still alive
         * 
         * @return the list of map task to submit
         */
        private LinkedList<Task> mapTaskToSubmit() {

            LinkedList<Task> taskToSubmit = new LinkedList<Task>();

            //search for map task
            for (Entry<TaskID, TreeSet<JobTask>> entry : successfullMapTaskAttempt.entrySet()) {
                TreeSet<JobTask> successfullyAttempt = entry.getValue();
                if (successfullyAttempt.isEmpty()) {
                    //no successfully attemp exist. Search for alive attempt existence
                    TaskID tID = entry.getKey();
                    TreeSet<JobTask> submittedAttempt = submittedMapTaskAttempt.get(tID);
                    if (submittedAttempt.isEmpty()) {
                        //No alive attempt exist for this task. Create new task attempt
                        int attemptNumber = getLastMapAttempt(tID) + 1;
                        TaskAttemptID newTaskAtt = new TaskAttemptID(tID, attemptNumber);
                        Configuration c = managedPrimaryJob.getJob().getConfiguration();
                        Task newTask = new MapTask(c, newTaskAtt);
                        taskToSubmit.add(newTask);
                    }
                }
            }

            //TODO lo stesso metodo per reduce task

            return taskToSubmit;
        }

        /**
         * Find all the reduce task that must be submitted.
         * A task must be submitted because no attempt exists
         * that are already successfully completed or still alive
         *
         * @return the list of map task to submit
         */
        private LinkedList<Task> reduceTaskToSubmit() {
            LinkedList<Task> taskToSubmit = new LinkedList<Task>();

            //search for reduce task
            for (Entry<TaskID, TreeSet<JobTask>> entry : successfullReduceTaskAttempt.entrySet()) {
                TreeSet<JobTask> successfullyAttempt = entry.getValue();
                if (successfullyAttempt.isEmpty()) {
                    //no successfully attemp exist. Search for alive attempt existence
                    TaskID tID = entry.getKey();
                    TreeSet<JobTask> submittedAttempt = submittedReduceTaskAttempt.get(tID);
                    if (submittedAttempt.isEmpty()) {
                        //No alive attempt exist for this task. Create new task attempt
                        int attemptNumber = getLastReduceAttempt(tID) + 1;
                        TaskAttemptID newTaskAtt = new TaskAttemptID(tID, attemptNumber);
                        Configuration c = managedPrimaryJob.getJob().getConfiguration();
                        Task newTask = new ReduceTask(c, newTaskAtt, successfullMapTaskAttempt);
                        taskToSubmit.add(newTask);
                    }
                }
            }

            return taskToSubmit;
        }

        /**
         * For each task in the list associate the slave node with lowest load.
         * After the association of a task, the slave load is increased to 
         * consider the load after task submission.
         *
         * @param taskList list of tasks that must be submitted
         * @return
         */
        private Map<NodeInfo, LinkedList<JobTask>> assign(LinkedList<Task> taskList) {

            Map<NodeInfo, LinkedList<JobTask>> ris = new HashMap<NodeInfo, LinkedList<JobTask>>();
            PriorityQueue<NodeInfo> currentSlaves = getCurrentSlaves();

            if (currentSlaves.isEmpty()) {
                LOG.log(Level.INFO, "{0} No slaves available to submit map tasks", new Object[]{LOG_PREFIX});
                return ris;
            }

            /*
             * For each task find the lowest load slave and add the relative
             * job task to the list
             */
            for (Task task : taskList) {
                NodeInfo slave = currentSlaves.remove();
                JobTask newJobTask = new JobTask(task, slave.getID());
                incrementLoad(slave, 1);
                currentSlaves.add(slave);

                LinkedList<JobTask> list = ris.get(slave);
                if (list == null) {
                    list = new LinkedList<JobTask>();
                    ris.put(slave, list);
                }
                list.add(newJobTask);

            }

            return ris;
        }

        /**
         *
         * @param tID
         * @return
         */
        private int getLastMapAttempt(TaskID tID) {
            int max = 0;
            TreeSet<JobTask> submitted = submittedMapTaskAttempt.get(tID);
            if (submitted != null && !submitted.isEmpty()) {
                JobTask last = submitted.last();
                if (last != null) {
                    max = Math.max(max, last.getTask().getTaskAttID().getId());
                }
            }


            TreeSet<JobTask> failed = failedMapTaskAttempt.get(tID);
            if (failed != null && !failed.isEmpty()) {
                JobTask last = failed.last();
                if (last != null) {
                    max = Math.max(max, last.getTask().getTaskAttID().getId());
                }
            }

            return max;
        }

        private int getLastReduceAttempt(TaskID tID) {
            int max = 0;
            TreeSet<JobTask> submitted = submittedReduceTaskAttempt.get(tID);
            if (submitted != null && !submitted.isEmpty()) {
                JobTask last = submitted.last();
                if (last != null) {
                    max = Math.max(max, last.getTask().getTaskAttID().getId());
                }
            }


            TreeSet<JobTask> failed = failedReduceTaskAttempt.get(tID);
            if (failed != null && !failed.isEmpty()) {
                JobTask last = failed.last();
                if (last != null) {
                    max = Math.max(max, last.getTask().getTaskAttID().getId());
                }
            }

            return max;
        }

        /**
         *
         * @return a PriorityQueue in which the node info of current available
         *          slaves are ordered by increasing load
         */
        private PriorityQueue<NodeInfo> getCurrentSlaves() {

            int initialCapacity = currentTaskAssignment.isEmpty() ? 1 : currentTaskAssignment.size();
            PriorityQueue<NodeInfo> ris = new PriorityQueue<NodeInfo>(initialCapacity, new NodeLoadComparator());
//            TreeMap<Double, NodeInfo> ris = new TreeMap<Double, NodeInfo>();
            Set<NodeInfo> allSlaves = nodeFSM.getNetworkModule().getAllSlaves();
            for (NodeInfo previous : currentTaskAssignment.keySet()) {
                NodeInfo current = find(allSlaves, previous.getID());
                if (current != null) {
                    ris.add(current);
                }
            }

            return ris;
        }

        /**
         * Increment the load value relatively to the adding of new 'numTaskAdded'
         * task
         * @param slave
         * @param numTaskAdded
         */
        private void incrementLoad(NodeInfo slave, int numTaskAdded) {
            slave.setLoad(slave.getLoad() + numTaskAdded);
        }

        private LinkedList<Task> constructTaskList(LinkedList<JobTask> jobTaskList) {
            LinkedList<Task> taskList = new LinkedList<Task>();
            for (JobTask jobTask : jobTaskList) {
                taskList.add(jobTask.getTask());
            }
            return taskList;
        }
    }
}
