/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule;

import java.io.IOException;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.entity.SlaveTask;
import p2pMapReduce.entity.Task;
import p2pMapReduce.entity.Task.TaskStatus;
import p2pMapReduce.entity.TaskRunner;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.InnerEvent;
import p2pMapReduce.nodeModule.event.MessageEvent;
import p2pMapReduce.nodeModule.nodeFsm.FSM;
import p2pMapReduce.nodeModule.nodeFsm.NodeFSM;
import p2pMapReduce.nodeModule.nodeFsm.NodeState;
import p2pMapReduce.util.LoggingUtil;

/**
 * Manages a {@link SlaveTask}.
 * Obtain an associated {@link TaskRunner} and wait its termination.
 * In case of failures or task completion it signals the event to the primary.
 * Every {@link #CHECK_TIMEOUT } period sends a check message to the primary.
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class TaskManager extends FSM implements Runnable {

    private static final transient Logger LOG = Logger.getLogger(TaskManager.class.getName());
    private String LOG_PREFIX;

    private static enum TimeoutEvent {

        CHECK, WAIT_PRIMARY
    }
    /**
     * Check message send period
     * NOTE: must be less than {@link p2pMapReduce.nodeModule.JobManager#MONITORING_TIMEOUT }
     */
    public static final int CHECK_TIMEOUT = JobManager.MONITORING_TIMEOUT / 3;
    /**
     * Time to wait after primary failure before make the task failed
     * if no primary update has been received
     */
    private static final int WAIT_PRIMARY_TIMEOUT = 3 * 60 * 1000;
    final TaskManagerState CHECK_PRIMARY;
    final TaskManagerState WAITING_PRIMARY;
    final TaskManagerState PRIMARY_UNAVAILABLE;
    private NodeFSM nodeFSM;
    private SlaveTask managedSlaveTask;
    private Thread executorThread;
    //TODO insert an execution state
    Timer timer;
    TimeoutTask timeoutTask = new TimeoutTask(TimeoutEvent.CHECK);
    private final Object lock;
    private boolean notified;
    private boolean stopped;

    public TaskManager(NodeFSM nodeFSM, SlaveTask slaveTask) {
        this.nodeFSM = nodeFSM;
        this.managedSlaveTask = slaveTask;
        this.LOG_PREFIX = "TaskManger(" + managedSlaveTask.getTask().getTaskAttIDAsString() + ")";
        this.lock = LOG_PREFIX + "lock";
        this.timer = new Timer(LOG_PREFIX + "Timer", true);
        CHECK_PRIMARY = new CheckPrimary();
        WAITING_PRIMARY = new WaitingPrimary();
        PRIMARY_UNAVAILABLE = new PrimaryUnavailable();
    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM methods
    ///////////////////////////////////////////////////////////////////////////
    private synchronized void timeout(Object param) {
        ((TaskManagerState) currentState()).timeout(param);
    }

    public synchronized void primaryMasterFailure(String failedPrimaryID) {
        ((TaskManagerState) currentState()).primaryMasterFailure(failedPrimaryID);
    }

    public synchronized void primaryUpdate(String updatedJobID, String primaryNodeID, double primaryLoad) {
        ((TaskManagerState) currentState()).primaryUpdate(updatedJobID, primaryNodeID, primaryLoad);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Public methods
    ///////////////////////////////////////////////////////////////////////////
    public synchronized void start() {
        if(executorThread!=null){
            LOG.log(Level.SEVERE, "{0} ERROR : Starting another thread", new Object[]{LOG_PREFIX});
        }
        executorThread = new Thread(this);
        executorThread.start();
    }

    public void run() {
        try {
            sendCheckMsg();
            transition(CHECK_PRIMARY);
            //Represents the parallel state EXECUTION
            LOG.log(Level.INFO, "{0} Starting task runner", new Object[]{LOG_PREFIX});
            TaskRunner taskRunner = managedSlaveTask.getTask().createTaskRunner(managedSlaveTask, nodeFSM.getNodeModule());
            TaskStatus completionStatus = taskRunner.call();
            handleTaskCompletion(completionStatus);


        } catch (Throwable ex) {
            LOG.log(Level.SEVERE, "{0} Task manager for task {1} interrupted by ex:\n{2}\n{3}",
                    new Object[]{
                        LOG_PREFIX,
                        managedSlaveTask.getTask().getTaskAttIDAsString(),
                        ex,
                        LoggingUtil.listStackTrace(ex.getStackTrace())});
            handleTaskCompletion(TaskStatus.FAILED);
        }
    }

    public synchronized void stop() {
        //XXX interrupt the thread cause the interruption of the message sending
        //from this thread to primary node to notify task completion
        //Simply don't force thread interrupt but wait for run() method to complete
//        executorThread.interrupt();
        if (timeoutTask != null) {
            timeoutTask.cancel();
            timer.cancel();
        }

        stopped = true;
        LOG.log(Level.INFO, LOG_PREFIX + " stopped");
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Private methods
    ///////////////////////////////////////////////////////////////////////////
    /**
     * 
     * @param taskStatus
     */
    private void handleTaskCompletion(TaskStatus taskStatus) {
        
        synchronized (lock) {
            if (stopped) {
                /*
                 * Stop viene modificato indirettamente all'interno del
                 * blocco sync(lock) in primary unavailable
                 */

                LOG.log(Level.INFO, "{0} Task execution completed with status {1} but task manager stopped", new Object[]{LOG_PREFIX, taskStatus});
                return;
            }

            LOG.log(Level.INFO, LOG_PREFIX + "Task execution completed with status {0}", new Object[]{taskStatus});
            managedSlaveTask.setTaskStatus(taskStatus);
            if (managedSlaveTask.getPrimaryID() == null) {
                LOG.log(Level.INFO, LOG_PREFIX + "Founded failed primary.Sleep until notified");
                while (!notified) {
                    //Primary failed wait for new one
                    try {
                        lock.wait();
                        LOG.log(Level.INFO, LOG_PREFIX + "Task executor wake up. notified=" + notified);
                    } catch (InterruptedException ex) {
                        LOG.log(Level.SEVERE, null, ex);
                    }
                }
            }
        }
        //Interrupt any action of task manager
        timeoutTask.cancel();
        Task t = managedSlaveTask.getTask();
        if (managedSlaveTask.getPrimaryID() == null) {
            //Primary unavailable
            LOG.log(Level.INFO, LOG_PREFIX + "Primary still unavailable.Return");
//                    managedSlaveTask.setTaskStatus(TaskStatus.FAILED);
//                    InnerEvent taskCompletedEv = new InnerEvent(EventType.TASK_COMPLETED, t.getTaskID(), TaskStatus.FAILED);
//                    nodeFSM.handleEvent(taskCompletedEv);

        } else {
            LOG.log(Level.INFO, LOG_PREFIX + "Notifing completion");
            InnerEvent taskCompletedEv = new InnerEvent(EventType.TASK_COMPLETED, t.getTaskAttIDAsString(), taskStatus);
            nodeFSM.handleEvent(taskCompletedEv);
        }



    }

    private void sendCheckMsg() {
        try {
            MessageEvent me = new MessageEvent(EventType.SLAVE_CHECK, managedSlaveTask.getTaskID());
            nodeFSM.getNodeModule().getNetworkModule().send(managedSlaveTask.getPrimaryID(), me);
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, LOG_PREFIX + ex.getLocalizedMessage(), ex);
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Fsm state classes
    ///////////////////////////////////////////////////////////////////////////
    private class TaskManagerState extends NodeState {
    }

    private class CheckPrimary extends TaskManagerState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, LOG_PREFIX + "Entering TASK_MANAGER.CHECK_PRIMARY");
            timeoutTask = new TimeoutTask(TimeoutEvent.CHECK);
            timer.schedule(timeoutTask, CHECK_TIMEOUT);
        }

        @Override
        protected void exitAction() {
            timeoutTask.cancel();
        }

        //TODO manca primary update perchè se non si accorge del fallimento
        //in tempo potrebbe essere già arrivato il nuovo primary
        @Override
        public void primaryMasterFailure(String failedPrimaryID) {
            LOG.log(Level.INFO, LOG_PREFIX + "PrimaryMaster " + failedPrimaryID + " failure. Transition to TASK_MANAGER.WAITING_PRIMARY");
            synchronized (lock) {
                managedSlaveTask.setPrimaryID(null);
            }
            transition(WAITING_PRIMARY);
        }

        @Override
        public void timeout(Object param) {
            if (param == TimeoutEvent.CHECK) {
                LOG.log(Level.INFO, LOG_PREFIX + "Timeout. Send check msg");
                timeoutTask = new TimeoutTask(TimeoutEvent.CHECK);
                timer.schedule(timeoutTask, CHECK_TIMEOUT);
                sendCheckMsg();
            }
        }
    }

    private class WaitingPrimary extends TaskManagerState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, LOG_PREFIX + "Entering TASK_MANAGER.WAITING_PRIMARY");
            timeoutTask = new TimeoutTask(TimeoutEvent.WAIT_PRIMARY);
            timer.schedule(timeoutTask, WAIT_PRIMARY_TIMEOUT);
        }

        @Override
        public void primaryUpdate(String updatedJobID, String primaryNodeID, double primaryLoad) {
            timeoutTask.cancel();
            LOG.log(Level.INFO, LOG_PREFIX + "PrimaryUpdate msg received: job {0} primaryID {1} ", new Object[]{updatedJobID, primaryNodeID});
            //ADDED sends to new primary a check msg
            //TODO controllare se sia possibile che il task sia già stato completato
            //e che il completamento sia già stato notifcato
            LOG.log(Level.INFO, LOG_PREFIX + "Sending check msg to the new primary");
            MessageEvent checkMsg = new MessageEvent(EventType.SLAVE_CHECK, managedSlaveTask.getTask().getTaskAttIDAsString());
            try {
                //TODO why send check message before data updating ???
                nodeFSM.getNodeModule().getNetworkModule().send(primaryNodeID, checkMsg);
            } catch (IOException ex) {
                ex.printStackTrace();
                LOG.log(Level.SEVERE, LOG_PREFIX + ex.getLocalizedMessage(), ex);
            }
            synchronized (lock) {
                managedSlaveTask.setPrimaryID(primaryNodeID);
                LOG.log(Level.INFO, LOG_PREFIX + "Notifing task");
                notified = true;
                lock.notifyAll();
            }
            LOG.log(Level.INFO, LOG_PREFIX + "Transition to TASK_MANAGER.CHECK_PRYMARY");
            transition(CHECK_PRIMARY);
        }

        @Override
        public void timeout(Object param) {
            if (param == TimeoutEvent.WAIT_PRIMARY) {
                LOG.log(Level.INFO, LOG_PREFIX + "Timeout. Transition to TASK_MANAGER.PRIMARY_UNAVAILABLE");
                transition(PRIMARY_UNAVAILABLE);
            }
        }
    }

    private class PrimaryUnavailable extends TaskManagerState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, LOG_PREFIX + "Entering TASK_MANAGER.PRIMARY_UNAVAILABLE");
            synchronized (lock) {

                LOG.log(Level.INFO, LOG_PREFIX + "Sending task failed completed ev");
                InnerEvent taskCompletedEv = new InnerEvent(EventType.TASK_COMPLETED, managedSlaveTask.getTask().getTaskAttIDAsString(), TaskStatus.FAILED);
                //send event to NODE.SLAVE
                nodeFSM.handleEvent(taskCompletedEv);
                LOG.log(Level.INFO, LOG_PREFIX + "Notifing task");
                notified = true;
                lock.notifyAll();
            }

        }
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Generic inner classes 
    ///////////////////////////////////////////////////////////////////////////
    private class TimeoutTask extends TimerTask {

        Object param;

        public TimeoutTask(Object param) {
            this.param = param;
        }

        @Override
        public void run() {
            timeout(param);
        }
    }
}
