/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule;

import p2pMapReduce.networkModule.NodeInfo;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.channels.FileChannel;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Set;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.fs.FileSystemView;
import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.JobID;
import p2pMapReduce.mapreduceModule.JobStatus;
import p2pMapReduce.mapreduceModule.JobStatus.State;
import p2pMapReduce.mapreduceModule.MRJobConfig;
import p2pMapReduce.mapreduceModule.conf.Configuration;
import p2pMapReduce.mapreduceModule.input.TextFileInputFormat;
import p2pMapReduce.networkModule.NetworkConfiguration;
import p2pMapReduce.networkModule.NetworkModuleFactory;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.entity.JobTask;
import p2pMapReduce.ftp.FTPManager;
import p2pMapReduce.nodeModule.event.EventIF;
import p2pMapReduce.nodeModule.event.EventListener;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.MessageEvent;
import p2pMapReduce.util.LoggingUtil;

/**
 * An utility class to start an user node, join the P2PMapReduce network 
 * and submit jobs.
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class UserNode {

    private static final transient Logger LOG = Logger.getLogger(UserNode.class.getName());
    /**
     * Time interval for master searching
     */
    public static final int MASTER_SEARCH_PERIOD = 5 * 1000;

    ///////////////////////////////////////////////////////////////////////////
    //  STATIC
    ///////////////////////////////////////////////////////////////////////////
    static {
        LoggingUtil.setLogging();
    }

    /**
     * Creates and starts a new UserNode
     * @return
     */
    public static UserNode createUserNode() {
        return new UserNode();
    }

    /**
     * Creates and starts a new UserNode providing
     * a specific {@link NetworkConfiguration}
     * @return
     */
    public static UserNode createUserNode(NetworkConfiguration nc) {
        return new UserNode(nc);
    }

    private static void copyFile(File f, String tmpDir) throws IOException {
        if (f.isDirectory()) {
            String dirPath = tmpDir + f.getName() + "/";
            File dir = new File(dirPath);
            dir.mkdirs();
            for (File f2 : f.listFiles()) {

                copyFile(f2, dirPath);
            }
        } else {
            String destPath = tmpDir + f.getName();
            File dest = new File(destPath);
            FileChannel outFC = new FileOutputStream(dest).getChannel();
            FileChannel inFC = new FileInputStream(f).getChannel();
            inFC.transferTo(0, inFC.size(), outFC);

        }
    }

    public static class UserJob {

        Job job;
        Runnable jobCompletionAction;
        String primaryID;

        public UserJob(Job job, Runnable jobCompletionAction) {
            this.job = job;
            this.jobCompletionAction = jobCompletionAction;
        }

        public Job getJob() {
            return job;
        }

        public void setJob(Job job) {
            this.job = job;
        }

        public Runnable getJobCompletionAction() {
            return jobCompletionAction;
        }

        public void setJobCompletionAction(Runnable jobCompletionAction) {
            this.jobCompletionAction = jobCompletionAction;
        }

        public String getPrimaryID() {
            return primaryID;
        }

        public void setPrimaryID(String primaryID) {
            this.primaryID = primaryID;
        }
    }

    private static class JobToAssign {

        UserJob userJob;
        NodeInfo master;

        public JobToAssign(UserJob job, NodeInfo master) {
            this.userJob = job;
            this.master = master;
        }

        public UserJob getUserJob() {
            return userJob;
        }

        public void setUserJob(UserJob job) {
            this.userJob = job;
        }

        public NodeInfo getMaster() {
            return master;
        }

        public void setMaster(NodeInfo master) {
            this.master = master;
        }
    }
    ///////////////////////////////////////////////////////////////////////////
    //  INSTANCE
    ///////////////////////////////////////////////////////////////////////////
    private NetworkConfiguration networkConfiguration;
    private NetworkModuleIF networkModule;
    /**
     * All the submitted job to this user node
     */
    private LinkedList<UserJob> subittedJob;
    /**
     * Job waiting and id to be assigned to a master node
     */
    private HashMap<Integer, JobToAssign> jobToAssign;
    /**
     * Job already assigned to a master node
     */
    private HashMap<String, UserJob> assignedJob;
    /**
     * A threaded object instructed to request id for submitted jobs
     */
    private JobSubmitter jobSubmitter;
    /**
     * A private counter of job id requests
     */
    private int requestCount;
    private Lock lock;
    private Condition notEmpty;

    private UserNode() {
        this(null);
    }

    private UserNode(NetworkConfiguration nc) {
        lock = new ReentrantLock(true);
        notEmpty = lock.newCondition();
        subittedJob = new LinkedList<UserJob>();
        jobToAssign = new HashMap<Integer, JobToAssign>();
        assignedJob = new HashMap<String, UserJob>();
        if (nc == null) {
            networkConfiguration = new NetworkConfiguration();
        } else {
            networkConfiguration = nc;
        }
        networkConfiguration.set(NetworkConfiguration.USER_NODE, true);
        networkModule = NetworkModuleFactory.newNetworkModule(networkConfiguration);
        networkModule.addEventListener(new UserNodeEventListener());
        networkModule.joinNetwork();
        jobSubmitter = new JobSubmitter();
        jobSubmitter.start();
    }

    /**
     * A way to submit a job through this user node
     * 
     * @param job
     * @throws IOException
     */
    public void submit(Job job) throws IOException {
        submit(job, null);
    }

    /**
     * Submit the job through this user node.
     *
     * @param job
     * @param jobCompletionAction A runnable object to execute when the job is
     * successfully completed. the execution will be performed in a new thread
     */
    public void submit(Job job, Runnable jobCompletionAction) {
        lock.lock();
        try {
            LOG.log(Level.INFO, "Adding new job to assign " + job);
            subittedJob.add(new UserJob(job, jobCompletionAction));
            notEmpty.signalAll();
        } finally {
            lock.unlock();
        }
    }

    public String getNodeID() {
        return networkModule.getNodeInfo().getID();
    }

    public FileSystemView getFileSystemView() {
        FileSystemView fsv = new FileSystemView(networkModule.getFtpManager().getBaseDir());
        return fsv;
    }

    public void addEventListener(EventListener eventListener) {
        networkModule.addEventListener(eventListener);
    }

    public HashMap<String, UserJob> getAssignedJob() {
        return assignedJob;
    }

    public void setAssignedJob(HashMap<String, UserJob> assignedJob) {
        this.assignedJob = assignedJob;
    }

    HashMap<Integer, JobToAssign> getJobToAssign() {
        return jobToAssign;
    }

    void setJobToAssign(HashMap<Integer, JobToAssign> jobToAssign) {
        this.jobToAssign = jobToAssign;
    }

    public LinkedList<UserJob> getSubittedJob() {
        return subittedJob;
    }

    public void setSubittedJob(LinkedList<UserJob> subittedJob) {
        this.subittedJob = subittedJob;
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Private methods
    ///////////////////////////////////////////////////////////////////////////
    private void requestId(UserJob userJob, NodeInfo master) {
        lock.lock();
        try {
            int requestNumber = requestCount++;
            JobToAssign jta = new JobToAssign(userJob, master);
            jobToAssign.put(requestNumber, jta);
            MessageEvent me = new MessageEvent(EventType.JOB_ID_REQUEST, requestNumber, networkModule.getNodeInfo());
            LOG.log(Level.INFO, "Requesting id for {0} with rn {1} to {2}", new Object[]{userJob.getJob(), requestNumber, master.getID()});
            networkModule.send(master, me);
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, null, ex);
        } finally {
            lock.unlock();
        }
    }

    private void jobIDReply(Integer requestNumber, JobID jobID) {
        try {
            JobToAssign jobToAss = jobToAssign.remove(requestNumber);
            Job job = jobToAss.getUserJob().getJob();
            job.setJobID(jobID);
            LOG.log(Level.INFO, "Received id reply for {0}\n rn={1} with id={2}", new Object[]{job, requestNumber, jobID});
            setDefaults(job);
            provideInput(job);
            assign(jobToAss.getUserJob(), jobToAss.getMaster());
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, null, ex);
        }
    }

    private void primaryUpdate(String jobID, String primaryID, Double primaryLoad) {
        assignedJob.get(jobID).primaryID = primaryID;
    }

    private void assign(UserJob userJob, NodeInfo master) {
        //TODO for better load balancing select the current lower load master
        NodeInfo selectedMaster = master;
        LOG.log(Level.INFO, "Selected master for {0} is :{1}", new Object[]{userJob.getJob().getJobID(), selectedMaster.getID()});
        MessageEvent submitJobMsg = new MessageEvent(EventType.JOB_ASSIGNED, userJob.getJob(), networkModule.getNodeInfo());
        try {
            LOG.log(Level.INFO, "Send assign message for {0} to {1}", new Object[]{userJob, master.getID()});
            networkModule.send(selectedMaster, submitJobMsg);
            userJob.setPrimaryID(master.getID());
            assignedJob.put(userJob.getJob().getJobID().toString(), userJob);
            LOG.log(Level.INFO, "Job {0} assigned to {1}", new Object[]{userJob.getJob().getJobID(), selectedMaster.getID()});
        } catch (IOException ex) {
            ex.printStackTrace();
            LOG.log(Level.SEVERE, null, ex);
        }
    }

    private void provideInput(Job job) throws FileNotFoundException, IOException {

        LOG.log(Level.INFO, "Providing input on ftp");
        FTPManager ftpM = networkModule.getFtpManager();
        String home = ftpM.getBaseDir();
        FileSystemView fsv = new FileSystemView(home);
        String jobDir = job.getJobID().toString() + "/";

//        TextFileInputFormat.addInputPath(job, jobDir);

        String paths = TextFileInputFormat.getLocalInputPath(job);
        String pathsArr[] = paths.split(",");
        fsv.mkdirs(jobDir + "input/");
        String inputDir = fsv.createPath(jobDir + "input/");
        for (String path : pathsArr) {
            File f = new File(path);
            copyFile(f, inputDir);
        }

        fsv.mkdir(jobDir + "jars/");
        String jarsDir = fsv.createPath(jobDir + "jars/");
        String jar = job.getJar();
        copyFile(new File(jar), jarsDir);
        LOG.log(Level.INFO, "Input provided");

    }

    /**
     * Configure all the job directory path
     * @param job
     */
    private void setDefaults(Job job) {

        Configuration conf = job.getConfiguration();

        String jobDir = job.getJobID().toString() + "/";
        conf.set(MRJobConfig.JOB_DIR_ATTR, jobDir);
        conf.set(MRJobConfig.JOB_INPUT_DIR_ATTR, jobDir + "input/");
        conf.set(MRJobConfig.JOB_INPUTSPLIT_DIR_ATTR, jobDir + "inputSplits/");
        conf.set(MRJobConfig.JOB_OUTPUT_DIR_ATTR, jobDir + "output/");
        conf.set(MRJobConfig.JOB_JAR_DIR_ATTR, jobDir + "jars/");

        if (conf.get(MRJobConfig.TASK_TEMP_DIR) == null) {
            conf.set(MRJobConfig.TASK_TEMP_DIR, jobDir + "tmp/");
        }
        String mapOutputDir = conf.get(MRJobConfig.MAP_OUTPUT_DIR_ATTR);
        if (mapOutputDir == null) {
            conf.set(MRJobConfig.MAP_OUTPUT_DIR_ATTR, conf.get(MRJobConfig.TASK_TEMP_DIR) + "map/output/");
        }
        String reduceInputDir = conf.get(MRJobConfig.REDUCE_INPUT_DIR_ATTR);
        if (reduceInputDir == null) {
            conf.set(MRJobConfig.REDUCE_INPUT_DIR_ATTR, conf.get(MRJobConfig.TASK_TEMP_DIR) + "reduce/input/");
        }
        String reduceOutputDir = conf.get(MRJobConfig.REDUCE_OUTPUT_DIR_ATTR);
        if (reduceOutputDir == null) {
            //XXX for now the same of job output dir
            conf.set(MRJobConfig.REDUCE_OUTPUT_DIR_ATTR, conf.get(MRJobConfig.JOB_OUTPUT_DIR_ATTR));
        }
    }

    /**
     * Explores the net every {@link #SEARCH_PERIOD} ms to find
     * a master node.
     * 
     * @return The lowest load master node
     */
    private NodeInfo waitMaster() {
        NodeInfo master = getLowestLoadMaster();
        while (master == null) {
            try {
                Thread.sleep(MASTER_SEARCH_PERIOD);
                master = getLowestLoadMaster();
            } catch (InterruptedException ex) {
                LOG.log(Level.SEVERE, null, ex);
            }
        }
        return master;
    }

    private NodeInfo getLowestLoadMaster() {
        Set<NodeInfo> allMasters = networkModule.getAllMasters();
        NodeInfo ris = null;
        double minLoad = Double.MAX_VALUE;
        for (NodeInfo nodeInfo : allMasters) {
            if (nodeInfo.getLoad() < minLoad) {
                ris = nodeInfo;
                minLoad = nodeInfo.getLoad();
            }
        }

        return ris;
    }

    private void retrieveReduceOutput(JobTask jt, NodeInfo primaryNodeInfo) {
        Job job = assignedJob.get(jt.getTask().getJobID()).getJob();
        LOG.log(Level.INFO, "Reducer {0} for job {1} completed. Retrieving output", new Object[]{jt.getTask().getTaskID().getId(), job.getJobID()});
        FTPManager ftpM = networkModule.getFtpManager();
        Configuration configuration = jt.getTask().getConf();
        boolean success = false;

        NodeInfo slaveNode = networkModule.find(jt.getSlaveID());
        if (slaveNode != null) {
            String srcFilePath = configuration.get(MRJobConfig.REDUCE_OUTPUT_DIR_ATTR);
            srcFilePath += "output_" + jt.getTask().getTaskID().getId();

            String destFilePath = configuration.get(MRJobConfig.JOB_OUTPUT_DIR_ATTR);
            destFilePath += "output_" + jt.getTask().getTaskID().getId();

            success = ftpM.downloadFile(slaveNode, srcFilePath, destFilePath);
        } else {
            LOG.log(Level.SEVERE, "Unable to retrieve node info for " + slaveNode);
        }

        MessageEvent me = null;
        if (success) {
            me = new MessageEvent(EventType.OUTPUT_RETRIEVING_COMPLETED, jt, true);
        } else {
            me = new MessageEvent(EventType.OUTPUT_RETRIEVING_COMPLETED, jt, false);
        }
        try {
            networkModule.send(primaryNodeInfo, me);
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, null, ex);
        }
    }

    private void jobCompleted(String jobID, State state) {
        //TODO notify mechanism to inform user. Use EventListener ?? 
        UserJob userJob = assignedJob.remove(jobID);
        Job job = userJob.getJob();
        LOG.log(Level.ALL, "Job {0} completed with status {1}", new Object[]{job.getJobID(), state});
        if (state == State.SUCCEEDED) {
            Runnable jobCompletionAction = userJob.getJobCompletionAction();
            if (jobCompletionAction != null) {
                LOG.log(Level.INFO, "Starting a new thread to execute completion action for " + job.getJobID());
                new Thread(jobCompletionAction).start();
            }
        }
    }

    private class UserNodeEventListener implements EventListener {

        public UserNodeEventListener() {
        }

        public void handleEvent(EventIF event) {
            LOG.log(Level.INFO, "Received new event: {0}", event);
            switch (event.getType()) {
                case JOB_ID_REPLY:
                    jobIDReply((Integer) event.getParam(0), (JobID) event.getParam(1));
                    break;
                case PRIMARY_UPDATE:
                    primaryUpdate((String) event.getParam(0), (String) event.getParam(1), (Double) event.getParam(2));
                    break;
                case REDUCE_COMPLETED:
                    retrieveReduceOutput((JobTask) event.getParam(0), (NodeInfo) event.getParam(1));
                    break;
                case JOB_COMPLETED:
                    jobCompleted((String) event.getParam(0), (JobStatus.State) event.getParam(1));
                    break;
                default:
                    ;
            }

        }
    }

    /**
     * A thread that manages jobs submission to master node
     */
    private class JobSubmitter extends Thread {

        public JobSubmitter() {
        }

        @Override
        public void run() {
            while (true) {
                UserJob nextJob = null;
                lock.lock();
                try {
                    while (subittedJob.isEmpty()) {
                        try {
                            notEmpty.await();
                        } catch (InterruptedException ex) {
                            LOG.log(Level.SEVERE, null, ex);
                        }
                    }

                    nextJob = subittedJob.removeFirst();
                    LOG.log(Level.INFO, "Next job to handle: " + nextJob);

                } finally {
                    lock.unlock();
                }

                //Wait to find a master for starting job submission
                NodeInfo master = waitMaster();
                //Starting job submission by requesting a new job id
                requestId(nextJob, master);
            }
        }
    }
}
