/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule.event;

import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.JobID;
import p2pMapReduce.nodeModule.JobManager;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.entity.BackupJob;
import p2pMapReduce.entity.JobTask;
import p2pMapReduce.entity.Task.TaskStatus;

/**
 * Enumeration of all the well known event type for the P2PMapReduce framework.
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public enum EventType {

    /**
     * Associated parameters array:
     * 0) String: id of the coordinator which sends the event
     */
    BECOME_MASTER,
    /**
     *
     * No parameters
     */
    BECOME_SLAVE,
    /**
     * Used by user node to request a new unique job id to
     * a master node.
     * 
     * Associated parameters array:
     * 0) Integer: user request number
     * 1) {@link NodeInfo: node info of user node
     */
    JOB_ID_REQUEST,
    /**
     * Used by a master node to respond to an user job id request.
     *
     * Associated parameters array:
     * 0) Integer: user request number
     * 1) {@link JobID}:  an unique job id
     */
    JOB_ID_REPLY,
    /**
     * Used to assign a new job to a master node.
     * 
     * Associated parameters array:
     * 0) {@link Job}: the assigned job object
     * 1) {@link NodeInfo}: Node info object of the user node that has submitted
     *   the job
     */
    JOB_ASSIGNED,
    /**
     * Used to reassign a job to a backup master node in consequence of the
     * primary failure
     *
     * Associated parameters array:
     * 0) {@link BackupJob}: The backup job object representing the job
     */
    JOB_REASSIGNED,
    /**
     * Used to notify the completion of a job.
     *
     * Associated parameters array:
     * 0) String: id of the completed job
     * 1) {@link JobStatus.State}: an object representation of the completion
     *    state
     */
    JOB_COMPLETED,
    /**
     * Used by primary node to assign a backup job to a master node.
     *
     * Associated parameters array:
     * 0) {@link PrimaryJob}: an object representing the job for which the backup
     *    job is assigned
     * 1) String: id of the primary node
     */
    BACKUP_JOB_ASSIGNED,
    /**
     * Used by primary nodes to sends job updates information to the backup master.
     *
     * Associated parameters array:
     * 0) String: id of the job
     * 1) {@code LinkedList<String>}: updated backup master list
     * 2) {@code LinkedList<JobTask>}: updated job task list
     * 
     */
    JOB_UPDATE,
    /**
     * Used by primary node to notify user that a reduce task has been completed.
     *
     * Associated parameters array:
     * 0) {@link JobTask}: the job task object relative to the completed reduce
     *    task
     * 1) {@link NodeInfo}: node info of the primary node
     *
     */
    REDUCE_COMPLETED,
    /**
     * Used by user node to notify the {@link JobManager} about
     * the success or failure of the reduce output retrieving.
     * 
     * Associated parameters array:
     * 0 {@link JobTask}: JobTask relative to the reduce task
     * 1 {@link Boolean}: true if the output has been retrieved
     * successfully, false otherwise
     */
    OUTPUT_RETRIEVING_COMPLETED,
    /**
     * Used to notify the completion of a backup job
     * 
     * Associated parameters array:
     * 0) String: id of the job for which the backup job has been completed
     */
    BACKUP_JOB_COMPLETED,
    /**
     * Used by a primary node to assign a task list to a slave node.
     * 
     * Associated parameters array:
     * 0) {@code LinkedList<Task>}: assigned task list
     * 1) String: id of the primary node
     */
    TASK_ASSIGNED,
    /**
     * Used by a slave node to notify the primary about a task completion.
     *
     * Associated parameters array:
     * 0) String: id of the task attempt
     * 1) {@link TaskStatus}: task status
     * 
     */
    TASK_COMPLETED,
    /**
     * Used by a slave node as an heartbeat message to inform a primary node
     * that a task it is managing is still alive
     *
     * Associated parameters array:
     * 0) String: id of the task
     */
    SLAVE_CHECK,
    /**
     * Coordinator Election Algorithm message
     *
     * Associated parameters array:
     * 0) String: the id of the node
     */
    COORD_ELECTION_MESSAGE,
    /**
     * Sent from the new coordinator to other masters.
     *
     * Associated parameters array:
     * 0) String: id of the new coordinator
     */
    NEW_COORDINATOR,
    /**
     * Inner event
     * Used to notify the failure of a primary node.
     *
     * Associated parameters array:
     * 0) String: id of the failed primary node
     */
    PRIMARY_FAILURE,
    /**
     * Primary Election Algorithm message
     *
     * Associated parameters array:
     * 0) String: id of the job
     * 1) String: id of the node
     * 2) Double: load value of the node
     *
     */
    PRIMARY_ELECTION_MESSAGE,
    /**
     * Used by the newly elected primary to inform all other interested nodes
     * that it has become the new primary node for a specified job
     *
     * Associated parameters array:
     * 0) String: id of the job for which a new primary has been elected.
     * 1) String: id of the new primary node for the job
     * 2) Double: load value of the new primary node
     */
    PRIMARY_UPDATE,
    /**
     * InnerEvent.
     * Used to notify the failure of the coordinator.
     * No parameters (the coordinator id must be already known)
     */
    COORDINATOR_FAILURE,
    /**
     * InnerEvent.
     * Used to notify the failure of a slave node.
     * 0) String: id of failed slave node
     */
    SLAVE_FAILURE,
    /**
     *
     * InnerEvent.
     * Used to notify the failure of a backup master node.
     * 0) String: id of failed backup master node
     */
    BACKUP_MASTER_FAILURE,
    //Low level event (network event)
    //TODO move to network module package
    /**
     * Inner event.
     * Low level event that notified a failure of a remote node.
     * Associated parameters array:
     * 0) String: id of the failed node
     */
    NODE_FAILURE,
    /**
     * Inner event
     * Creation of a new connection
     * Associated parameters array:
     * 0) The connection object (ConnectionIF)
     */
    NEW_CONNECTION,
    /**
     *
     * Crash of a connection
     */
    CRASH_EVENT,
    /**
     *
     * Closing of a connection
     */
    CLOSE_EVENT,
    /**
     *
     * Used in connection start up to send the local node info
     * to the remote endpoint
     */
    REMOTE_NODEINFO,
}
