/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule.nodeFsm;

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.mapreduceModule.JobStatus;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.nodeModule.TaskManager;
import p2pMapReduce.entity.SlaveTask;
import p2pMapReduce.entity.Task;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.MessageEvent;
import p2pMapReduce.util.LoggingUtil;

/**
 * Inner state of Slave macro state
 * A slave node remains in this state until all the job for the task submitted
 * to this node are completed
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
//TODO refactoring: accorpare tutte gli stati dello slave in un unico macro stato slave
//o meglio estendere super stato slave
public class Active extends NodeState {

    private static final transient Logger LOG = Logger.getLogger(Active.class.getName());
    NodeFSM nodeFSM;

    public Active(NodeFSM aThis) {
        this.nodeFSM = aThis;
    }

    @Override
    protected void entryAction() {
        updateLoad(nodeFSM.getNodeModule().getSlaveTaskMap().size());
    }

    @Override
    public void taskAssigned(LinkedList<Task> assignedTaskList, String primaryID) {
        LOG.log(Level.INFO, "New task list assigned: {0} from {1}", new Object[]{assignedTaskList, primaryID});
        for (Task t : assignedTaskList) {
            SlaveTask st = nodeFSM.getNodeModule().getSlaveTask(t.getTaskAttIDAsString());
            if (st != null) {
                LOG.log(Level.SEVERE, "ERROR Task with ID:{0} already exists with slave task:{1}.Stopping the old one", new Object[]{t.getTaskAttIDAsString(), st});
                nodeFSM.getNodeModule().removeSlaveTask(t.getTaskAttIDAsString());
                TaskManager tm = nodeFSM.getNodeModule().removeTaskManager(t.getTaskAttIDAsString());
                tm.stop();
            }
            st = new SlaveTask(t, primaryID);
            TaskManager tm = new TaskManager(nodeFSM, st);
            nodeFSM.getNodeModule().addSlaveTask(st);
            nodeFSM.getNodeModule().addTaskManager(t.getTaskAttIDAsString(), tm);
            tm.start();
        }
        updateLoad(assignedTaskList.size());
        LOG.log(Level.INFO, "TaskManagers started.");
    }

    @Override
    public void taskCompleted(String taskAttID, Task.TaskStatus taskStatus) {
        LOG.log(Level.INFO, "Task {0} completed with status {1}", new Object[]{taskAttID, taskStatus});
        //XXX only task manager are removed on task completion
        //the slave task will be removed on job completion to make the slave node
        //able to comunicate the task completion state to a new primary master
        //in the future
        //XXX by not removing slave tasks the node remaining in the active state
        //until the job is completed. Not bad
        //TODO controllare se possibile concurrent modification exception
        SlaveTask st = nodeFSM.getNodeModule().getSlaveTask(taskAttID);
        TaskManager tm = nodeFSM.getNodeModule().removeTaskManager(taskAttID);
        tm.stop();

        String primaryID = st.getPrimaryID();
        if (primaryID == null) {
            LOG.log(Level.INFO, "Failed primary for task " + taskAttID + ". No notification to primary is possible");
        } else {
            try {
                MessageEvent taskCompletedMsg = new MessageEvent(EventType.TASK_COMPLETED, taskAttID, taskStatus);
                LOG.log(Level.INFO, "Propagate event to primary " + primaryID);
                nodeFSM.getNodeModule().getNetworkModule().send(primaryID, taskCompletedMsg);
            } catch (IOException ex) {
                ex.printStackTrace();
                LOG.log(Level.SEVERE, null, ex);
            }
        }

        updateLoad(-1);

        if (nodeFSM.getNodeModule().getSlaveTaskMap().isEmpty()) {
            LOG.log(Level.INFO, "No slave task remainig.Transition to SLAVE.IDLE");
            nodeFSM.transition(nodeFSM.IDLE);
        } else {
            LOG.log(Level.INFO, "Task remaining:\n{0}", LoggingUtil.elementByLine(nodeFSM.getNodeModule().getSlaveTaskMap().values()));
        }
    }

    @Override
    public void jobCompleted(String jobID, JobStatus.State jobState) {
        LOG.log(Level.INFO, "Job completion received for " + jobID + " Removing all the associated slave tasks");
        //Remove all the slave task associated to this job
        Set<Entry<String, SlaveTask>> entrySet = nodeFSM.getNodeModule().getSlaveTaskMap().entrySet();
        for (Iterator<Entry<String, SlaveTask>> it = entrySet.iterator(); it.hasNext();) {
            Entry<String, SlaveTask> entry = it.next();
            if (entry.getValue().getTask().getJobID().equals(jobID)) {
                it.remove();
            }
        }

        if (nodeFSM.getNodeModule().getSlaveTaskMap().isEmpty()) {
            LOG.log(Level.INFO, "No slave task remainig.Transition to SLAVE.IDLE");
            nodeFSM.transition(nodeFSM.IDLE);
        } else {
            LOG.log(Level.INFO, "Task remaining:\n{0}", LoggingUtil.elementByLine(nodeFSM.getNodeModule().getSlaveTaskMap().values()));
        }

    }

    @Override
    public void primaryMasterFailure(String failedPrimaryID) {
        LOG.log(Level.INFO, "PrimaryMaster " + failedPrimaryID + " failed. Propagating event to interested TaskManagers");
        for (SlaveTask st : nodeFSM.getNodeModule().getSlaveTaskMap().values()) {
            if (failedPrimaryID.equals(st.getPrimaryID())) {
                //FIXME non c'è bisogno di utilizzare un inner event ?
                TaskManager tm = nodeFSM.getNodeModule().getTaskManager(st.getTask().getTaskAttIDAsString());
                if (tm != null) {
                    tm.primaryMasterFailure(failedPrimaryID);
                } else {
                    LOG.log(Level.INFO, "PrimaryMaster " + failedPrimaryID + " failed.TaskManager for slave task" + st + " not exists");
                }
            }
        }
    }

    @Override
    public void primaryUpdate(String updatedJobID, String primaryNodeID, double updatedPrimaryLoad) {
        LOG.log(Level.INFO, "PrimaryUpdate received for job:{0} with primary:{1}.Propagating event to interested TaskManagers.", new Object[]{updatedJobID, primaryNodeID});
        for (SlaveTask st : nodeFSM.getNodeModule().getSlaveTaskMap().values()) {
            if (st.getTask().getJobID().equals(updatedJobID)) {
                //FIXME non c'è bisogno di utilizzare un inner event ?
//                InnerEvent ie = new InnerEvent(EventType.PRIMARY_UPDATE, updatedJobID,primaryNodeID);
                TaskManager tm = nodeFSM.getNodeModule().getTaskManager(st.getTask().getTaskAttIDAsString());
                if (tm != null) {
                    tm.primaryUpdate(updatedJobID, primaryNodeID, updatedPrimaryLoad);
                } else {
                    LOG.log(Level.INFO, "TaskManager for slave task" + st + " not exists. Send completion event (" + st.getTaskStatus() + ") to new primary");
                    MessageEvent taskCompletedMsg = new MessageEvent(EventType.TASK_COMPLETED, st.getTaskID(), st.getTaskStatus());
                    try {
                        nodeFSM.getNodeModule().getNetworkModule().send(primaryNodeID, taskCompletedMsg);
                    } catch (IOException ex) {
                        LOG.log(Level.SEVERE, null, ex);
                    }

                }
            }
        }
    }

    private void updateLoad(int numTasks) {
        NodeInfo thisNode = nodeFSM.getNodeModule().getNodeInfo();
        thisNode.setLoad(thisNode.getLoad() + numTasks);
        LOG.log(Level.INFO, "Current load={0}", thisNode.getLoad());
        nodeFSM.getNetworkModule().publish();
    }
}
