/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule.nodeFsm;

import java.io.IOException;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.MessageEvent;
import p2pMapReduce.util.P2PMapReduceConfig;

/**
 * Inner state of {@link Master} macro state, that takes care
 * of master/slave ratio maintenance
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class CoordinationFSM extends FSM {

    private static final transient Logger LOG = Logger.getLogger(CoordinationFSM.class.getName());
    private static final String CHECKPERIOD_ATTR = "coordinator.checkperiod";
    private static final String MASTER_SLAVE_RATIO_ATTR = "coordinator.master.ratio";
    private static final int ELECTION_TIME;
    /**
     * Coordinator check period
     */
    private static final int CHECK_PERIOD;
    /**
     * 1 indicates as many masters as slaves
     * 0 indicates only one master
     * Values between 0 and 1 indicate the master/slave ratio
     */
    public static final double MASTER_SLAVE_RATIO;

    static {
        //FIXME made configurable ??
        ELECTION_TIME = 1000 * 5;

        Properties p = P2PMapReduceConfig.getConfiguration();
        CHECK_PERIOD = 1000 * Integer.parseInt(p.getProperty(CHECKPERIOD_ATTR, "180"));
        MASTER_SLAVE_RATIO = Double.parseDouble(p.getProperty(MASTER_SLAVE_RATIO_ATTR, "0.2"));

    }
    final CoordinationState NOT_COORDINATOR;
    final CoordinationState WAITING_COORDINATOR;
    final CoordinationState ELECTING_COORDINATOR;
    final CoordinationState COORDINATOR;
    final CoordinationState ACTIVE;
    final CoordinationState IDLE;
    private NodeFSM nodeFSM;
    private MessageEvent newCoorinatorMsg;
    private MessageEvent electionMsg;
    //TODO valutare utilizzo ScheduledThreadPoolExecutor
    Timer timer = new Timer("CoordinationFSM Timer", true);
    TimeoutTask timeoutTask = new TimeoutTask();

    public CoordinationFSM(NodeFSM nodeFSM) {
        this.nodeFSM = nodeFSM;
        NOT_COORDINATOR = new NotCoordinator();
        WAITING_COORDINATOR = new WaitingCoordinator();
        ELECTING_COORDINATOR = new ElectingCoordinator();
        COORDINATOR = new Coordinator();
        ACTIVE = new Active();
        IDLE = new Idle();
        newCoorinatorMsg = new MessageEvent(EventType.NEW_COORDINATOR, nodeFSM.getNodeModule().getNodeID());
        electionMsg = new MessageEvent(EventType.COORD_ELECTION_MESSAGE, nodeFSM.getNodeModule().getNodeID());

    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM methods
    ///////////////////////////////////////////////////////////////////////////
    //XXX Devono essere tutti synchronized per mutua esclusione tra eventi interni
    //e eventi esterni
    public synchronized void timeout(Object param) {
        ((CoordinationState) currentState()).timeout(param);
    }

    public synchronized void coordinatorFailure() {
        ((CoordinationState) currentState()).coordinatorFailure();
    }

    public synchronized void electionMessage(String senderID) {
        ((CoordinationState) currentState()).electionMessage(senderID);
    }

    public synchronized void newCoordinator(String senderID) {
        ((CoordinationState) currentState()).newCoordinator(senderID);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Private methods
    ///////////////////////////////////////////////////////////////////////////
    private boolean lowerThan(String otherID) {
        NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();
        String nodeID = netModule.getNodeInfo().getID();
        return nodeID.compareTo(otherID) < 0;
    }

    private Set<NodeInfo> removeNotIdle(Set<NodeInfo> nodes) {
        for (Iterator<NodeInfo> it = nodes.iterator(); it.hasNext();) {
            NodeInfo nodeInfo = it.next();
            if (nodeInfo.getLoad() > 0) {
                it.remove();
            }
        }

        return nodes;
    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM state classes
    ///////////////////////////////////////////////////////////////////////////
    private abstract class CoordinationState extends NodeState {

        @Override
        protected void exitAction() {
            timeoutTask.cancel();
        }
    }

    private class NotCoordinator extends CoordinationState {

        @Override
        protected void entryAction() {

            //TODO is correct?
            if (nodeFSM.getNodeModule().getCoordinatorID() == null) {
                //TODO critico nelle fasi iniziali quando non si conoscono gli altri master
                //Si potrebbe aspettare random e accetare anche qui messaggi new coordinator
                //oppure passare subito all'elezione
                int nextTimeout = getNextTimeout(ELECTION_TIME, 2 * ELECTION_TIME);
                LOG.log(Level.CONFIG, "Entrering COORDINATOR.NOT_COORDINATOR\nCoordinator not founded. Transition to ELECTING_COORDINATOR after " + nextTimeout);
                //XXX automa modification for testing
                timeoutTask = new TimeoutTask();
                timer.schedule(timeoutTask, nextTimeout);
//                transition(ELECTING_COORDINATOR);
            } else {
                LOG.log(Level.CONFIG, "Entrering COORDINATOR.NOT_COORDINATOR\nCoordinator founded.");
            }

        }

        @Override
        public void newCoordinator(String senderID) {
            timeoutTask.cancel();
            LOG.log(Level.CONFIG, "Received new coordinator id: " + senderID + ".Transition to NOT_COORDINATOR");
            nodeFSM.getNodeModule().setCoordinatorID(senderID);
            transition(NOT_COORDINATOR);
        }

        @Override
        public void coordinatorFailure() {
            LOG.log(Level.CONFIG, "Coordinator failure detected. Transition to ELECTING_COORDINATOR");
            timeoutTask.cancel();
            transition(ELECTING_COORDINATOR);
        }

        //XXX automa modification for testing
        @Override
        public void timeout(Object param) {
            //TODO is correct?
            if (nodeFSM.getNodeModule().getCoordinatorID() == null) {
                LOG.log(Level.CONFIG, "Coordinator not founded. Transition to ELECTING_COORDINATOR");
                transition(ELECTING_COORDINATOR);
            } else {
                LOG.log(Level.CONFIG, "Coordinator founded.");
            }
        }
    }

    private class ElectingCoordinator extends CoordinationState {

        @Override
        protected void entryAction() {
            //Send election message to all master with higher identifier
            LOG.log(Level.CONFIG, "Entering COORDINATOR.ELECTING_COORDINATOR: sending election messages to higher masters");
            NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();
            Set<NodeInfo> masters = netModule.getOtherMasters();
            LOG.log(Level.CONFIG, "Available masters: " + masters);
            for (NodeInfo master : masters) {
                try {
                    LOG.log(Level.CONFIG, "Sending to " + master.getID());
                    netModule.send(master, electionMsg);
                    LOG.log(Level.CONFIG, "Sent");
                } catch (IOException ex) {
                    ex.printStackTrace();
                    LOG.log(Level.SEVERE, null, ex);
                }
            }

            timeoutTask = new TimeoutTask();
            timer.schedule(timeoutTask, ELECTION_TIME);
            LOG.log(Level.CONFIG, "Messages sent. Waitinig for " + ELECTION_TIME);

        }

        @Override
        public void electionMessage(String senderID) {

            boolean higherID = lowerThan(senderID);
            if (higherID) {
                //An election message is received from another master with
                //an higher identifier
                timeoutTask.cancel();
                LOG.log(Level.CONFIG, "Received a response from higher(" + senderID + ").Transition to WAITING_COORDINATOR");
                transition(WAITING_COORDINATOR);
            } else {
                //An election message is received from another master with 
                //a lower identifier
                LOG.log(Level.CONFIG, "Received a response from lower(" + senderID + "). Sending response.");
                NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();
                try {
                    netModule.send(senderID, electionMsg);
                } catch (IOException ex) {
                    ex.printStackTrace();
                    LOG.log(Level.SEVERE, null, ex);

                }
            }
        }

        @Override
        public void newCoordinator(String senderID) {
            timeoutTask.cancel();
            LOG.log(Level.CONFIG, "Received new coordinator id: " + senderID + ".Transition to NOT_COORDINATOR");
            nodeFSM.getNodeModule().setCoordinatorID(senderID);
            transition(NOT_COORDINATOR);
        }

        @Override
        public void timeout(Object param) {
            //NO respose are received.
            //Become new coordinator
            LOG.log(Level.CONFIG, "Becoming the new coordinator");
            transition(COORDINATOR);
        }
    }

    private class WaitingCoordinator extends CoordinationState {

        @Override
        protected void entryAction() {
            LOG.log(Level.CONFIG, "Entering COORDINATOR.WAITING_COORDINATOR");
            timeoutTask = new TimeoutTask();
            timer.schedule(timeoutTask, ELECTION_TIME);
        }

        @Override
        public void newCoordinator(String senderID) {
            timeoutTask.cancel();
            LOG.log(Level.CONFIG, "Received new coordinator id: " + senderID);
            nodeFSM.getNodeModule().setCoordinatorID(senderID);
            transition(NOT_COORDINATOR);
        }

        @Override
        public void timeout(Object param) {
            LOG.log(Level.CONFIG, "Coordinator id not received in time.Transition to ELECTING_COORDINATOR");
            transition(ELECTING_COORDINATOR);
        }
    }

    private class Coordinator extends CoordinationState {

        @Override
        protected void entryAction() {

            //Notify other master nodes that this node is the new coordinator
            LOG.log(Level.CONFIG, "Entering COORDINATOR.NEW_COORDINATOR.\nSending my id as new coordinator to other masters");
            nodeFSM.getNodeModule().setCoordinatorID(nodeFSM.getNodeModule().getNodeInfo().getID());
            NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();
            Set<NodeInfo> masters = netModule.getOtherMasters();
            LOG.log(Level.CONFIG, "Available masters: " + masters);
            for (NodeInfo master : masters) {
                try {
                    LOG.log(Level.FINE, "Sending to " + master.getID());
                    netModule.send(master, newCoorinatorMsg);
                    LOG.log(Level.FINE, "Sent");
                } catch (IOException ex) {
                    ex.printStackTrace();
                    LOG.log(Level.SEVERE, null, ex);
                }
            }
            LOG.log(Level.CONFIG, "Messages sent.Transition to ACTIVE");
            transition(ACTIVE);
        }

        @Override
        public void electionMessage(String senderID) {
            try {
                //Election message received while i'm already a coordinator
                //Send direcly new coordinator msg
                LOG.log(Level.CONFIG, "Received an election message from: " + senderID + ". Sending new coordinator msg");
                nodeFSM.getNodeModule().getNetworkModule().send(senderID, newCoorinatorMsg);
            } catch (IOException ex) {
                ex.printStackTrace();
                LOG.log(Level.SEVERE, null, ex);
            }


        }

        @Override
        public void newCoordinator(String senderID) {
            //New coordinator id received while i'm already a coordinator
            //To assure net convergence leave the higher master to continue as coordinator

            boolean lower = lowerThan(senderID);
            if (lower) {
                timeoutTask.cancel();
                LOG.log(Level.CONFIG, "Received new higher coordinator id: " + senderID + ".Transition to NOT_COORDINATOR");
                nodeFSM.getNodeModule().setCoordinatorID(senderID);
                transition(NOT_COORDINATOR);
            } else {
                LOG.log(Level.CONFIG, "Received new lower coordinator id: " + senderID + ".Sending NEW_COORDINATOR message.");
                try {
                    nodeFSM.getNodeModule().getNetworkModule().send(senderID, newCoorinatorMsg);
                } catch (IOException ex) {
                    ex.printStackTrace();
                    Logger.getLogger(CoordinationFSM.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
    }

    private class Active extends Coordinator {

        @Override
        protected void entryAction() {
            LOG.log(Level.CONFIG, "Entering COORDINATOR.ACTIVE");
            //Check actions
            NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();

            Set<NodeInfo> masters = netModule.getOtherMasters();
            Set<NodeInfo> slaves = netModule.getOtherSlaves();
            double numMasters = masters.size() + 1;//this node is also a master
            double numSlaves = slaves.size();
            double desiredMasters = 1;
            if (MASTER_SLAVE_RATIO > 0) {
                desiredMasters = Math.floor((numMasters + numSlaves) * MASTER_SLAVE_RATIO);
            }
            double delta = desiredMasters - numMasters;
            LOG.log(Level.INFO, "Masters: {0} Slaves: {1} Master %: {2} Desired Master: {3} delta: {4}", new Object[]{numMasters, numSlaves, MASTER_SLAVE_RATIO, desiredMasters, delta});
            if (delta > 0) {
                //more masters required
                LOG.log(Level.CONFIG, delta + " new master required.");
                MessageEvent becomeMasterMsg = new MessageEvent(EventType.BECOME_MASTER, nodeFSM.getNodeModule().getNodeID());
                int count = 0;
                for (NodeInfo slaveNode : removeNotIdle(slaves)) {
                    try {
                        LOG.log(Level.CONFIG, "Sending becomeMaster to " + slaveNode.getID());
                        netModule.send(slaveNode, becomeMasterMsg);
                        count++;
                        if (count == delta) {
                            break;
                        }
                    } catch (IOException ex) {
                        ex.printStackTrace();
                        Logger.getLogger(CoordinationFSM.class.getName()).log(Level.SEVERE, null, ex);
                    }
                }
            } else if (delta < 0) {
                //more slaves required
                delta = -delta;
                LOG.log(Level.CONFIG, delta + " new slaves required.");
                MessageEvent becomeSlaveMsg = new MessageEvent(EventType.BECOME_SLAVE);
                int count = 0;
                for (NodeInfo masterNode : removeNotIdle(masters)) {
                    try {
                        LOG.log(Level.CONFIG, "Sending becomeSlave to " + masterNode.getID());
                        netModule.send(masterNode, becomeSlaveMsg);
                        count++;
                        if (count == delta) {
                            break;
                        }
                    } catch (IOException ex) {
                        ex.printStackTrace();
                        LOG.log(Level.SEVERE, null, ex);
                    }
                }
            } else {
                LOG.log(Level.CONFIG, "New masters or slaves not required");
            }

            transition(IDLE);
        }
    }

    private class Idle extends Coordinator {

        @Override
        protected void entryAction() {

            //Start the timer
            LOG.log(Level.CONFIG, "Entering COORDINATOR.IDLE");
            timeoutTask = new TimeoutTask();
            timer.schedule(timeoutTask, CHECK_PERIOD);
        }

        @Override
        public void timeout(Object param) {
            LOG.log(Level.CONFIG, "timeout");
            transition(ACTIVE);
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Generic inner classes
    ///////////////////////////////////////////////////////////////////////////
    private class TimeoutTask extends TimerTask {

        @Override
        public void run() {
            timeout(null);
        }
    }
}
