/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule.nodeFsm;

import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.mapreduceModule.JobStatus;
import p2pMapReduce.nodeModule.Constants;
import p2pMapReduce.nodeModule.TaskManager;
import p2pMapReduce.entity.SlaveTask;
import p2pMapReduce.entity.Task;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.MessageEvent;

/**
 *
 * Inner state of Slave macro state.
 * A slave node is in this state if it has no task to execute
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class Idle extends NodeState {

    private static final transient Logger LOG = Logger.getLogger(Idle.class.getName());
    private static final int MIN_TIMEOUT = 1 * 60 * 1000;
    private static final int MAX_TIMEOUT = 3 * 60 * 1000;
    //TODO move into NodeState
    NodeFSM nodeFSM;
    Timer timer = new Timer("Idle Timer", true);
    TimeoutTask timeoutTask = new TimeoutTask();

    public Idle(NodeFSM nodeFSM) {
        this.nodeFSM = nodeFSM;
    }

    @Override
    protected void entryAction() {
        int nextTimeout = getNextTimeout(MIN_TIMEOUT, MAX_TIMEOUT);
        LOG.log(Level.INFO, "entryAction: waiting for " + nextTimeout / 1000 / 60 + " min");
        nodeFSM.getNodeModule().getNodeInfo().setRole(Constants.SLAVE);
        nodeFSM.getNodeModule().getNetworkModule().publish();
        nodeFSM.getNodeModule().setCoordinatorID(null);
        timeoutTask = new TimeoutTask();
        timer.schedule(timeoutTask, nextTimeout);

    }

    @Override
    protected void exitAction() {
        timeoutTask.cancel();
    }

    @Override
    public void becomeMaster(String coordID) {
        LOG.log(Level.INFO, "BecomeMaster msg received from: " + coordID + ".Transition to MASTER");
        timeoutTask.cancel();
        nodeFSM.getNodeModule().setCoordinatorID(coordID);
        nodeFSM.transition(nodeFSM.MASTER);
        nodeFSM.getNodeModule().getNetworkModule().becomeSlave();
    }

    @Override
    public void taskAssigned(LinkedList<Task> assignedTaskList, String primaryID) {
        LOG.log(Level.INFO, "New task list assigned: {0} from {1}", new Object[]{assignedTaskList, primaryID});
        timeoutTask.cancel();
        for (Task t : assignedTaskList) {

            SlaveTask st = new SlaveTask(t, primaryID);
            TaskManager tm = new TaskManager(nodeFSM, st);
            nodeFSM.getNodeModule().addSlaveTask(st);
            nodeFSM.getNodeModule().addTaskManager(t.getTaskAttIDAsString(), tm);
            tm.start();
        }
        LOG.log(Level.INFO, "TaskManager started.Transition to SLAVE.ACTIVE");
        nodeFSM.transition(nodeFSM.ACTIVE);
    }

    @Override
    public void primaryUpdate(String updatedJobID, String primaryNodeID, double updatedPrimaryLoad) {
        LOG.log(Level.INFO, "PrimaryUpdate received for job:{0} with primary:{1}.Try to finding slave task.", new Object[]{updatedJobID, primaryNodeID});
        for (SlaveTask st : nodeFSM.getNodeModule().getSlaveTaskMap().values()) {
            if (st.getTask().getJobID().equals(updatedJobID)) {
                LOG.log(Level.INFO, "Slave task founded:{0}. Send completion event ({1}) to new primary", new Object[]{st, st.getTaskStatus()});
                MessageEvent taskCompletedMsg = new MessageEvent(EventType.TASK_COMPLETED, st.getTaskID(), st.getTaskStatus());
                try {
                    nodeFSM.getNodeModule().getNetworkModule().send(primaryNodeID, taskCompletedMsg);
                } catch (IOException ex) {
                    LOG.log(Level.SEVERE, null, ex);
                }

            }
        }
    }

    @Override
    public void jobCompleted(String jobID, JobStatus.State jobState) {
        LOG.log(Level.INFO, "Job completion received for " + jobID + " Removing all the associated slave tasks");
        //Remove all the slave task associated to this job
        Set<Entry<String, SlaveTask>> entrySet = nodeFSM.getNodeModule().getSlaveTaskMap().entrySet();
        for (Iterator<Entry<String, SlaveTask>> it = entrySet.iterator(); it.hasNext();) {
            Entry<String, SlaveTask> entry = it.next();
            if (entry.getValue().getTask().getJobID().equals(jobID)) {
                it.remove();
            }
        }

    }

    @Override
    public void timeout(Object param) {
        LOG.log(Level.INFO, "Transition to CHECK_MASTER");
        nodeFSM.transition(nodeFSM.CHECK_MASTER);
    }

    private class TimeoutTask extends TimerTask {

        @Override
        public void run() {
            timeout(null);
        }
    }
}
