/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule.nodeFsm;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.JobID;
import p2pMapReduce.mapreduceModule.JobStatus;
import p2pMapReduce.mapreduceModule.TaskAttemptID;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.nodeModule.JobManager;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.entity.BackupJob;
import p2pMapReduce.entity.JobTask;
import p2pMapReduce.entity.PrimaryJob;
import p2pMapReduce.entity.Task.TaskStatus;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.MessageEvent;

/**
 * Inner state of {@link Master} macro state, that take care
 * of jobs management
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class ManagementFSM extends FSM {

    private static final transient Logger LOG = Logger.getLogger(ManagementFSM.class.getName());
    final MANAGEMENTState NOT_PRIMARY;
    final MANAGEMENTState PRIMARY;
    private NodeFSM nodeFSM;
    //TODO move to node module???
    /**
     * For each job id holds the relative JobManager
     */
    private HashMap<String, JobManager> jobManagerMap;
    /**
     * A counter atomically incremented. Used to create unique job id
     */
    private int jobCount;
    //TODO valutare utilizzo ScheduledThreadPoolExecutor
    Timer timer = new Timer("ManagementFSM Timer", true);
    TimeoutTask timeoutTask = new TimeoutTask();

    public ManagementFSM(NodeFSM nodeFSM) {
        this.nodeFSM = nodeFSM;
        NOT_PRIMARY = new NotPrimary();
        PRIMARY = new Primary();
        jobManagerMap = new HashMap<String, JobManager>();
    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM methods
    ///////////////////////////////////////////////////////////////////////////
    public synchronized void jobIDRequest(int requestNumber, NodeInfo userNodeInfo) {
        ((MANAGEMENTState) currentState()).jobIDRequest(requestNumber, userNodeInfo);
    }

    public synchronized void jobAssigned(Job job, NodeInfo userNodeInfo) {
        ((MANAGEMENTState) currentState()).jobAssigned(job, userNodeInfo);
    }

    public synchronized void jobReassigned(BackupJob bJ) {
        ((MANAGEMENTState) currentState()).jobReassigned(bJ);
    }

    public synchronized void slaveCheck(String taskID) {
        ((MANAGEMENTState) currentState()).slaveCheck(taskID);
    }

    /**
     * In PRIMARY a node can receive a taskCompleted message that must be
     * propagate to the relative JobManager
     *
     * @param taskID
     * @param taskStatus
     */
    public synchronized void taskCompleted(String taskID, TaskStatus taskStatus) {
        ((MANAGEMENTState) currentState()).taskCompleted(taskID, taskStatus);
    }

    public synchronized void outputRetrieved(JobTask reduceJT, boolean completionState) {
        ((MANAGEMENTState) currentState()).outputRetrieved(reduceJT, completionState);
    }

    public synchronized void jobCompleted(String jobID, JobStatus.State jobState) {
        ((MANAGEMENTState) currentState()).jobCompleted(jobID, jobState);
    }

    public synchronized void slaveFailure(String failedSlaveID) {
        ((MANAGEMENTState) currentState()).slaveFailure(failedSlaveID);
    }

    public synchronized void backupMasterFailure(String failedBackupID) {
        ((MANAGEMENTState) currentState()).backupMasterFailure(failedBackupID);
    }

    public synchronized void timeout(Object param) {
        ((MANAGEMENTState) currentState()).timeout(param);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Private methods
    ///////////////////////////////////////////////////////////////////////////
    private String retrieveJobIDOf(String taskAttID) {
        //FIXME necessario un metodo per recuperare l'id del job al quale il task appartiene in modo più efficiente
        TaskAttemptID taskAttemptID = TaskAttemptID.forName(taskAttID);
        return taskAttemptID.getJobID().toString();

    }

    private void startJobManager(PrimaryJob pJ, boolean isReassigned) {
        JobManager jM = new JobManager(nodeFSM, pJ, isReassigned);
        jobManagerMap.put(pJ.getJobID(), jM);
        jM.start();
    }

    /**
     * Per ora si utilizza per assicurarsi che gli
     * id dei job gestiti siano univoci
     * @param job
     */
    private void checkID(Job job) {
        String jobID = job.getJobID().toString();
        for (PrimaryJob pj : nodeFSM.getNodeModule().getPrimaryJobList()) {
            if (jobID.equals(pj.getJobID())) {
                throw new IllegalStateException("Received a job with not unique id. Job: " + job + "\n equals to \n" + pj);
            }
        }
        for (BackupJob bj : nodeFSM.getNodeModule().getBackupJobList()) {
            if (jobID.equals(bj.getJobID())) {
                throw new IllegalStateException("Received a job with not unique id. Job: " + job + "\n equals to \n" + bj);
            }
        }
    }

    private void updateLoad(int numTasks) {
        NodeInfo thisNode = nodeFSM.getNodeModule().getNodeInfo();
        //XXX every primary job is equals to 2 backup jobs
        thisNode.setLoad(thisNode.getLoad() + 2 * numTasks);
        LOG.log(Level.INFO, "Current load=" + thisNode.getLoad());
        nodeFSM.getNetworkModule().publish();
    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM state classes
    ///////////////////////////////////////////////////////////////////////////
    private class MANAGEMENTState extends NodeState {

        @Override
        public void jobIDRequest(int requestNumber, NodeInfo userNodeInfo) {

            JobID jobID = new JobID(nodeFSM.getNodeModule().getNodeID(), jobCount);
            LOG.log(Level.INFO, "New job id requested " + jobID);
            jobCount++;

            MessageEvent jobIDReply = new MessageEvent(EventType.JOB_ID_REPLY, requestNumber, jobID);
            try {
                nodeFSM.getNetworkModule().send(userNodeInfo, jobIDReply);
            } catch (IOException ex) {
                LOG.log(Level.SEVERE, ex.getLocalizedMessage(), ex);
            }
        }

        @Override
        public void jobAssigned(Job job, NodeInfo userNodeInfo) {
            LOG.log(Level.INFO, "New job assigned: " + job + ".Transition to MASTER.PRIMARY");
//            JobID jobID = new JobID(nodeFSM.getNodeModule().getNodeID(), jobCount);
//            jobCount++;
//            job.setJobID(jobID);
            checkID(job);
            PrimaryJob pj = new PrimaryJob(job, userNodeInfo);
            nodeFSM.getNodeModule().addPrimaryJob(pj);
            startJobManager(pj, false);
            updateLoad(1);
        }

        @Override
        public void jobReassigned(BackupJob bJ) {
            LOG.log(Level.INFO, "New job reassigned: " + bJ);
            checkID(bJ.getJob());
            //TODO controllare che nella backupMasterList non ci sia anche questo stesso nodo
            PrimaryJob pJ = new PrimaryJob(bJ);
            nodeFSM.getNodeModule().addPrimaryJob(pJ);
            String jobID = bJ.getJob().getJobID().toString();
            String nodeID = nodeFSM.getNodeModule().getNodeID();
            NetworkModuleIF netModule = nodeFSM.getNodeModule().getNetworkModule();
            //XXX for primary election algorithm needs this primaryUpdate msg are already sent
            //by the BackupJobManager which has wone the election. 
//            MessageEvent primaryUpdateMsg = new MessageEvent(EventType.PRIMARY_UPDATE, jobID, nodeID, 0);
//            LOG.log(Level.INFO, "Sending update msg to BackupMaster nodes");
//            for (String backupMasterID : pJ.getBackupMasterList()) {
//                try {
//                    netModule.send(backupMasterID, primaryUpdateMsg);
//                } catch (IOException ex) {
//                    ex.printStackTrace();
//                    Logger.getLogger(MANAGEMENTFSM.class.getName()).log(Level.SEVERE, null, ex);
//                }
//            }

            //XXX These msg can be sent from here using a 0 dummy value for 3rd parameter
            MessageEvent primaryUpdateMsg = new MessageEvent(EventType.PRIMARY_UPDATE, jobID, nodeID, 0D);
            LOG.log(Level.INFO, "Sending update msg to slave nodes");
            //XXX Sending primary update for all not failed tasks in order to
            //create a connection towards the slave node and so make the job manager
            //to be notified of slave crash or make the job manager able to know
            //about node crashed during the election
            //Used to trace already notified slaves
            Set<String> notifiedSlave = new HashSet<String>();
            //Used to track the slave nodes for which send failed and so to consider
            //as failed
            Set<String> failedSlaves = new HashSet<String>();
            for (JobTask jobTask : pJ.getJobTaskList()) {
                try {
                    if (jobTask.getTaskStatus() != TaskStatus.FAILED && !notifiedSlave.contains(jobTask.getSlaveID())) {
                        notifiedSlave.add(jobTask.getSlaveID());
                        netModule.send(jobTask.getSlaveID(), primaryUpdateMsg);
                    }
                } catch (IOException ex) {
                    //TODO consider all tasks from this slave failed and blacklist the node
                    failedSlaves.add(jobTask.getSlaveID());
                    LOG.log(Level.SEVERE, null, ex);
                }
            }

            for (JobTask jobTask : pJ.getJobTaskList()) {
                if (jobTask.getTaskStatus() != TaskStatus.FAILED && failedSlaves.contains(jobTask.getSlaveID())) {
                    jobTask.setTaskStatus(TaskStatus.FAILED);
                }
            }
            LOG.log(Level.INFO, "Sending update msg to user node");
            NodeInfo userNodeInfo = pJ.getUserNodeInfo();
            try {
                netModule.send(userNodeInfo, primaryUpdateMsg);
            } catch (IOException ex) {
                LOG.log(Level.SEVERE, null, ex);
            }

            startJobManager(pJ, true);
            updateLoad(1);
        }

        @Override
        protected void exitAction() {
            timeoutTask.cancel();
        }
    }

    private class NotPrimary extends MANAGEMENTState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, "Entering MANAGEMENT.NOT_PRIMARY");
        }

        @Override
        public void jobAssigned(Job job, NodeInfo userNodeInfo) {
            super.jobAssigned(job, userNodeInfo);
            LOG.log(Level.INFO, "Transition to MANAGEMENT.PRIMARY");
            transition(PRIMARY);
        }

        @Override
        public void jobReassigned(BackupJob bJ) {
            super.jobReassigned(bJ);
            LOG.log(Level.INFO, "Transition to MANAGEMENT.PRIMARY");
            transition(PRIMARY);
        }
    }

    private class Primary extends MANAGEMENTState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, "Entering MANAGEMENT.PRIMARY");
        }

        @Override
        public void jobAssigned(Job job, NodeInfo userNodeInfo) {
            super.jobAssigned(job, userNodeInfo);
        }

        @Override
        public void jobReassigned(BackupJob bJ) {
            super.jobReassigned(bJ);
        }

        @Override
        public void slaveCheck(String taskID) {
            LOG.log(Level.INFO, "Received slave check for task " + taskID);
            String jobID = retrieveJobIDOf(taskID);
            JobManager jm = jobManagerMap.get(jobID);
            jm.slaveCheck(taskID);
        }

        @Override
        public void slaveFailure(String failedSlaveID) {
            //FIXME controllare se efficiente
            LOG.log(Level.INFO, "Slave failure event received for slave " + failedSlaveID + ".Propagate it to the interested JobManagers");
            Set<String> jobIDs = new HashSet<String>();
            //First find the jobs ids of the JobTasks relative to the failed node
            for (PrimaryJob primaryJob : nodeFSM.getNodeModule().getPrimaryJobList()) {
                for (JobTask jobTask : primaryJob.getJobTaskList()) {
                    if (failedSlaveID.equals(jobTask.getSlaveID())) {
                        jobIDs.add(primaryJob.getJobID());
                        //for this job the failure will be notified
                        //so go on
                        continue;
                    }
                }
            }
            //Notify to JobManagers relative to the founded jobs
            for (String jobID : jobIDs) {
                JobManager jobM = jobManagerMap.get(jobID);
                jobM.slaveFailure(failedSlaveID);
            }

        }

        @Override
        public void backupMasterFailure(String failedBackupID) {
            LOG.log(Level.INFO, "BackupMaster " + failedBackupID + " failed.Propagate it to the interested JobManagers");
            Set<String> jobIDs = new HashSet<String>();
            //First find the jobs ids of the JobTasks relative to the failed node
            for (PrimaryJob primaryJob : nodeFSM.getNodeModule().getPrimaryJobList()) {
                if (primaryJob.getBackupMasterList().contains(failedBackupID)) {
                    jobIDs.add(primaryJob.getJobID());
                }
            }
            //Notify to JobManager relative to the founded job
            for (String jobID : jobIDs) {
                JobManager jobM = jobManagerMap.get(jobID);
                jobM.backupMasterFailure(failedBackupID);
            }
        }

        /**
         * In PRIMARY a node can receive a taskCompleted message that must be
         * propagate to the relative JobManager
         * 
         * @param taskAttID
         * @param taskStatus
         */
        @Override
        public void taskCompleted(String taskAttID, TaskStatus taskStatus) {
            String jobID = retrieveJobIDOf(taskAttID);
            LOG.log(Level.INFO, "TaskCompleted event received for task " + taskAttID + " with task status " + taskStatus + ".Propagate it to JobManager for the job " + jobID);
            JobManager jobM = jobManagerMap.get(jobID);
            jobM.taskCompleted(taskAttID, taskStatus);
        }

        @Override
        public void outputRetrieved(JobTask reduceJT, boolean completionState) {
            String jobID = reduceJT.getTask().getJobID();
            JobManager jobM = jobManagerMap.get(jobID);
            jobM.outputRetrieved(reduceJT, completionState);
        }

        @Override
        public void jobCompleted(String jobID, JobStatus.State jobState) {
            LOG.log(Level.INFO, "Job {0} completed with status {1}", new Object[]{jobID, jobState});
            PrimaryJob pj = nodeFSM.getNodeModule().removePrimaryJob(jobID);
            JobManager jM = jobManagerMap.remove(jobID);
            jM.stop();
            updateLoad(-1);

            //Propagate msg to backup masters
            MessageEvent bjCompletedMsg = new MessageEvent(EventType.BACKUP_JOB_COMPLETED, jobID);
            LOG.log(Level.INFO, "Sending completion msg to BackupMaster nodes");
            for (String nodeID : pj.getBackupMasterList()) {
                try {
                    nodeFSM.getNodeModule().getNetworkModule().send(nodeID, bjCompletedMsg);

                } catch (IOException ex) {
                    ex.printStackTrace();
                    LOG.log(Level.SEVERE, null, ex);
                }
            }

            MessageEvent jCompletedMsg = new MessageEvent(EventType.JOB_COMPLETED, jobID, jobState);
            //Propagate msg to slaves
            //TODO optimization: should not consider failed (or blacklisted) nodes
            //for now we use JobManager.isBlackListed method 
            //NOTE: slaveID is not null also for failed jobtask ? 
            LOG.log(Level.INFO, "Sending completion msg to slaves nodes");
            Set<String> notified = new HashSet<String>();
            for (JobTask jobTask : pj.getJobTaskList()) {
                String slaveID = jobTask.getSlaveID();
                if (slaveID != null && !notified.contains(slaveID) && !jM.isBlackListed(slaveID)) {
                    try {
                        notified.add(slaveID);
                        nodeFSM.getNetworkModule().send(slaveID, jCompletedMsg);
                    } catch (IOException ioex) {
                        ioex.printStackTrace();
                        LOG.log(Level.SEVERE, null, ioex);
                    }
                }
            }

            //Propagate msg to user
            LOG.log(Level.INFO, "Sending completion msg to user node");
            NodeInfo userNodeInfo = pj.getUserNodeInfo();
            try {
                nodeFSM.getNetworkModule().send(userNodeInfo, jCompletedMsg);
            } catch (IOException ex) {
                ex.printStackTrace();
                LOG.log(Level.SEVERE, null, ex);
            }

            if (nodeFSM.getNodeModule().getPrimaryJobList().isEmpty()) {
                LOG.log(Level.INFO, "No primary jobs available.Transition to NOT_PRIMARY");
                transition(NOT_PRIMARY);
            }

        }
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Generic inner classes
    ///////////////////////////////////////////////////////////////////////////
    private class TimeoutTask extends TimerTask {

        @Override
        public void run() {
            timeout(null);
        }
    }
}
