/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule.nodeFsm;

import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.JobStatus;
import p2pMapReduce.nodeModule.Constants;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.entity.BackupJob;
import p2pMapReduce.entity.JobTask;
import p2pMapReduce.entity.PrimaryJob;
import p2pMapReduce.entity.Task.TaskStatus;

/**
 * A FSM macro state representing a master node.
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class Master extends NodeState {

    private static final transient Logger LOG = Logger.getLogger(Master.class.getName());
    private NodeFSM nodeFSM;
    private ManagementFSM managementFSM;
    private RecoveryFSM recoveryFSM;
    private CoordinationFSM coordinationFSM;

    public Master(NodeFSM aThis) {
        this.nodeFSM = aThis;
        coordinationFSM = new CoordinationFSM(nodeFSM);
        managementFSM = new ManagementFSM(nodeFSM);
        recoveryFSM = new RecoveryFSM(nodeFSM);
    }

    @Override
    protected void entryAction() {
        LOG.log(Level.INFO, "Entering MASTER macro state");
        nodeFSM.getNodeModule().getNodeInfo().setRole(Constants.MASTER);
        nodeFSM.getNodeModule().getNetworkModule().publish();
        coordinationFSM.transition(coordinationFSM.NOT_COORDINATOR);
        managementFSM.transition(managementFSM.NOT_PRIMARY);
        recoveryFSM.transition(recoveryFSM.NOT_BACKUP);
    }

    @Override
    public void becomeSlave() {

        if (idle()) {
            LOG.log(Level.INFO, "Received becomeSlave message in MASTER.IDLE.Transition to SLAVE.IDLE");
            nodeFSM.transition(nodeFSM.IDLE);
            nodeFSM.getNodeModule().getNetworkModule().becomeSlave();
        } else {
            LOG.log(Level.INFO, "Received becomeSlave message but not in MASTER.IDLE state");
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Coordination events
    ///////////////////////////////////////////////////////////////////////////
    @Override
    public void electionMessage(String senderID) {
        //Only coordination state could by affected by this event
        coordinationFSM.electionMessage(senderID);
    }

    @Override
    public void newCoordinator(String senderID) {
        //Only coordination state could by affected by this event
        coordinationFSM.newCoordinator(senderID);
    }

    @Override
    public void coordinatorFailure() {
        //Only coordination state could by affected by this event
        coordinationFSM.coordinatorFailure();
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Managment event 
    ///////////////////////////////////////////////////////////////////////////
    @Override
    public void jobIDRequest(int requestNumber, NodeInfo userNodeInfo) {
        managementFSM.jobIDRequest(requestNumber, userNodeInfo);
    }

    @Override
    public void jobAssigned(Job job, NodeInfo userNodeInfo) {
        managementFSM.jobAssigned(job, userNodeInfo);
    }

    @Override
    public void slaveCheck(String taskID) {
        managementFSM.slaveCheck(taskID);
    }

    @Override
    public void jobReassigned(BackupJob bJ) {
        managementFSM.jobReassigned(bJ);
    }

    /**
     * The MANGMENT macro state can receive a taskCompleted message that must be
     * propagate to the relative JobManager
     *
     * @param taskAttID
     * @param taskStatus
     */
    @Override
    public void taskCompleted(String taskAttID, TaskStatus taskStatus) {
        managementFSM.taskCompleted(taskAttID, taskStatus);
    }

    @Override
    public void outputRetrieved(JobTask reduceJT, boolean completionState) {
        managementFSM.outputRetrieved(reduceJT, completionState);
    }

    @Override
    public void jobCompleted(String jobID, JobStatus.State jobState) {
        managementFSM.jobCompleted(jobID, jobState);
    }

    @Override
    public void slaveFailure(String failedSlaveID) {
        managementFSM.slaveFailure(failedSlaveID);
    }

    @Override
    public void backupMasterFailure(String failedBackupID) {
        managementFSM.backupMasterFailure(failedBackupID);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Recovery events
    ///////////////////////////////////////////////////////////////////////////
    @Override
    public void primaryMasterFailure(String primaryID) {
        //Failure of a primary node notified.
        //Surely this node is a backup node of the  failed primary node
        recoveryFSM.primaryMasterFailure(primaryID);
    }

    @Override
    public void backupJobAssigned(PrimaryJob primaryJob, String primaryNodeID) {
        recoveryFSM.backupJobAssigned(primaryJob, primaryNodeID);
    }

    @Override
    public synchronized void primaryElectionMessage(String jobID, String senderID, double senderLoad) {
        recoveryFSM.primaryElectionMessage(jobID, senderID, senderLoad);
    }

    @Override
    public synchronized void primaryUpdate(String jobID, String updatedPrimaryID, double updatedPrimaryLoad) {
        recoveryFSM.primaryUpdate(jobID, updatedPrimaryID, updatedPrimaryLoad);
    }

    @Override
    public void jobUpdate(String jobID, LinkedList<String> updatedBackupMasterList, LinkedList<JobTask> updatedJobTaskList) {
        recoveryFSM.jobUpdate(jobID, updatedBackupMasterList, updatedJobTaskList);
    }

    @Override
    public void backupJobCompleted(String completedJobID) {
        recoveryFSM.backupJobCompleted(completedJobID);
    }

    /**
     * Indicates if this master node is in idle macro state, or it is in
     * NOT_COORDINATOR, NOT_PRIMARY and NOT_BACKUP states
     * 
     * @return true if this master node is in idle state
     */
    private boolean idle() {
        return coordinationFSM.currentState() == coordinationFSM.NOT_COORDINATOR
                && managementFSM.currentState() == managementFSM.NOT_PRIMARY
                && recoveryFSM.currentState() == recoveryFSM.NOT_BACKUP;
    }
}
