/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule.nodeFsm;

import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.JobStatus;
import p2pMapReduce.networkModule.ConnectionIF;
import p2pMapReduce.networkModule.NetworkModuleIF;
import p2pMapReduce.nodeModule.Constants;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.nodeModule.event.EventListener;
import p2pMapReduce.nodeModule.NodeModuleIF;
import p2pMapReduce.entity.BackupJob;
import p2pMapReduce.entity.JobTask;
import p2pMapReduce.entity.PrimaryJob;
import p2pMapReduce.entity.SlaveTask;
import p2pMapReduce.entity.Task;
import p2pMapReduce.entity.Task.TaskStatus;
import p2pMapReduce.nodeModule.event.EventIF;
import p2pMapReduce.nodeModule.event.EventType;
import p2pMapReduce.nodeModule.event.InnerEvent;
import p2pMapReduce.nodeModule.nodeFsm.NodeState.StateName;

/**
 * A finite state machine that implements the node behavior.
 * 
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class NodeFSM implements EventListener {

    private static final Logger LOG = Logger.getLogger(NodeFSM.class.getName());
    /**
     * Simple states
     */
    final NodeState IDLE;
    final NodeState CHECK_MASTER;
    final NodeState ACTIVE;
    /**
     * Composite(parallel) state
     */
    final NodeState MASTER;
    //TODO eliminare ed estendere FSM ???
    private FSM fsm;
    private NodeModuleIF nodeModule;

    public NodeFSM(NodeModuleIF nodeMod) {
        this.nodeModule = nodeMod;
        IDLE = new Idle(this);
        CHECK_MASTER = new Check_Master(this);
        ACTIVE = new Active(this);
        MASTER = new Master(this);
        fsm = new FSM();
        //starts the node fsm mechanism
        fsm.transition(IDLE);
    }

    public NodeModuleIF getNodeModule() {
        return nodeModule;
    }

    public NetworkModuleIF getNetworkModule() {
        return nodeModule.getNetworkModule();
    }

    /**
     * The entry method to submit an event to handle.
     * The event handling is being in mutual exclusion in order to
     * avoid leaving the node state inconsistent.
     * @param event
     */
    public synchronized void handleEvent(EventIF event) {
        LOG.log(Level.INFO, "Received new event: {0}", event);
        switch (event.getType()) {
            case TASK_ASSIGNED:
                taskAssigned((LinkedList<Task>) event.getParam(0), (String) event.getParam(1));
                break;
            case TASK_COMPLETED:
                taskCompleted((String) event.getParam(0), (TaskStatus) event.getParam(1));
                break;
            case BECOME_MASTER:
                becomeMaster((String) event.getParam(0));
                break;
            case BECOME_SLAVE:
                becomeSlave();
                break;
            case COORD_ELECTION_MESSAGE:
                electionMessage((String) event.getParam(0));
                break;
            case NEW_COORDINATOR:
                newCoordinator((String) event.getParam(0));
                break;
            case COORDINATOR_FAILURE:
                coordinatorFailure();
                break;
            case SLAVE_FAILURE:
                slaveFailure((String) event.getParam(0));
                break;
            case BACKUP_MASTER_FAILURE:
                backupMasterFailure((String) event.getParam(0));
                break;
            case JOB_ID_REQUEST:
                jobIDRequest((Integer) event.getParam(0), (NodeInfo) event.getParam(1));
                break;
            case JOB_ASSIGNED:
                jobAssigned((Job) event.getParam(0), (NodeInfo) event.getParam(1));
                break;
            case JOB_REASSIGNED:
                jobReassigned((BackupJob) event.getParam(0));
                break;
            case SLAVE_CHECK:
                slaveCheck((String) event.getParam(0));
                break;
            case OUTPUT_RETRIEVING_COMPLETED:
                outputRetrieved((JobTask) event.getParam(0), (Boolean) event.getParam(1));
                break;
            case JOB_COMPLETED:
                jobCompleted((String) event.getParam(0), (JobStatus.State) event.getParam(1));
                break;
            case BACKUP_JOB_ASSIGNED:
                backupJobAssigned((PrimaryJob) event.getParam(0), (String) event.getParam(1));
                break;
            case PRIMARY_FAILURE:
                primaryMasterFailure((String) event.getParam(0));
                break;
            case PRIMARY_ELECTION_MESSAGE:
                primaryElectionMessage((String) event.getParam(0), (String) event.getParam(1), (Double) event.getParam(2));
                break;
            case PRIMARY_UPDATE:
                primaryUpdate((String) event.getParam(0), (String) event.getParam(1), (Double) event.getParam(2));
                break;
            case JOB_UPDATE:
                jobUpdate((String) event.getParam(0), (LinkedList<String>) event.getParam(1), (LinkedList<JobTask>) event.getParam(2));
                break;
            case BACKUP_JOB_COMPLETED:
                backupJobCompleted((String) event.getParam(0));
                break;
            case NODE_FAILURE:
                handleNodeFailure(event);
                break;
            case CRASH_EVENT:
                handleCrashEvent(event);
                break;
            case NEW_CONNECTION:
                //nothing to do
                break;
            default:
                LOG.log(Level.WARNING, "Event type not handled = {0}", event.getType());

        }
    }

    public synchronized void transition(State nextState) {
        fsm.transition(nextState);
    }

    public State getCurrentState() {
        return fsm.currentState();
    }

    //FIXME devono essere per forza sinc?? forse si per garantire
    //tranisizioni di stato consistenti
    //TODO tutti questi metodi dovrebbero non essere utilizzabili all'esterno
    //direttamente per cui da rendere protected
    protected synchronized void timeout(Object param) {
        //TODO gli eventi interni di time out dovrebbero essere notificati
        //tutti tramite questo metodo per garantire la mutua esclusione
        ((NodeState) fsm.currentState()).timeout(param);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Slave macro state events
    ///////////////////////////////////////////////////////////////////////////
    public synchronized void taskAssigned(LinkedList<Task> assignedTaskList, String primaryID) {
        ((NodeState) fsm.currentState()).taskAssigned(assignedTaskList, primaryID);
    }

    public synchronized void taskCompleted(String taskID, TaskStatus taskStatus) {
        ((NodeState) fsm.currentState()).taskCompleted(taskID, taskStatus);
    }

    public synchronized void becomeMaster(String coordID) {
        ((NodeState) fsm.currentState()).becomeMaster(coordID);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Master macro state events
    ///////////////////////////////////////////////////////////////////////////
    public synchronized void becomeSlave() {
        ((NodeState) fsm.currentState()).becomeSlave();
    }

    ///////////////////////////////////////////////////////////////////////////
    //      Coordination macro state events
    ///////////////////////////////////////////////////////////////////////////
    public synchronized void electionMessage(String senderID) {
        ((NodeState) fsm.currentState()).electionMessage(senderID);
    }

    public synchronized void newCoordinator(String senderID) {
        ((NodeState) fsm.currentState()).newCoordinator(senderID);
    }

    public synchronized void coordinatorFailure() {
        ((NodeState) fsm.currentState()).coordinatorFailure();
    }

    public synchronized void slaveFailure(String failedSlaveID) {
        ((NodeState) fsm.currentState()).slaveFailure(failedSlaveID);
    }

    public synchronized void backupMasterFailure(String failedBackupID) {
        ((NodeState) fsm.currentState()).backupMasterFailure(failedBackupID);
    }

    ///////////////////////////////////////////////////////////////////////////
    //      Management macro state events.
    ///////////////////////////////////////////////////////////////////////////
    public synchronized void jobIDRequest(Integer requestNumber, NodeInfo nodeInfo) {
        ((NodeState) fsm.currentState()).jobIDRequest(requestNumber, nodeInfo);
    }

    public synchronized void jobAssigned(Job job, NodeInfo userNodeInfo) {
        ((NodeState) fsm.currentState()).jobAssigned(job, userNodeInfo);
    }

    public synchronized void jobReassigned(BackupJob bJ) {
        ((NodeState) fsm.currentState()).jobReassigned(bJ);
    }

    public synchronized void slaveCheck(String taskID) {
        ((NodeState) fsm.currentState()).slaveCheck(taskID);
    }

    private synchronized void outputRetrieved(JobTask jobTask, Boolean completionState) {
        ((NodeState) fsm.currentState()).outputRetrieved(jobTask, completionState);
    }

    public synchronized void jobCompleted(String jobID, JobStatus.State jobState) {
        ((NodeState) fsm.currentState()).jobCompleted(jobID, jobState);
    }

    ///////////////////////////////////////////////////////////////////////////
    //      Recovery macro state events
    ///////////////////////////////////////////////////////////////////////////
    public synchronized void backupJobAssigned(PrimaryJob primaryJob, String primaryNodeID) {
        ((NodeState) fsm.currentState()).backupJobAssigned(primaryJob, primaryNodeID);
    }

    public synchronized void primaryMasterFailure(String primaryID) {
        ((NodeState) fsm.currentState()).primaryMasterFailure(primaryID);
    }

    public synchronized void primaryElectionMessage(String jobID, String senderID, double senderLoad) {
        ((NodeState) fsm.currentState()).primaryElectionMessage(jobID, senderID, senderLoad);
    }

    public synchronized void primaryUpdate(String updatedJobID, String primaryNodeID, double primaryLoad) {
        ((NodeState) fsm.currentState()).primaryUpdate(updatedJobID, primaryNodeID, primaryLoad);
    }

    public synchronized void jobUpdate(String jobID, LinkedList<String> updatedBackupMasterList, LinkedList<JobTask> updatedJobTaskList) {
        ((NodeState) fsm.currentState()).jobUpdate(jobID, updatedBackupMasterList, updatedJobTaskList);
    }

    public synchronized void backupJobCompleted(String completedJobID) {
        ((NodeState) fsm.currentState()).backupJobCompleted(completedJobID);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Private methods
    ///////////////////////////////////////////////////////////////////////////
    /**
     * Translate a NODE_FAILURE event into one of SLAVE_FAILURE,MASTER_FAILURE
     * or COORDINATOR_FAILURE event
     * @param event
     */
    private void handleNodeFailure(EventIF event) {
        String failedNodeID = (String) event.getParam(0);
        LOG.log(Level.INFO, "Detected node failure: " + failedNodeID);

        if (nodeModule.getNodeInfo().getRole() == Constants.MASTER) {
            handleNodeFailureAsMaster(failedNodeID);

        } else if (nodeModule.getNodeInfo().getRole() == Constants.SLAVE) {
            handleNodeFailureAsSlave(failedNodeID);
        }

    }

    private void handleNodeFailureAsSlave(String failedNodeID) {
        //if i'm a slave node i'm only interested in primary failure for my SlaveTask
        for (SlaveTask st : nodeModule.getSlaveTaskMap().values()) {
            if (failedNodeID.equals(st.getPrimaryID())) {
                InnerEvent primaryMasterFailureEv = new InnerEvent(EventType.PRIMARY_FAILURE, failedNodeID);
                handleEvent(primaryMasterFailureEv);
            }
        }
    }

    private void handleNodeFailureAsMaster(String failedNodeID) {
        //Check if Coordinator failure first
        if (failedNodeID.equals(nodeModule.getCoordinatorID())) {
            nodeModule.setCoordinatorID(null);
            InnerEvent coordFailureEv = new InnerEvent(EventType.COORDINATOR_FAILURE);
            handleEvent(coordFailureEv);
        }
        //1.if i am a Master i am interested in backup master failure for my PrimaryJobs
        for (PrimaryJob pj : nodeModule.getPrimaryJobList()) {
            if (pj.getBackupMasterList().contains(failedNodeID)) {
                InnerEvent backupMasterFailureEv = new InnerEvent(EventType.BACKUP_MASTER_FAILURE, failedNodeID);
                handleEvent(backupMasterFailureEv);
                //one backup failure notification is enough
                //the management state will multiplex the event
                break;
            }
        }
        //2. if i'm a Master i'm also interested in primary failure for my BackupJobs
        for (BackupJob bj : nodeModule.getBackupJobList()) {
            if (failedNodeID.equals(bj.getPrimaryID())) {
                InnerEvent primaryMasterFailure = new InnerEvent(EventType.PRIMARY_FAILURE, failedNodeID);
                handleEvent(primaryMasterFailure);
                //one primary failure notification is enough
                //the recovery state will multiplexing the event
                break;
            }
        }
        //3.if i'm a Master i'm also interested in slave failure for my PrimaryJobs
        for (PrimaryJob pj : nodeModule.getPrimaryJobList()) {
            for (JobTask jt : pj.getJobTaskList()) {
                if (failedNodeID.equals(jt.getSlaveID())) {
                    InnerEvent slaveFailureEv = new InnerEvent(EventType.SLAVE_FAILURE, failedNodeID);
                    handleEvent(slaveFailureEv);
                    //one slave failure notification is enough
                    //the management state will multiplexing the event
                    return;
                }
            }
        }
    }

    private void handleCrashEvent(EventIF event) {
        ConnectionIF conn = (ConnectionIF) event.getParam(0);
        String nodeID = conn.getRemoteNodeInfo().getID();
        LOG.log(Level.INFO, "Detected crash event on the connection toward : " + nodeID);
        //FIXME Considering crash event like node failure event.
        handleNodeFailure(new InnerEvent(EventType.NODE_FAILURE, nodeID));

    }

    /**
     * XXX Dangerous forcing of state transition
     * Only for testing purpose, will be removed
     * @deprecated 
     */
    public synchronized void forceTransition(StateName stateName) {

        switch (stateName) {
            case IDLE:
                transition(IDLE);
                break;
            case CHECK_MASTER:
                transition(CHECK_MASTER);
                break;
            case ACTIVE:
                transition(ACTIVE);
                break;
            case MASTER:
                transition(MASTER);
                break;
            default:
                throw new AssertionError();
        }
    }
}
