/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule.nodeFsm;

import java.util.LinkedList;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.mapreduceModule.Job;
import p2pMapReduce.mapreduceModule.JobID;
import p2pMapReduce.mapreduceModule.JobStatus;
import p2pMapReduce.nodeModule.JobManager;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.entity.BackupJob;
import p2pMapReduce.entity.JobTask;
import p2pMapReduce.entity.PrimaryJob;
import p2pMapReduce.entity.Task;

/**
 * A generic {@link NodeFSM} state.
 * 
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public abstract class NodeState extends State {

    public static enum StateName {

        IDLE,
        CHECK_MASTER,
        ACTIVE,
        MASTER
    }
    private static final transient Logger LOG = Logger.getLogger(NodeState.class.getName());

    public static int getNextTimeout(int minTimeout, int maxTimeout) {
        Random r = new Random();
        return minTimeout + r.nextInt(maxTimeout - minTimeout);
    }

    private static void logWarning(String name) {
        StackTraceElement stackTraceElems[] = Thread.currentThread().getStackTrace();
        StringBuilder sb = new StringBuilder();
        sb.append("Calling a NodeState method: ").append(name).append(" from\n");
        for (int i = 1; i < stackTraceElems.length; i++) {
            StackTraceElement ste = stackTraceElems[i];
            sb.append(ste.toString()).append("\n");

        }

        LOG.log(Level.WARNING, sb.toString());
    }

    /**
     * A generic timeout method.
     * The param object is useful to handle different timeout event
     * @param param
     */
    public void timeout(Object param) {
        logWarning("timeout");
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Slave macro state events
    ///////////////////////////////////////////////////////////////////////////
    public void taskAssigned(LinkedList<Task> assignedTaskList, String primaryID) {
        logWarning("taskAssigned");
    }

    public void taskCompleted(String taskAttID, Task.TaskStatus taskStatus) {
        logWarning("taskCompleted");
    }

    public void becomeMaster(String coordID) {
        logWarning("becomeMaster");
    }

    ///////////////////////////////////////////////////////////////////////////
    // Master macro state events
    ///////////////////////////////////////////////////////////////////////////
    public void becomeSlave() {
        logWarning("becomeSlave");
    }

    //Coordination macro state event
    /**
     * Invoked when an election message is received from another master
     * In ELECTING_COORDINATOR state from a master which is taking part in
     * the election algorithm
     * In WAITING_COORDINATOR state from the new coordinator
     * @param senderID The node id of the message sender
     */
    public void electionMessage(String senderID) {
        logWarning("electionMessage");
    }

    /**
     * Invoked when a new coordinator id is received by another master.
     * @param senderID The node id of the message sender that has become the
     *        new Coordinator
     */
    public void newCoordinator(String senderID) {
        logWarning("newCoordinator");
    }

    public void coordinatorFailure() {
        logWarning("coordinatorFailure");
    }

    public void slaveFailure(String failedSlaveID) {
        logWarning("slaveFailure");
    }

    public void backupMasterFailure(String failedBackupID) {
        logWarning("backupMasterFailure");
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Managment macro state event
    ///////////////////////////////////////////////////////////////////////////
    /**
     * Used by an user node to request a new unique {@link JobID }
     * 
     * @param userNodeInfo
     */
    public void jobIDRequest(int requestNumber, NodeInfo userNodeInfo) {
        logWarning("jobIDRequest");
    }

    public void jobAssigned(Job job, NodeInfo userNodeInfo) {
        logWarning("jobAssigned");
    }

    public void jobReassigned(BackupJob bJ) {
        logWarning("jobReassigned");
    }

    public void jobCompleted(String jobID, JobStatus.State jobState) {
        logWarning("jobCompleted");
    }

    public void slaveCheck(String taskID) {
        logWarning("slaveCheck");
    }

    /**
     * Received by {@link JobManager}.
     * Notify the success state of {@code reduceJT} output retrieving by user
     * node.
     * 
     * @param reduceJT
     * @param completionState
     */
    public void outputRetrieved(JobTask reduceJT, boolean completionState) {
        logWarning("outputRetrieved");
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Recovery macro state events
    ///////////////////////////////////////////////////////////////////////////
    public void backupJobAssigned(PrimaryJob primaryJob, String primaryNodeID) {
        logWarning("backupJobAssigned");
    }

    public synchronized void primaryMasterFailure(String primaryID) {
        logWarning("primaryMasterFailure");
    }

    public synchronized void primaryElectionMessage(String jobID, String senderID, double senderLoad) {
        logWarning("primaryElectionMessage");
    }

    public synchronized void primaryUpdate(String jobID, String updatedPrimaryID, double updatedPrimaryLoad) {
        logWarning("primaryUpdate");
    }

    public void jobUpdate(String jobID, LinkedList<String> updatedBackupMasterList, LinkedList<JobTask> updatedJobTaskList) {
        logWarning("jobUpdate");
    }

    public void backupJobCompleted(String completedJobID) {
        logWarning("backupJobCompleted");
    }
}
