/*
 * P2P-MapReduce
 *
 * Copyright (C) 2011 University of Calabria
 * Dept. of Electronics, Computer Science and Systems
 */
package p2pMapReduce.nodeModule.nodeFsm;

import p2pMapReduce.nodeModule.BackupJobManager;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;
import p2pMapReduce.networkModule.NodeInfo;
import p2pMapReduce.entity.BackupJob;
import p2pMapReduce.entity.JobTask;
import p2pMapReduce.entity.PrimaryJob;
import p2pMapReduce.util.LoggingUtil;

/**
 * Inner state of {@link Master} macro state, that take care
 * of jobs backup and recovery
 *
 * @author Francesco De Luca f.deluca86@gmail.com
 */
public class RecoveryFSM extends FSM {

    private static final Logger LOG = Logger.getLogger(ManagementFSM.class.getName());
    final RecoveryState NOT_BACKUP;
    final RecoveryState BACKUP;
    private NodeFSM nodeFSM;
    /**
     * For each job id holds the relative backup job manager
     */
    private HashMap<String, BackupJobManager> backupJobManagerMap;
    //TODO valutare utilizzo ScheduledThreadPoolExecutor
    Timer timer = new Timer("RecoveryFSM Timer", true);
    TimeoutTask timeoutTask = new TimeoutTask();

    public RecoveryFSM(NodeFSM nodeFSM) {
        this.nodeFSM = nodeFSM;
        NOT_BACKUP = new NotBackup();
        BACKUP = new Backup();
        backupJobManagerMap = new HashMap<String, BackupJobManager>();
    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM methods
    ///////////////////////////////////////////////////////////////////////////
    public synchronized void timeout(Object param) {
        ((RecoveryState) currentState()).timeout(param);
    }

    public synchronized void backupJobAssigned(PrimaryJob primaryJob, String primaryNodeID) {
        ((RecoveryState) currentState()).backupJobAssigned(primaryJob, primaryNodeID);
    }

    public synchronized void primaryMasterFailure(String primaryID) {
        ((RecoveryState) currentState()).primaryMasterFailure(primaryID);
    }

    public synchronized void primaryElectionMessage(String jobID, String senderID, double senderLoad) {
        ((RecoveryState) currentState()).primaryElectionMessage(jobID, senderID, senderLoad);
    }

    public synchronized void primaryUpdate(String jobID, String updatedPrimaryID, double updatedPrimaryLoad) {
        ((RecoveryState) currentState()).primaryUpdate(jobID, updatedPrimaryID, updatedPrimaryLoad);
    }

    public synchronized void jobUpdate(String jobID, LinkedList<String> updatedBackupMasterList, LinkedList<JobTask> updatedJobTaskList) {
        ((RecoveryState) currentState()).jobUpdate(jobID, updatedBackupMasterList, updatedJobTaskList);
    }

    public synchronized void backupJobCompleted(String completedJobID) {
        ((RecoveryState) currentState()).backupJobCompleted(completedJobID);
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Private methods
    ///////////////////////////////////////////////////////////////////////////
    private void startBackupJobManager(BackupJob bJ) {
        BackupJobManager bjManager = new BackupJobManager(nodeFSM, bJ);
        backupJobManagerMap.put(bJ.getJobID(), bjManager);
        bjManager.start();
    }

    private void updateLoad(int numTasks) {
        NodeInfo thisNode = nodeFSM.getNodeModule().getNodeInfo();
        thisNode.setLoad(thisNode.getLoad() + numTasks);
        LOG.log(Level.INFO, "Current load=" + thisNode.getLoad());
        nodeFSM.getNetworkModule().publish();
    }

    ///////////////////////////////////////////////////////////////////////////
    //  FSM state classes
    ///////////////////////////////////////////////////////////////////////////
    private class RecoveryState extends NodeState {

        @Override
        public void backupJobAssigned(PrimaryJob primaryJob, String primaryNodeID) {
            LOG.log(Level.INFO, "New backup job assigned: {0} from {1}", new Object[]{primaryJob, primaryNodeID});
            BackupJob bJ = new BackupJob(primaryJob, primaryNodeID);
            nodeFSM.getNodeModule().addBackupJob(bJ);
            startBackupJobManager(bJ);
            updateLoad(1);
        }
    }

    private class NotBackup extends RecoveryState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, "Entering RECOVERY.NOT_BACKUP");
        }

        @Override
        public void backupJobAssigned(PrimaryJob primaryJob, String primaryNodeID) {
            super.backupJobAssigned(primaryJob, primaryNodeID);
            transition(BACKUP);
        }
    }

    private class Backup extends RecoveryState {

        @Override
        protected void entryAction() {
            LOG.log(Level.INFO, "Entering RECOVERY.BACKUP");
        }

        @Override
        public void backupJobAssigned(PrimaryJob primaryJob, String primaryNodeID) {
            super.backupJobAssigned(primaryJob, primaryNodeID);
        }

        @Override
        public synchronized void primaryMasterFailure(String primaryID) {
            LOG.log(Level.INFO, "Primary master failure for node {0}", new Object[]{primaryID});
            //To avoid concurrent modification in consequence of calling backupJobCompleted when a
            //backupJobManager becomes the new primary
            Set<BackupJobManager> bjManagers = new HashSet<BackupJobManager>();
            for (BackupJobManager bjM : backupJobManagerMap.values()) {
                if (primaryID.equals(bjM.getPrimaryID())) {
                    bjManagers.add(bjM);
                }
            }

            for (BackupJobManager bjM : bjManagers) {
                bjM.primaryMasterFailure(primaryID);
            }
        }

        @Override
        public synchronized void primaryElectionMessage(String jobID, String senderID, double senderLoad) {
            LOG.log(Level.INFO, "Primary election message received for job {0} from {1}", new Object[]{jobID, senderID});
            BackupJobManager bJManager = backupJobManagerMap.get(jobID);
            bJManager.primaryElectionMessage(jobID, senderID, senderLoad);
        }

        @Override
        public void primaryUpdate(String jobID, String updatedPrimaryID, double updatedPrimaryLoad) {
            LOG.log(Level.INFO, "Primary update received for {0} from {1}", new Object[]{jobID, updatedPrimaryID});
            BackupJobManager bJManager = backupJobManagerMap.get(jobID);
            bJManager.primaryUpdate(jobID, updatedPrimaryID, updatedPrimaryLoad);
        }

        @Override
        public void jobUpdate(String jobID, LinkedList<String> updatedBackupMasterList, LinkedList<JobTask> updatedJobTaskList) {
            LOG.log(Level.INFO, "Job update received for {0}" + " updateBackupMasterList:\n{1}\nupdatedJobList:\n{2}",
                    new Object[]{jobID, LoggingUtil.elementByLine(updatedBackupMasterList), LoggingUtil.elementByLine(updatedJobTaskList)});
            //TODO spostare in bJManager
            BackupJob bJ = nodeFSM.getNodeModule().getBackupJob(jobID);
            bJ.setBackupMasterList(updatedBackupMasterList);
            for (JobTask updatedJobTask : updatedJobTaskList) {
                bJ.update(updatedJobTask);
            }
            bJ.updateProgress();

        }

        @Override
        public void backupJobCompleted(String completedJobID) {
            LOG.log(Level.INFO, "BackupJob completion received for: " + completedJobID);
            //FIXME bj is already removed ?
            BackupJob bJ = nodeFSM.getNodeModule().removeBackupJob(completedJobID);
            BackupJobManager bJManager = backupJobManagerMap.remove(completedJobID);
            bJManager.stop();
            updateLoad(-1);
            if (nodeFSM.getNodeModule().getBackupJobList().isEmpty()) {
                LOG.log(Level.INFO, "No more backup job managed.Transition to RECOVERY.NOT_BACKUP");
                transition(NOT_BACKUP);
            }

        }
    }

    ///////////////////////////////////////////////////////////////////////////
    //  Generic inner classes
    ///////////////////////////////////////////////////////////////////////////
    private class TimeoutTask extends TimerTask {

        @Override
        public void run() {
            timeout(null);
        }
    }
}
